{Random.pas - random numbers for Delphi 1, 2, 3 (32bits)}
{Copyright (c) 1997 Engineering Objects International}
{Generated 4/23/97 11:33:21 AM   by   ClassBuilder 4.0 for Delphi}

{Version 2.0 07/20/1997}

{$X+ extended syntax}
{$Q- overflow checking off}
{$R- range checking off; all 32bits are fully used.}

{$IFDEF Win32}
  {$OPTIMIZATION  ON}
{$ENDIF}

unit random;

(*The classes form a shallow hierarchy for generating random numbers:

  TeoRandomBase   the abstract interface to the classes. (DO NOT INSTANTIATE)

    TeoRandomGood    Park & Miller minimal standard congruential generator.
    TeoRandomBetter  Park & Miller with a Bays & Durham shuffle
    TeoRandomBest    L'Ecuyer's two-series combo plus a shuffle for a period > 2 e18
    TeoRandomQAD     Quick and dirty series of 'not very random' numbers

    TeoRandomMarsagliaBase    Base class for the two Marsaglia PRNG (DO NOT INSTANTIATE)
      TeoRandomMarsagliaFast  fast version of the Marsaglia generator
      TeoRandomMarsaglia      top-quality version of the Marsaglia generator.


How to use the Object Pascal classes (see notes for C++Builder below)
=====================================================================
1.  Choose a class that meets your needs, say TeoRandomGood.

2.  Include the 'Random' unit in the 'uses' clause of your unit.

3.  Declare a variable of the class (or, declare it as a field
    of your own class), say 'r'.

4.  Pick a positive integer value to seed the generator. One value
    is usually as good as another, but if you will want to later restart
    the sequence at the same point, you must 'remember' the seed value.

5.  Create the random number generator with the statement
       r := TeoRandomGood.Create(seed);

6.  As you need 'random' doubles in the range 0.0 .. 1.0
   (endpoints excluded), access the Value method:
        aRandomNumber := r.Value;

7.  To restart the sequence with another seed, use
      r.Reset(anotherseed);

8.  To begin a new sequence in the same state as r, use the Clone
    constructor: p := TeoRandomGood.Clone(r); r and p are now two
    independent generators starting from the same point.

9.  To set an instantiated generator (say, r) to the same state
    as another (say, p), use the Copy method:
      r.Copy(p);
    r and p are now two independent generators starting from
    the same point.

10. To save the generator for a later restart from a stream,
    use the Put method. Assuming 'S' is opened with the statement
      S := TFileStream.Create('saveit.str',fmCreate);
    the complete code to save the random number generator at its
    current state is
      r.Put(S);

11. To create a new generator from one saved to a stream, call the
    Get constructor on an opened stream. Assuming 'S' is opened
    with the statement
      S := TFileStream.Create('saveit.str',fmOpenRead);
    the complete code to construct the random number generator
    at its previous state is
      r := TeoRandomGood.Get(S);


C++Builder notes
================
1. The C++Builder framework allows the use of Object Pascal
   units. To add the suite to your project: From the
   Project | Add to project menu, open the Add to project dialog.
   Click on file type and select *.pas. Use the File Explorer to
   locate Random.pas. Select it, click Open. Random is now
   added to your project.

2. In the editor, activate the C++Builder .cpp file in which you
   wish to use Random. From the File | Include unit header menu,
   choose Random.

3. In your code, instantiate an object of the appropriate type
   following this simple but correct outline:

     TeoRandomGood *r = new TeoRandomGood(1); //  new instance

     double z = r->Value(); //  use instance

        // etc

     delete r;

   where the 'vcl' (read: 'pascal') classes must be instantiated
   as pointer to object.

4. To avoid forgetting the 'delete r;' statement, you might
   consider the 'resource allocation is initialization' rule,
   following Stroustrup:

      class MyRandom {
        TeoRandomGood *r;
      public:
        MyRandom(const int seed)
          {r = new TeoRandomGood(seed);}
        ~MyRandom()
          {delete r;}
        double Value()
          {return r->Value();}
      };

   and instantiate your class with

      MyRandom r(yourIntSeedHere);

   r will be destroyed automatically when control exits its scope.
   Check the .cpp and .h files included with the distribution for
   an example.

5. Notes for Object Pascal apply as above.


Caveats
=======
Note that in some classes, computational effort is traded off
for better statistical behavior.  That is, 'good' distributions
take more time & effort to compute. For example, the superior
statistical qualities of the Marsaglia PRNG are not obvious 
if you only need *short* streams of random numbers
(i.e. only 20-30 numbers). Instead, you might use TeoRandomQAD.

Take the names of the generators with a grain of salt; TeoRandomBest
is only 'best' in a limited set of circumstances. On the other hand,
the quick and dirty series IS quick and dirty.

There is a lot of code here, but it would be a mistake to extract
pieces from the unit. Instead, compile the whole thing; the Delphi
smart linker will only link in the code that you actually use.

For very-high-speed random numbers that may be acceptable to your
specific situation, see the references and other sources. For example,
the time of day can sometimes be used with success, if you only need
a single number.


References:
===========
  Press, Teukolsky, Vettering, and Flannery,
  "NUMERICAL RECIPES in FORTRAN", 2e, Cambridge University
  Press, 1992 (reprinted with corrections 1994), ch 7, 1.
  See the text for other random number generators.

  Original authors of the Marsaglia PRNG are
             Arif Zaman (arif@stat.fsu.edu) and
             George Marsaglia (geo@stat.fsu.edu).
  George Marsaglia is "the" guru for random number generators;
  this generator has been put to scientific scrutiny (see the
  comments for TeoRandomMarsaglia) and has been found to be
  of excellent statistical quality.

  See also comments in the code.
*)

INTERFACE
uses
  Classes;

  {IntegerType defined for various systems}
{$ifdef WIN32}
const
  MAXINTEGERTYPE = MAXINT;
type
  IntegerType = integer;
{$else  WINDOWS 3.x}
const
  MAXINTEGERTYPE = MAXLONGINT;
type
  IntegerType = longint;
{$endif}


  {
  // TeoRandomBase
  // Quasi abstract class whence to derive random number generators
  }

type
  TeoRandomBase = class
    constructor Create(const seed: integerType);
    constructor Clone(const aTeoRandomBase: TeoRandomBase);
    constructor Get(S: TStream);

    destructor Destroy; override;

    procedure Copy(const aTeoRandomBase: TeoRandomBase);
    procedure Put(S: TStream);

      {member functions}
    procedure Reset(const seed: integerType);  virtual;  abstract; {// restart sequence with 'seed'}
    function Seed: integerType;                     {//  returns (perhaps modified) seed}
    function Value: double;  virtual;  abstract; {//  random in (0..1), not [0..1]}

  protected {for use by derivers}
    procedure SetTheSeed(const seed: integerType);
    procedure SetTheValue(const aValue: integerType);
    function  TheValueIs: integerType;

  private
    theSeed: integerType;
    theValue: integerType;  {//  changes with each call}
  end;  {TeoRandomBase}


  {
  // TeoRandomGood
  // The Minimal Standard which a random number generator must meet
  // (ran0 of Numerical Recipes)
  }

  TeoRandomGood = class( TeoRandomBase )
    constructor Create(const seed: integerType);

    destructor Destroy; override;

      {member functions}
    procedure Reset(const seed: integerType);  override;
    function Value: double;  override;
  end;  {TeoRandomGood}


  {
  // TeoRandomBetter
  // The Bays & Durham algorithm shuffles the values produced
  // by the Minimal Standard random number generator
  // (ran1 of Numerical Recipes)
  }

const
  TeoRandomBetterTableSize = 32;

type
  TeoRandomBetterTableArray = array[1..TeoRandomBetterTableSize] of integerType;

type
  TeoRandomBetter = class( TeoRandomBase )
    constructor Create(const seed: integerType);
    constructor Clone(const aTeoRandomBetter: TeoRandomBetter);
    constructor Get(S: TStream);

    destructor Destroy; override;

    procedure Copy(const aTeoRandomBetter: TeoRandomBetter);
    procedure Put(S: TStream);

      {member functions}
    procedure Reset(const seed: integerType);  override;
    function Value: double;  override;

  private
    auxValue: integerType;
    Table: TeoRandomBetterTableArray;
    function AuxValueIs: integerType;  {declared, but never used}
    function AuxValueSet(const aInteger: integerType): integerType;  {declared, but never used}
    function Calc: integerType;
  end;  {TeoRandomBetter}

  {
  // TeoRandomBest
  // Combined series of L'Ecuyer plus the Bays & Durham
  // shuffle produces a long period sequence (> 2 * 1.0e18).
  // (ran2 of Numerical Recipes)
  }

type
  TeoRandomBest = class( TeoRandomBetter )
    constructor Create(const seed: integerType);
    constructor Clone(const aTeoRandomBest: TeoRandomBest);
    constructor Get(S: TStream);

    destructor Destroy; override;

    procedure Copy(const aTeoRandomBest: TeoRandomBest);
    procedure Put(S: TStream);

      {member functions}
    procedure Reset(const seed: integerType);  override;
    function Value: double;  override;

  private
    yaValue: integerType;
    function Calc1: integerType;
    function Calc2: integerType;
    function YaValueIs: integerType;  {declared, but never used}
    function YaValueSet(const aInteger: integerType): integerType;  {declared, but never used}
  end;  {TeoRandomBest}



  {
  // TeoRandomQAD
  // Quick and Dirty random number generator
  // (ranqd2 of Numerical Recipes)
  // Good for short runs of numbers that
  // don't need to meet any statistical criteria
  }

  TeoRandomQAD = class( TeoRandomBase )
    constructor Create(const seed: integertype);

    destructor Destroy; override;

      {member functions}
    procedure Reset(const seed: integerType);  override;
    function Value: double;  override;
  end;  {TeoRandomQAD}



  { ===== Base class for the two Marsaglia PRNG =====

    Original authors of the PRNG are
               Arif Zaman (arif@stat.fsu.edu) and
               George Marsaglia (geo@stat.fsu.edu).
    George Marsaglia is "the" guru for random number generators;
    this generator has been put to scientific scrutiny (see the
    comments for TeoRandomMarsaglia) and has been found to be
    of excellent statistical quality.

    Do NOT use this PRNG if you only need *short* streams of
    random numbers (i.e. only 20-30 numbers). You'll waste time.
  }

  TeoRandomMarsagliaBase = class(TeoRandomBase)
  public
    { Creates PRNG and initializes it: ResetSWB(Seed, 7654321) }
    constructor Create(const Seed: integerType);

    { The Marsaglia generator takes TWO seeds as input,
      suggested values are: 1234567, 7654321 }
    constructor CreateSWB(const Seed, SWBseed: integerType);
    constructor Clone(const aTeoRandomMarsagliaBase: TeoRandomMarsagliaBase);
    constructor Get(S: TStream);

    procedure Copy(const aTeoRandomMarsagliaBase: TeoRandomMarsagliaBase);
    procedure Put(S: TStream);

     { member functions }

    procedure Reset(const Seed: integerType); override; {calls ResetSWB(Seed, 7654321)}
    procedure ResetSWB(const Seed, SWBseed: integerType); virtual;

    function SWBSeed: integerType;

  private
    theCarryFlag: byte;{ The carry flag for the SWB generator  }

    { The counter counts up to the table size to retrieve elements;
      if elements are no longer available the table is refilled }
    randomElement: integerType;

    { additional Seed for the SWB generator;
      this seed will NEVER change

      "theSeed: longint", a private member of the base class
      will not change either }
    theSWBseed: longint;

    { This array contains the seed mixed with a congruential. }
    { The generator will initialize the seed array with these numbers
      (commented out), if the generator itself is initialized with
      the parameters 1234567, 7654321 }
    SWBArray: array[0..37-1] of longint { =
     ($0D45D69A,
      $9DB73B1A, $A84604E8, $7C5F0CA5, $BC0196CE, $FF4CB42E, $7ACA6BE3,
      $A9ED2A5A, $6405A8F7, $AC00D4F8, $BDD1FC77, $064F9DC5, $F10AB737,
      $781293D3, $8410C2D2, $1C6587DB, $7D8F8F0F, $F3DCC4EA, $B965F99F,
      $9A0094D1, $65976D1C, $09173DC1, $8E38B992, $84701D44, $14F0E1B9,
      $E8A1EC5F, $1E925A12, $E77A0B5B, $CDB5926E, $D16260C8, $C917E806,
      $519076AA, $7BF6C21C, $808C0C90, $3E93C7B8, $707D6EA0, $F1DB698D ) } ;

    { theSeed = $1C3a128F; in case of the above initialisation }
  end;


  { ===== Fast version of the Marsaglia generator =====

    It takes the two 32bit values that are used to calculate
    the double directly from the SWBArray.

    The resulting stream of random numbers has slightly less
    statistical quality (the period is not that long, but it
    is still extremely long as compared to other generators)
    than the "normal" version (below). The randomness properties
    should satisfy all but the most demanding users.

    This class generates numbers slightly faster than the
    "normal" version; benchmarks timed with Delphi 3
    indicate that:
      It runs about 30% faster than TeoRandomMarsaglia
      It runs approximately as fast as TeoRandomGood
      It runs about 15% faster than TeoRandomBetter
      It runs about 55% (!) faster than TeoRandomBest
  }
  TeoRandomMarsagliaFast = class(TeoRandomMarsagliaBase)
  public
    constructor Clone(const aTeoRandomMarsagliaFast: TeoRandomMarsagliaFast);
    procedure Copy(const aTeoRandomMarsagliaFast: TeoRandomMarsagliaFast);

    function Value: double; override;
  private
    procedure SWBfill;
    function i32bit: LongInt;
    function i31bit: LongInt;
  end;



  { ===== Top-quality version of the Marsaglia RNG. =====
    Speed is sacrificed to get the best statistical
    characteristics possible for the resulting stream of random
    numbers. Nevertheless this generator is VERY fast:

    Benchmarks timed with Delphi 3 indicate that:
      It runs at about 70% of the speed of TeoRandomMarsaglia
      It runs at about 70% of the speed of TeoRandomGood
      It runs about 15% slower than TeoRandomBetter
      It runs about 33% (!) FASTER than TeoRandomBest


    This PRNG can be characterised as follows:

    - Extremely long period (more than 10^356---that's more than
      10^270 numbers for each atom in the universe, in case you
      want to simulate creation).

    - Combines two different types of generators, to achieve a very
      thorough mixing.

    - All possible values of 37 consecutive numbers occur with the
      correct frequencies in the long run.

    - Single precision reals (by far the most common) are guaranteed
      to have full precision in the fraction (mantissa).
      Almost all generators  produce reals by dividing a 32 (or 31) bit
      integer by 2^32 (2^31).  This means the smallest possible
      random reals are small multiples of 2^(-32).  This implementation
      will produce random reals down to 2^(-64) with the proper
      frequencies.  This makes it impossible to get a 0, which
      avoids the rare, but irritating program-stopping situation
      that arises from taking a logarithm of, or dividing by, zero.


    The principal component of this generator is the Subtract-with-Borrow (SWB)
    generator that is described in the Zaman/Marsaglia paper
    `A New Class of Random Number Generators', Annals of Applied
    Probability V1 462-480, 1991.

    They use a 148 byte seed array, to obtain an astronomically large
    period, while satisfying all the usual theoretical and experimental
    tests for randomness.

    The other component of this generator is the congruential generator with
    multiplier 69069 and base 2^32. This is a very well known, reliable
    (but short period) generator, tried and tested. It is, for example,
    the generator built into VAX's.

    The results of both of these generators are xor'ed to provide the
    bytes which form the output of this object. }

  TeoRandomMarsaglia = class(TeoRandomMarsagliaBase)
  public
    constructor Clone(const aTeoRandomMarsaglia: TeoRandomMarsaglia);
    constructor Get(S: TStream);

    procedure Copy(const aTeoRandomMarsaglia: TeoRandomMarsaglia);
    procedure Put(S: TStream);

    function Value: double; override;
  private
    RandomArray: array[0..37-1] of longint;

    procedure SWBfill;
    function i32bit: LongInt;
    function i31bit: LongInt;
  end;


{]====================================================[}

                     IMPLEMENTATION

uses
  SysUtils;

constructor TeoRandomBase.Create(const seed: integerType);
begin
  inherited Create;
  Reset(seed);
end;  {Create}

constructor TeoRandomBase.Clone(const aTeoRandomBase: TeoRandomBase);
begin
  inherited Create;
  theSeed := aTeoRandomBase.theSeed;
  theValue := aTeoRandomBase.theValue;
end;  {Clone}

constructor TeoRandomBase.Get(S: TStream);
begin
  inherited Create;
  S.ReadBuffer(theSeed,SizeOf(theSeed));
  S.ReadBuffer(theValue,SizeOf(theValue));
end;  {Get}

destructor TeoRandomBase.Destroy;  {override}
begin
  inherited Destroy;
end;  {Destroy}

procedure TeoRandomBase.Copy(const aTeoRandomBase: TeoRandomBase);
begin
  if Self = aTeoRandomBase then Exit;  {don't clobber self}
  theSeed := aTeoRandomBase.theSeed;
  theValue := aTeoRandomBase.theValue;
end;  {Copy}

procedure TeoRandomBase.Put(S: TStream);
begin
  S.WriteBuffer(theSeed,SizeOf(theSeed));
  S.WriteBuffer(theValue,SizeOf(theValue));
end;  {Put}

function TeoRandomBase.Seed: integerType;
begin
  Result := theSeed;
end; {Seed}

procedure TeoRandomBase.SetTheSeed(const seed: integerType);
begin
  theSeed := seed;
  theValue := seed;
end;

procedure TeoRandomBase.SetTheValue(const aValue: integerType);
begin
  theValue := aValue;
end;

function TeoRandomBase.TheValueIs: integerType;
begin
  Result := theValue;
end;



{--- TeoRandomGood ---}

constructor TeoRandomGood.Create(const seed: integerType);
begin
  inherited Create(seed);
end;  {Create}

destructor TeoRandomGood.Destroy;  {override}
begin
  inherited Destroy;
end;  {Destroy}

procedure TeoRandomGood.Reset(const seed: integerType);  {override}
const
  MASK = 123459876;
begin
  theSeed := seed xor MASK;
  if theSeed = 0 then theSeed := 1;
  theValue := theSeed;
end;  {Reset}

function TeoRandomGood.Value: double;  {override}
const
  a = 16807;
  q = MAXINTEGERTYPE div a;
  r = MAXINTEGERTYPE mod a;
  Minv = 1.0 / MAXINTEGERTYPE;
var
  k: integerType;
begin
  k := theValue div q;
  theValue := a * (theValue - k * q) - r * k; {// Schrage's technique}
  if theValue < 0 then Inc(theValue,MAXINTEGERTYPE);
  Result := theValue * Minv;
end;  {Value}



{--- TeoRandomBetter ---}

constructor TeoRandomBetter.Create(const seed: integerType);
begin
  inherited Create(seed);
end;  {Create}

constructor TeoRandomBetter.Clone(const aTeoRandomBetter: TeoRandomBetter);
begin
  inherited Clone(aTeoRandomBetter);
  auxValue := aTeoRandomBetter.auxValue;
  Table := aTeoRandomBetter.Table;
end;  {Clone}

constructor TeoRandomBetter.Get(S: TStream);
begin
  inherited Get(S);
  S.ReadBuffer(auxValue,SizeOf(auxValue));
  S.ReadBuffer(Table,SizeOf(Table));
end;  {Get}

destructor TeoRandomBetter.Destroy;  {override}
begin
  inherited Destroy;
end;  {Destroy}

procedure TeoRandomBetter.Copy(const aTeoRandomBetter: TeoRandomBetter);
begin
  if Self = aTeoRandomBetter then Exit;  {don't clobber self}
  inherited Copy(aTeoRandomBetter);
  auxValue := aTeoRandomBetter.auxValue;
  Table := aTeoRandomBetter.Table;
end;  {Copy}

procedure TeoRandomBetter.Put(S: TStream);
begin
  inherited Put(S);
  S.WriteBuffer(auxValue,SizeOf(auxValue));
  S.WriteBuffer(Table,SizeOf(Table));
end;  {Put}

function TeoRandomBetter.Calc: integerType;
const
  a = 16807;
  q = MAXINTEGERTYPE div a;
  r = MAXINTEGERTYPE mod a;
var
  k: integerType;
begin
  k := theValue div q;
  theValue := a * (theValue - k * q) - r * k; {// Schrage's technique}
  if theValue < 0 then Inc(theValue,MAXINTEGERTYPE);
  Result := theValue;
end;  {Calc}

procedure TeoRandomBetter.Reset(const seed: integerType);  {override}
const
  MASK = 123459876;
var
  i: integerType;
begin
  theSeed := seed xor MASK;
  if theSeed = 0 then theSeed := 1;
  theValue := theSeed;
    {//  fill the Table}
  for i := 1 to 8 do begin  {//  but 1st, a few warmups}
    Calc;
  end;
    {//  ok, NOW fill the table}
  for i := TeoRandomBetterTableSize downto 1 do begin
    Table[i] := Calc;
  end;
  auxValue := Table[1];
end;  {Reset}

function TeoRandomBetter.Value: double;  {override}
const
  NDIV = 1 + (MAXINTEGERTYPE - 1) div TeoRandomBetterTableSize;
  MaxR = 1.0 - 1.2e-7;  {//  maximum value to return}
  Minv = 1.0 / MAXINTEGERTYPE;
var
  j: integerType;
begin
  theValue := Calc;               {//  calc a new value}
  j := 1 + (auxValue div NDIV);   {//  calc an index. j is in [1..32]}
  auxValue := Table[j];           {//  return value at index}
  Table[j] := theValue;           {//  save new value}
  Result := auxValue * Minv;      {//  float it}
  if Result > MaxR then Result := MaxR;
end;  {Value}

function TeoRandomBetter.AuxValueIs: integerType;
begin
  Result := auxValue;
end; {AuxValueIs}

function TeoRandomBetter.AuxValueSet(const aInteger: integerType): integerType;
begin
  Result := auxValue;
  auxValue := aInteger;
end; {AuxValueSet}



{--- TeoRandomBest ---}

constructor TeoRandomBest.Create(const seed: integerType);
begin
  inherited Create(seed);
end;  {Create}

constructor TeoRandomBest.Clone(const aTeoRandomBest: TeoRandomBest);
begin
  inherited Clone(aTeoRandomBest);
  yaValue := aTeoRandomBest.yaValue;
end;  {Clone}

constructor TeoRandomBest.Get(S: TStream);
begin
  inherited Get(S);
  S.ReadBuffer(yaValue,SizeOf(yaValue));
end;  {Get}

destructor TeoRandomBest.Destroy;  {override}
begin
  inherited Destroy;
end;  {Destroy}

procedure TeoRandomBest.Copy(const aTeoRandomBest: TeoRandomBest);
begin
  if Self = aTeoRandomBest then Exit;  {don't clobber self}
  inherited Copy(aTeoRandomBest);
  yaValue := aTeoRandomBest.yaValue;
end;  {Copy}

procedure TeoRandomBest.Put(S: TStream);
begin
  inherited Put(S);
  S.WriteBuffer(yaValue,SizeOf(yaValue));
end;  {Put}

function TeoRandomBest.Calc1: integerType;
const
  M = 2147483563;
  a = 40014;
  q = 53668;
  r = 12211;
var
  k: integerType;
begin
  k := theValue div q;
  theValue := a * (theValue - k * q) - r * k; {// Schrage's technique}
  if theValue < 0 then Inc(theValue,M);
  Result := theValue;
end;  {Calc1}

function TeoRandomBest.Calc2: integerType;
const
  M = 2147483399;
  a = 40692;
  q = 52774;
  r = 3791;
var
  k: integerType;
begin
  k := yaValue div q;
  yaValue := a * (yaValue - k * q) - r * k; {// Schrage's technique}
  if yaValue < 0 then Inc(yaValue,M);
  Result := yaValue;
end;  {Calc2}

procedure TeoRandomBest.Reset(const seed: integerType);  {override}
const
  MASK = 123459876;
var
  i: integerType;
begin
  theSeed := seed xor MASK;
  if theSeed = 0 then theSeed := 1;
  theValue := theSeed;
  yaValue := theValue;
    {//  fill the Table}
  for i := 1 to 8 do begin  {//  but 1st, a few warmups}
    Calc1;
  end;
    {//  ok, NOW fill the table}
  for i := TeoRandomBetterTableSize downto 1 do begin
    Table[i] := Calc1;
  end;
  auxValue := Table[1];
end;  {Reset}

function TeoRandomBest.Value: double;  {override}
const
  M = 2147483563; {//  as in Calc1}
  NDIV = 1 + (M - 1) div TeoRandomBetterTableSize;
  MaxR = 1.0 - 1.2e-7;  {//  maximum value to return}
  Minv = 1.0 / M;
var
  j: integerType;
begin
  theValue := Calc1;              {//  calc new values}
  yaValue  := Calc2;
  j := 1 + (auxValue div NDIV);   {//  calc an index. j is in [1..32]}
  auxValue := Table[j] - yaValue; {//  combine series}
  Table[j] := theValue;           {//  save new value}
  if auxValue < 1 then Inc(auxvalue,M); {//  ensure subscript remains in range}
  Result := auxValue * Minv;      {//  float it}
  if Result > MaxR then Result := MaxR;
end;  {Value}

function TeoRandomBest.YaValueIs: integerType;
begin
  Result := yaValue;
end; {YaValueIs}

function TeoRandomBest.YaValueSet(const aInteger: integerType): integerType;
begin
  Result := yaValue;
  yaValue := aInteger;
end; {YaValueSet}



{--- TeoRandomQAD ---}

constructor TeoRandomQAD.Create(const seed: integertype);
begin
  inherited Create(seed);
end;  {Create}

destructor TeoRandomQAD.Destroy;  {override}
begin
  inherited Destroy;
end;  {Destroy}

procedure TeoRandomQAD.Reset(const seed: integerType);  {override}
begin
  SetTheSeed(seed);
end;  {Reset}

function TeoRandomQAD.Value: double;  {override}
const
  a = 1664525;    {D. Knuth}
  c = 1013904223; {H. W. Lewis}
  expomask = $3f800000; {produces a single 'x' such that 1 <= x <= 2}
  mantmask = $007FFFFF; {23 low order bits}
type
{$ifdef WIN32}
  overlay = record
    case integer of
    0: (i: integer);
    1: (s: single);
  end;
{$else  WINDOWS 3.x}
  overlay = record
    case integer of
    0: (i: longint);
    1: (s: single);
  end;
{$endif WIN32}
var
  trick: overlay;
begin
  theValue := a * TheValueIs + c;
  {Using masks, morph theValue into a single, return as double}
  trick.i := expomask or (theValue and mantmask);
  Result := trick.s - 1.0;
end;  {Value}



{--- TeoRandomMarsagliaBase ---}

constructor TeoRandomMarsagliaBase.Create(const Seed: integerType);
begin
  inherited Create(Seed);
  Reset(Seed);
end;

constructor TeoRandomMarsagliaBase.CreateSWB(const Seed, SWBseed: integerType);
begin
  inherited Create(Seed);
  ResetSWB(Seed, SWBseed);
end;

constructor TeoRandomMarsagliaBase.Clone(const aTeoRandomMarsagliaBase: TeoRandomMarsagliaBase);
begin
  inherited Clone(aTeoRandomMarsagliaBase);
  Copy(aTeoRandomMarsagliaBase);
end;


procedure TeoRandomMarsagliaBase.Copy(const aTeoRandomMarsagliaBase: TeoRandomMarsagliaBase);
begin
  inherited Copy(aTeoRandomMarsagliaBase);

  theCarryFlag := aTeoRandomMarsagliaBase.theCarryFlag;
  randomElement:= aTeoRandomMarsagliaBase.randomElement;
  theSWBseed   := aTeoRandomMarsagliaBase.theSWBseed;

  SWBArray     := aTeoRandomMarsagliaBase.SWBArray;
end;

{ streaming code for all generators }
procedure TeoRandomMarsagliaBase.Put(S: TStream);
begin
  inherited Put(S);
  { base class:
  S.WriteBuffer(theSeed,  SizeOf(theSeed));
  S.WriteBuffer(theValue, SizeOf(theValue)); }

  S.WriteBuffer(theCarryFlag  ,SizeOf(theCarryFlag));
  S.WriteBuffer(randomElement ,SizeOf(randomElement));
  S.WriteBuffer(theSWBseed    ,SizeOf(theSWBseed));

  S.WriteBuffer(SWBArray      ,SizeOf(SWBArray));
end;  {Put}

constructor TeoRandomMarsagliaBase.Get(S: TStream);
begin
  inherited Get(S);
  { base class:
   S.ReadBuffer(theSeed , SizeOf(theSeed));
   S.ReadBuffer(theValue, SizeOf(theValue)); }

  S.ReadBuffer(theCarryFlag,  SizeOf(theCarryFlag));
  S.ReadBuffer(randomElement, SizeOf(randomElement));
  S.ReadBuffer(theSWBseed,    SizeOf(theSWBseed));

  S.ReadBuffer(SWBArray,      SizeOf(SWBArray));
end;

procedure TeoRandomMarsagliaBase.Reset(const Seed: integerType);
begin
  { inherited Reset(Seed)  --  virtual; abstract; method }

  { initialize with a default second seed }
  ResetSWB(Seed, 7654321);
end;

procedure TeoRandomMarsagliaBase.ResetSWB(const Seed, SWBseed: integerType);
var
  i, j: integer;
  tidbits: LongInt;
begin
  { default: Seed=1234567, SWBSeed=7654321. Really.}

  tidbits := 0;

  { initialize data members: seed }
  theSeed := Seed*2+1;
  theSWBseed := SWBseed;

  { fill data member: table }
  for i :=0 to 37-1 do
  begin
    for j := 32 downto 1 do
    begin
      theSeed := theSeed * 69069;
      theSWBseed := theSWBseed xor (theSWBseed shr 15);
      theSWBseed := theSWBseed xor (theSWBseed shl 17);
      tidbits := (tidbits shr 1) or ($80000000 and (theSeed xor theSWBseed));
    end;

    SWBArray[i] := tidbits;
  end;

  { initialize data member: carry flag }
  theCarryFlag := 0;

  { initialize data member: current random element
      make sure that this points behind the last element of the table,
      so that the table is refilled the moment it is accessed }
  randomElement := 37 + 1;
end;


function TeoRandomMarsagliaBase.SWBSeed: integerType;
begin
  Result := theSWBSeed;
end;


{--- TeoRandomMarsagliaFast ---}

constructor TeoRandomMarsagliaFast.Clone(const aTeoRandomMarsagliaFast: TeoRandomMarsagliaFast);
begin
  inherited Clone(aTeoRandomMarsagliaFast);
  Copy(aTeoRandomMarsagliaFast);
end;

procedure TeoRandomMarsagliaFast.Copy(const aTeoRandomMarsagliaFast: TeoRandomMarsagliaFast);
begin
  inherited Copy(aTeoRandomMarsagliaFast);
  { nothing else }
end;

procedure TeoRandomMarsagliaFast.SWBfill;

{
  This is the SWB operation on a very long integer;
  it has been left in for clearer representation,
  but has been "inlined" into the loop for improved
  speed.

  procedure SWB(var c: byte; var x,y,z: longint);
  begin
    if y < 0 then
    begin
      z := x - y - c;
      c := ord((z < 0) or (x >= 0));  // OR
    end
    else
    begin
      z := x - y - c;
      c := ord((z < 0) and (x >= 0)); // AND
    end;
  end;
}


{$IFNDEF Win32} { use this for 16bit only }
var
  i : integer;
begin
{
  Here we see the stylish implementation of the SWB
  operation; as mentioned above it is NOT used, in order
  to minimize call overhead. If you rearrange the
  appropriate comments this code will work flawlessly.

  for i := 0 to 24-1 do
    SWB(theCarryFlag, SWBArray[i + 37 - 24],
                     SWBArray[i],
                     SWBArray[i]);

  for i := 24 to 37-1 do
    SWB(theCarryFlag, SWBArray[i  - 24],
                     SWBArray[i],
                     SWBArray[i]);
}


{ Version with "inlined" SWB - faster, with no call overhead }
  for i := 0 to 24-1 do { 0 to 23 = 24 loops }
    if SWBArray[i] < 0 then
    begin
      SWBArray[i] := SWBArray[i + 37 - 24] - SWBArray[i] - theCarryFlag;
      theCarryFlag := ord((SWBArray[i] < 0) or (SWBArray[i + 37 - 24] >= 0));
    end
    else
    begin
      SWBArray[i] := SWBArray[i + 37 - 24] - SWBArray[i] - theCarryFlag;
      theCarryFlag := ord((SWBArray[i] < 0) and (SWBArray[i + 37 - 24] >= 0));
    end;

  for i := 24 to 37-1 do { 24 to 37-1 = 13 loops }
    if SWBArray[i] < 0 then
    begin
      SWBArray[i] := SWBArray[i  - 24] - SWBArray[i] - theCarryFlag;
      theCarryFlag := ord((SWBArray[i] < 0) or (SWBArray[i  - 24] >= 0));
    end
    else
    begin
      SWBArray[i] := SWBArray[i  - 24] - SWBArray[i] - theCarryFlag;
      theCarryFlag := ord((SWBArray[i] < 0) and (SWBArray[i  - 24] >= 0));
    end;
end; { 16bit implementation }

{$ELSE}
begin{ 32 bit implementation }
  asm
      { must preserve these registers }
      push edi
      push esi

      { Self is always passed in eax for methods }
      lea edi, [eax].SWBArray
      lea esi, [eax].SWBArray + 13*4

      { set carry flag to what it was the last time we
        left this routine }
      { destroy eax }
      mov  ah, [eax].theCarryFlag.byte
      sahf

      mov ecx, 24             {    will do first loop 24 times       }
    @1:
      { On a 80386, the instructions `lodsd', `stosd' and `loop'
        all change registers automatically as noted in parentheses   }
      lodsd                   {   eax = x[i+13]     ( esi = esi+4 )  }
      sbb eax, [edi]
      stosd                   {    y[i] = eax       ( edi = edi+4 )  }
(* SLOW VERSION
      loop @1                 {                     ( ecx = ecx-1 )  }
*)
      dec ecx { this is faster than LOOP }
      jnz @1

      mov  esi, Self
      lea  esi, [esi].SWBArray
      { edi is set up already }

      mov ecx,13              {    will do next loop 13 times       }
    @2:
      lodsd                   {   eax = x[i-24]     ( esi = esi+4 ) }
      sbb eax, [edi]
      stosd                   {    x[i] = eax       ( edi = edi+4 ) }
(* SLOW VERSION
      loop @2                 {                     ( ecx = ecx-1 )  }
*)
      dec ecx { this is faster than LOOP }
      jnz @2

      {  save the carry flag for next time }
      lahf
      mov esi, Self
      mov [esi].theCarryFlag.byte, ah

      { restore values we pushed }
      pop esi
      pop edi
  end;
end;
{$ENDIF Win32}

function TeoRandomMarsagliaFast.i32bit: LongInt;
begin
  if randomElement < 37-1 then
  begin
    inc(randomElement);
    Result := SWBArray[randomElement];
  end
  else
  begin
    randomElement := 0;
    SWBfill;
    Result := SWBArray[0];
  end;
end;

function TeoRandomMarsagliaFast.i31bit: LongInt;
begin
  if randomElement < 37-1 then
  begin
    inc(randomElement);
    Result := SWBArray[randomElement];
  end
  else
  begin
    randomElement := 0;
    SWBfill;
    Result := SWBArray[0];
  end;

  Result := Result and $7FFFFFFF;
end;


function TeoRandomMarsagliaFast.Value: double;
const
  two2neg31 = ( (2.0/$10000) / $10000 ) ;
  two2neg32 = ( (1.0/$10000) / $10000 ) ;
{ This is slower, but looks less messed up:
var
  temp: LongInt;
begin
  temp := i32bit;
  Result := ( temp * two2neg32 + i31bit) * two2neg31;
end; }
begin
  Result := ( i32bit * two2neg32 + i31bit) * two2neg31;
end;



{--- TeoRandomMarsaglia ---}

constructor TeoRandomMarsaglia.Clone(const aTeoRandomMarsaglia: TeoRandomMarsaglia);
begin
  inherited Clone(aTeoRandomMarsaglia);
  Copy(aTeoRandomMarsaglia);
end;

procedure TeoRandomMarsaglia.Copy(const aTeoRandomMarsaglia: TeoRandomMarsaglia);
begin
  inherited Copy(aTeoRandomMarsaglia);

  RandomArray := aTeoRandomMarsaglia.RandomArray;
end;

procedure TeoRandomMarsaglia.Put(S: TStream);
begin
  inherited Put(S);

  S.WriteBuffer(RandomArray,SizeOf(RandomArray));
end;

constructor TeoRandomMarsaglia.Get(S: TStream);
begin
  inherited Get(S);

  S.ReadBuffer(RandomArray, SizeOf(RandomArray));
end;

procedure TeoRandomMarsaglia.SWBfill;
{
  See the comments in TeoRandomMarsagliaFast.SWBfill

  procedure SWB(var c: byte; var x,y,z: longint);
  begin
    if y < 0 then
    begin
      z := x - y - c;
      c := ord((z < 0) or (x >= 0));
    end
    else
    begin
      z := x - y - c;
      c := ord((z < 0) and (x >= 0));
    end;
  end;
}

{$IFNDEF Win32}
var
  i : integer;
begin
{
  for i := 0 to 24-1 do
    SWB(theCarryFlag, SWBArray[i + 37 - 24],
                     SWBArray[i],
                     SWBArray[i]);

  for i := 24 to 37-1 do
    SWB(theCarryFlag, SWBArray[i  - 24],
                     SWBArray[i],
                     SWBArray[i]);
}

  for i := 0 to 24-1 do
    if SWBArray[i] < 0 then
    begin
      SWBArray[i] := SWBArray[i + 37 - 24] - SWBArray[i] - theCarryFlag;
      theCarryFlag := ord((SWBArray[i] < 0) or (SWBArray[i + 37 - 24] >= 0));
    end
    else
    begin
      SWBArray[i] := SWBArray[i + 37 - 24] - SWBArray[i] - theCarryFlag;
      theCarryFlag := ord((SWBArray[i] < 0) and (SWBArray[i + 37 - 24] >= 0));
    end;

  for i := 24 to 37-1 do
    if SWBArray[i] < 0 then
    begin
      SWBArray[i] := SWBArray[i  - 24] - SWBArray[i] - theCarryFlag;
      theCarryFlag := ord((SWBArray[i] < 0) or (SWBArray[i  - 24] >= 0));
    end
    else
    begin
      SWBArray[i] := SWBArray[i  - 24] - SWBArray[i] - theCarryFlag;
      theCarryFlag := ord((SWBArray[i] < 0) and (SWBArray[i  - 24] >= 0));
    end;

  { This is the additional feature for the "normal" PRNG
    each element is XORed with a multiplicative congruential
    generator }

  { apply congruential generator }
  for i := 0 to 37-1 do
  begin
    theSeed := theSeed * 69069;
    RandomArray[i] := SWBArray[i] xor theSeed;
  end;

end; { 16bit version }
{$ELSE}
begin { 32bit version }
  asm

      { See the comments in TeoRandomMarsagliaFast.SWBfill }

      { must preserve these registers }
      push edi
      push esi

      { Self is always passed in eax for methods }
      lea edi, [eax].SWBArray
      lea esi, [eax].SWBArray + 13*4

      { destroy eax }
      mov  ah, [eax].theCarryFlag.byte
      sahf

      mov ecx, 24
    @1:
      lodsd
      sbb eax, [edi]
      stosd
   (* loop @1 { this is slow } *)
      dec ecx { this is faster than LOOP }
      jnz @1

      mov  esi, Self
      lea  esi, [esi].SWBArray
      { edi is set up already }

      mov ecx,13
    @2:
      lodsd
      sbb eax, [edi]
      stosd
   (* loop @2 { this is slow } *)
      dec ecx { this is faster than LOOP }
      jnz @2

      lahf
      mov esi, Self
      mov [esi].theCarryFlag.byte, ah


      { XOR the elements of SWBArray with a congruential generator and put the
        result in RandomArray, reset the counter and the pointer. }
      mov  eax, [esi].theSeed     { initialize congruential generator }

      lea  edi, [esi].RandomArray { destination array }
      lea  esi, [esi].SWBArray    { source array }

      mov  ecx,37

      push ebx                    { we must not modify ebx }
      mov  ebx,69069

    @c:
      mul  ebx
      mov  edx,eax
      lodsd                       { eax   := [esi];        esi := esi + 4 }
      xor  eax,edx                { eax   := eax XOR edx                  }
      stosd                       { [edi] := eax;          edi := edi + 4 }
      mov  eax,edx
   (* loop @c { this is slow } *)
      dec ecx { this is faster than LOOP }
      jnz @c

      pop  ebx                    { restore ebx }

      mov  esi, Self
      mov  [esi].theSeed.DWORD, eax
      { update congruential generator with the new, current seed }

      { restore values we pushed }
      pop esi
      pop edi
  end;
end;
{$ENDIF Win32}

function TeoRandomMarsaglia.i32bit: LongInt;
begin
  if randomElement < 37-1 then
  begin
    inc(randomElement);
    Result := RandomArray[randomElement];
  end
  else
  begin
    randomElement := 0;
    SWBfill;
    Result := RandomArray[0];
  end;
end;

function TeoRandomMarsaglia.i31bit: LongInt;
begin
  if randomElement < 37-1 then
  begin
    inc(randomElement);
    Result := RandomArray[randomElement];
  end
  else
  begin
    randomElement := 0;
    SWBfill;
    Result := RandomArray[0];
  end;

  Result := Result and $7FFFFFFF;
end;

function TeoRandomMarsaglia.Value: double;
const
  two2neg31 = ( (2.0/$10000) / $10000 ) ;
  two2neg32 = ( (1.0/$10000) / $10000 ) ;
begin
  Result := ( i32bit * two2neg32 + i31bit) * two2neg31;
end;

{initialization}
{finalization}
end. {Random.pas}
