unit s_mcmc;
(*Entire unit by MS.
Markov chain Monte Carlo integration methods and related procedures applicable both to
- maximum likelihood
- Bayesian
problems involving intractable integrals.
The emphasis is on general tools that can be adapted to solving many problems.
Specific applications can be found in particular in <S_mcmc_I> <S_mcmc_II>, <S_re>, and others.*)

(*--------------------------------------------------------------------------------------------------*)

interface

(*--------------------------------------------------------------------------------------------------*)

uses sysutils, classes, eight, s_constants, slib, s_dat, s_base, s_matrices, Math, amrandomx;

type MCMC = object(ModelSpecification)

procedure CheckNetwork(var ascii: textfile);
(*Writes present network to file.*)

end; (*MCMC*)

const tin = 1e-16; (*Serves as threshold.*)

var
  stage: (*Stage of the information process: needed to communicate between distinct Delphi / Pascal objects.*)
    byte;
  gibbs1, (*Gibbs sampler: rate parameter.*)
  gibbs2, (*Gibbs sampler: (inverse) variance-covariance matrix of actor-dependent random effects.*)
  newpath, (*Path candidate accepted or not.*)
  mproposal, (*Multi-block or one-block Metropolis-Hastings.*)
  success1, (*<False>: rescale candidate-generating density corresponding to parameter vector.*)
  success2, (*<False>: rescale candidate-generating density corresponding to actor-dependent random effects.*)
  indscale, (*Individual scale factors for rescaling the variance-covariance matrices of candidate-generating densities or not.*)
  mlecovre: (*Stochastic approximation (Fisher scoring) step or not: precision matrix of the actor-dependent random effects or not.*)
    boolean;
  reflectiontheta: (*Metropolis-Hastings, candidate-generating density: point in Euclidean n-space about which vector-valued candidates are reflected.*)
    pvec;

(*-------------------------------------*)
(*Markov chain Monte Carlo: information*)
(*-------------------------------------*)

procedure MCMC_intro(var MaxLike, Bayesian: boolean; var actors: integer; var scaleRE_i: R_Ar; var ascii: TFileStream);
(*Write basic description of the Markov chain Monte Carlo
sampling scheme to the output file.*)

procedure MCMC_report(var stage: byte; var maxlike: boolean; var dim1: byte; var theta: pvec;
var randomeffects: boolean; var dim2: byte; var sigma: pmatrix; var actors: integer; var re: rrarray;
var Bayesian: boolean; var blackbox: blackarray; var thetafile, covrefile, refile, slrcovrefile, arrefile, adllfile,
arpathfile, arcovrefile, arthetafile: textfile);
(*Report information on how the Markov chain behaves to files.*)

procedure MCMC_final_report(var Bayesian: boolean; var ascii: TFileStream; var randomeffects: boolean);
(*Autocorrelations of functionals whose expectation is to be approximated: basic information.*)

(*--------------------------------------*)
(*Monte Carlo s.e. / sampling efficiency*)
(*--------------------------------------*)

procedure MC_Error_I(var dim: byte; var vector1: pvec; var ldim1: integer; var matrix1: pmatrix; var matrix3: rrarray; var ldim2: integer; var ascii: TFileStream; var input: shortstring);
(*Evaluate Monte Carlo batch means estimator of the Monte Carlo s.e. introduced by
estimating the desired integrals by ergodic averages.*)

procedure MC_Error_II(var dim: byte; var matrix2: pmatrix; var ldim1: integer; var mc_covre: rrrarray; var matrix1: pmatrix; var ldim2: integer; var ascii: TFileStream; var input: shortstring);
(*Evaluate Monte Carlo batch means estimator of the Monte Carlo s.e. introduced by
estimating the desired integrals by ergodic averages.*)

procedure MC_Error_III(var eb: rrarray; var actors: integer; var dim: byte; var ldim1: integer; var mc_re: rrrarray; var matrix3: rrarray;
var ldim2: integer; var ascii: TFileStream; var input: shortstring);
(*Evaluate Monte Carlo batch means estimator of the Monte Carlo s.e. introduced by
estimating the desired integrals by ergodic averages.*)

procedure Autocorrelation(var autocov1: R_Ar; var dim: byte; var ldim: integer; var autovector: rrarray; var lag: byte; var file1: textfile); overload;
procedure Autocorrelation(var autocov1: R_Ar; var dim: byte; var ldim: integer; var autovector: rrarray; var lag: byte; var file1, file2: textfile); overload;
procedure Autocorrelation(var autocov2: rrarray; var dim: byte; var actors, ldim: integer; var automatrix: rrrarray; var lag: byte; var file1: textfile); overload;

procedure Autocorrelation(var autocov1: R_Ar; var dim: byte; var ldim: integer; var autovector: rrarray; var lag: byte; var file1: TFileStream); overload;
procedure Autocorrelation(var autocov1: R_Ar; var dim: byte; var ldim: integer; var autovector: rrarray; var lag: byte; var file1:  TFileStream; var file2: textfile); overload;
procedure Autocorrelation(var autocov2: rrarray; var dim: byte; var actors, ldim: integer; var automatrix: rrrarray; var lag: byte; var file1: TFileStream); overload;
(*Evaluate the lag <lag> sample auto-covariance/correlation function. If the acceptance rate of the Metropolis-Hastings algorithm is pushed towards
some specified value, then a fair comparison between autocorrelation and iid sampling would be based on every 1/optimal rate-th observation.*)

function ISE(const autocov: array of real; const dim: byte): real;
(*Evaluate Initial Sequence Estimates of the Monte Carlo variance,
in particular the Initial Monotone Sequence Estimate;
see Geyer (1992) and the Gilks et al. (1996), p. 50, which contains some minor errors.*)

procedure IEFactor(var ascii: TFileStream; var ldim: integer; var dim: byte; var autovector: rrarray; var ISE_var: R_Ar); overload;
procedure IEFactor(var ascii: TFileStream; const indicator: byte; var iterations: integer; var dimension1: integer; var dimension2: byte; var automatrix: rrrarray; var ISE_var: rrarray); overload;
(*Evaluate
- lag <t> autocorrelations, <t = 0, 1, 2, ..., m>.
- Monte Carlo standard errors as estimated by Initial Sequence Estimates (Geyer, 1992)
- inefficiency factors (Chib, 2004)
which is all written to the file <ascii>.
*)

(*--------------------------------------------------*)
(*Sampling from probability density / mass functions*)
(*--------------------------------------------------*)
(*These are some routines that are frequently needed, complemented
by routines in <UltrasSim>, <AMXRAndom>, <RanGen>.*)

function RescaleCGD(const observed, desired, iterations, tolerance: integer; var number: byte; var success: boolean): real;
(*Rescale candidate-generating density to achieve desired acceptance rate:
method I in Browne's (1998) thesis <Applying MCMC methods to multi-level models>.*)

procedure SampleMVN(var mvndensities: R_Ar; const actors: integer; const dim: byte; const cholfac: pmatrix;
const scaleRE_i: R_Ar; var mvn: rrarray); overload;
(*Algorithm:
- sample <z ~ N(0, I)>, where <I> is the identity matrix.
- set <mvn = C z>, where <C> is the Cholesky factor of variance-covariance matrix <Sigma>,
  so that <mvn ~ N(0, Sigma)>.
Note that <C> and hence <Sigma> are rescaled.*)
function SampleMVN(const dim: byte; const cholfac: pmatrix; const scale: real; var vector1: pvec): real; overload;
(*Algorithm:
- sample <z ~ N(0, I)>, where <I> is the identity matrix.
- set <mvn = C z>, where <C> is the Cholesky factor of variance-covariance matrix <Sigma>,
  so that <mvn ~ N(0, Sigma)>.
Note that <C> and hence <Sigma> are rescaled.*)

function SampleGamma(var ldim: integer; var real1: real): real;
(*Generate one draw from the Gamma distribution with parameters <ldim> and <real1>: see Robert & Casella, 1999, p. 45.*)

procedure SampleW(var dim: byte; var ldim: integer; var cholfac, ssigma: pmatrix);
(*Generate one draw from the Wishart distributionl, parametrized such that $E(\Sigma^{-1}) = m \Omega$, where $m$ is the degree of freedoms parameter, and $\Omega$ the scale matrix.*)

procedure CandidateMH(var mvn0: rrarray; const actors: integer; const dim: byte; var mvndensities: R_Ar; const cholfac: pmatrix;
const scaleRE_i: R_Ar; const cdensity: byte; var mvn: rrarray; const re, intermediate: rrarray); overload;
(*Generates candidates for Metropolis-Hastings algorithms with multivariate normal candidate-generating densities: one or more vectors: one vector.*)

function CandidateMH(const dim: byte; const cholfac: pmatrix; const scale: real;
var vector1: pvec; const cdensity: byte; var vector10: pvec; const vector2, vector3: pvec): real; overload;
(*Generates candidates for Metropolis-Hastings algorithms with multivariate normal candidate-generating densities.*)

(*---------------------------------------------*)
(*Evaluate probability density / mass functions*)
(*---------------------------------------------*)
(*Comment:
functions to evaluate probability mass / density functions up to a constant of proportionality.*)

function EvaluatePoisson(var ldim1: integer; var real1: real; const real2: real; const ldim2: integer): real;
(*Evaluate Poisson kernel.*)

function EvaluateGamma(var dim: byte; var vector1: pvec; var gamma1: real; var vector2: pvec; var gamma2: real): real;
(*Evaluate Gamma kernel with parameters <gamma1> and <gamma2> for each coordinate of some vector valued random variate.*)

function EvaluateMVN(var mvndensities: R_Ar; const actors: integer; const dim: byte; const re: rrarray; const isigma: pmatrix): real; overload;
(*Evaluate multivariate Gaussian kernel: one or more vectors.*)

function EvaluateMVN(const dim: byte; const vector1: pvec; const isigma: pmatrix): real; overload;
(*Evaluate multivariate Gaussian kernel: one vector.*)

function EvaluateFunctionMVN1(var mvndensities: R_Ar; var ldim: integer; var dim: byte; var re: rrarray; var scaleRE_i: R_Ar): real;
(*Evaluate multivariate Gaussian kernel: one or more observations, variance-covariance matrix equal to identity matrix.*)

function EvaluateFunctionMVN2(var mvndensities: R_Ar; const actors: integer; const dim: byte; const re: rrarray; const isigma: pmatrix; const scaleRE_i: R_Ar): real; overload;
(*Evaluate multivariate Gaussian: one or more vectors, variance-covariance matrix rescaled.*)

function EvaluateFunctionMVN2(const dim: byte; const vector1: pvec; const isigma: pmatrix; const scale: real): real; overload;
(*Evaluate multivariate Gaussian: one vector, variance-covariance matrix rescaled.*)

function EvaluateMVNactor(const dim: byte; const re: rrarray; const ldim: integer; const isigma: pmatrix): real;
(*Evaluate multivariate Gaussian: one specified vector of the two or more vectors.*)

function EvaluateW(var omega: pmatrix; var degrees: integer; var dim: byte; var sigma: pmatrix): real;
(*Evaluate the Wishart kernel, which corresponds to the conjugate prior probability density of
the variance-covariance matrix of the random effects.*)

(*------------------------------*)
(*Some Metropolis-Hastings tools*)
(*------------------------------*)
(*There are many more in other units, such as <S_ml> and <S_re>.*)

function DecisionMH(var acceptancepr, real1, real2: real; var blackbox: blackarray): boolean;
(*Decide whether or not the candidate is to be accepted.*)

(*-----------------------*)
(*Output-related routines*)
(*-----------------------*)

procedure MCMC_Output_Files_I(const input, input1, input2, input3: shortstring);
procedure MCMC_Output_Files_II(const input: shortstring; const no_columns: integer; const input1, input2, input3: shortstring);
(*Monte Carlo output procedures.*)

procedure Get_MCMC_Output(const Bayesian: boolean; const no_columns: integer);
(*Obtain (and transform) MCMC output.*)

procedure MatlabPlot(var ascii: textfile);
(*Adds some convenient Matlab code that enables columnwise plots.*)

procedure MatlabPlotExtra(var ascii: textfile);
(*Some Matlab code.*)

procedure GenerateGnu(const input, input1, input2, input3: shortstring);
(*Generates Gnuplot source code to make simple trace plots and so on.*)

(*--------------------------------------------------------------------------------------------------*)

implementation

(*--------------------------------------------------------------------------------------------------*)

procedure MCMC.CheckNetwork(var ascii: textfile);
(*Writes current network to file.*)
var i, j: integer;
begin
for i := 1 to n do
  begin
  writeln(ascii);
  for j := 1 to n do
    begin
    if yn(i,j) then write(ascii,' 1')
    else write(ascii,' 0');
    end;
  end;
writeln(ascii); writeln(ascii);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MCMC_intro(var MaxLike, Bayesian: boolean; var actors: integer; var scaleRE_i: R_Ar; var ascii: TFileStream);
(*Write basic description of the Markov chain Monte Carlo
sampling scheme to the output file.*)
var i: integer;
x: string;
begin
if (MaxLike or Bayesian) then
  begin
  gibbs1 := true; (*Rate parameter: Gibbs samper versus Metropolis-Hastings.*)
  gibbs2 := true; (*(Inverse) variance-covariance matrix: Gibbs sampler versus Metropolis-Hastings.*)
  mproposal := true; (*Multi-block Metropolis-Hastings versus one-block Metropolis-Hastings: actor-dependent random effects.*)
  success1 := false; (*<False>: rescale candidate-generating density corresponding to parameter vector.*)
  success2 := false; (*<False>: rescale candidate-generating density corresponding to actor-dependent random effects.*)
  if (scalebeta <> 0) then
    begin
    success1 := true;
    blackbox[12] := scalebeta;
    end;
  if ((scalere <> 0) and randomeffects) then
    begin
    success2 := true;
    for i := 1 to actors do scaleRE_i[i] := scalere;
    blackbox[11] := scalere;
    end;
  indscale := false; (*Common scale factor for the random effects, false: individual scale factors for the random effects.*)
  mlecovre := true; (*Variance-covariance matrix simply estimated by stochastic approximation.*)
  Heading(2,ascii,'Markov chain Monte Carlo algorithm');
  x := 'Latent data:'+eol;
  ascii.Write(pchar(x)^,length(x));
  x := '- paths: Metropolis-Hastings.'+eol;
  ascii.Write(pchar(x)^,length(x));
  ascii.Write(pchar(eol)^,length(eol));// writeln(ascii);
  if randomeffects then
    begin
    x := '- actor-dependent random effects: Metropolis-Hastings: ';
    ascii.Write(pchar(x)^,length(x));
    if (cdensity2 = 1) then begin
      x := 'random walk: ';
      ascii.Write(pchar(x)^,length(x));
    end
    else if (cdensity2 = 2) then begin
        x := 'autoregressive density: ';
        ascii.Write(pchar(x)^,length(x));
    end
    else if (cdensity2 = 3) then begin
      x := 'independence sampler: ';
      ascii.Write(pchar(x)^,length(x));
    end;
    if mproposal then begin
      x := 'multi-block.';
      ascii.Write(pchar(x)^,length(x));
    end
    else begin
      x := 'one-block.';
      ascii.Write(pchar(x)^,length(x));
    end;
    if (scalere = 0) then begin
      x := 'Variance-covariance matrix of candidate-generating (multivariate) Gaussian density to be rescaled.'+eol;
      ascii.Write(pchar(x)^,length(x));
    end
    else begin
      x := format('%s%11.6f%s',['Variance-covariance matrix of candidate-generating (multivariate) Gaussian density to be rescaled by factor ',scalere,'.'])+eol;
     ascii.Write(pchar(x)^,length(x));
    end;
    ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
   end;
  if Bayesian then
    begin
    x := 'Parameters: '+eol;
    ascii.Write(pchar(x)^,length(x));
    ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
    x := '- rate parameters: ';
    ascii.Write(pchar(x)^,length(x));
    if gibbs1 then begin
        x := 'Gibbs.'+eol;
        ascii.Write(pchar(x)^,length(x));
    end
    else begin
      x := 'Metropolis-Hastings: independence sampler.'+eol;
      ascii.Write(pchar(x)^,length(x));
    end;
    ascii.Write(pchar(x)^,length(x));//writeln(ascii);
    x := '- objective function parameters: Metropolis-Hastings: ';
    ascii.Write(pchar(x)^,length(x));
    if (cdensity1 = 1) then begin
      x := 'random walk.'+eol;
      ascii.Write(pchar(x)^,length(x));
    end
    else if (cdensity1 = 2) then begin
      x := 'autoregressive density.'+eol;
      ascii.Write(pchar(x)^,length(x));
    end
    else if (cdensity1 = 3) then begin
      x := 'independence sampler.'+eol;
      ascii.Write(pchar(x)^,length(x));
    end;
    if (scalebeta = 0) then begin
      x := 'Variance-covariance matrix of candidate-generating (multivariate) Gaussian density to be rescaled.'+eol;
      ascii.Write(pchar(x)^,length(x));
    end
    else begin
      x := format('%s%11.6f%s',['Variance-covariance matrix of candidate-generating (multivariate) Gaussian density to be rescaled by factor ',scalebeta,'.'])+eol;
     ascii.Write(pchar(x)^,length(x));
    end;
    ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
    if randomeffects then
      begin
      x := '- (inverse) variance-covariance matrix: ';
      ascii.Write(pchar(x)^,length(x));
      if gibbs2 then begin
        x := 'Gibbs.'+eol;
        ascii.Write(pchar(x)^,length(x));
      end
      else begin
        x := 'Metropolis-Hastings: random walk.'+eol;
        ascii.Write(pchar(x)^,length(x));
      end;
      ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
      end;
    end
  else
    begin
    if randomeffects then
      begin
      if mlecovre then begin
        x := '(Inverse) variance-covariance matrix estimated by stochastic approximation.'+eol;
        ascii.Write(pchar(x)^,length(x));
      end
      else begin
        x := '(Inverse) variance-covariance estimated by solution to maximum likelihood equation.'+eol;
        ascii.Write(pchar(x)^,length(x));
      end;
      ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
      end;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MCMC_report(var stage: byte; var maxlike: boolean; var dim1: byte; var theta: pvec;
var randomeffects: boolean; var dim2: byte; var sigma: pmatrix; var actors: integer; var re: rrarray;
var Bayesian: boolean; var blackbox: blackarray; var thetafile, covrefile, refile, slrcovrefile, arrefile, adllfile,
arpathfile, arcovrefile, arthetafile: textfile);
(*Report information on how the Markov chain behaves to files.*)
var
  i, j: integer;
  minimum: real;
  condition: boolean;
  phase2, phase3: set of byte;
  eigen: pvec;
  v1, v2: pmatrix;
begin
condition := true;
if condition then exit;(*Skip the present procedure: do not write to files.*)
phase2 := [20..29];
phase3 := [30..39];
if (stage in phase2) then (*Phase 2.*)
  begin
  if maxlike then (*In the Bayesian algorithm, phase 2 serves as burn-in.*)
    begin
    VectorWrite(thetafile,dim1,theta); (*Writes estimates of fixed effects (parameter) to file to produce time series.*)
    if randomeffects then MatrixWriteCov(dim2,covrefile,sigma); (*To detect non-convergence: time series of variance-covariance matrix of random effects.*)
    end;
  end;
if (stage in phase3) then (*Phase 3.*)
  begin
  if Bayesian then VectorWrite(thetafile,dim1,theta); (*Writes estimates of fixed effects (parameter) to file to produce time series.*)
  if randomeffects then
    begin
    MatrixWritePlot(randomeffects,actors,dim2,refile,re); (*Writes last sampled actor-dependent random effects vector to file to produce time series.*)
    if Bayesian then
      begin
      MatrixWriteCov(dim2,covrefile,sigma); (*To detect non-convergence: trace plot of variance-covariance matrix of random effects.*)
      if (dim2 > 1) then (*Smallest latent root of variance-covariance matrix of actor-dependent random effects.*)
        begin
        i := dim2;
        SetMatrix(dim2,dim2,v1,sigma);
        Jacobi(i,v1,eigen,v2,j); (*Spectral decomposition.*)
        minimum := eigen[1]; (*Get smallest latent root.*)
        for i := 2 to dim2 do
          begin
          if (eigen[i] < minimum) then minimum := eigen[i];
          end;
        WriteReal(slrcovrefile,minimum);
        end;
      end;
    end;
  end;
WriteReal(randomeffects,arREfile,blackbox[1]); (*Writes acceptance rate of Metropolis-Hastings algorithm to sample random effects.*)
WriteReal(randomeffects,adllfile,blackbox[3]); (*Writes augmented-data log-likelihood (up to proportionality constant).*)
condition := true;
WriteReal(condition,arPATHfile,blackbox[4]); (*Writes acceptance rate of Metropolis-Hastings algorithm to sample paths: minimum acceptance rate across periods.*)
condition := (Bayesian and randomeffects);
WriteReal(condition,arCovREfile,blackbox[5]); (*Writes acceptance rate of Gibbs or Metropolis-Hastings algorithm to variance-covariance matrices of random effects.*)
WriteReal(Bayesian,arThetafile,blackbox[6],blackbox[7]); (*Writes acceptance rate of Gibbs saMetropolis-Hastings algorithm to sample rate parameters (minimum acceptance rate across periods) and objective function parameters.*)
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MCMC_final_report(var Bayesian: boolean; var ascii: TFileStream; var randomeffects: boolean);
(*Autocorrelations of functionals whose expectation is to be approximated: basic information.*)
var i: byte;
x: string;
begin
if (Bayesian or randomeffects) then (*Bayesian, Hierarchical Bayesian modeling, or frequentist random effects model.*)
  begin
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  x := 'Alternative estimates of the Monte Carlo s.e. (batch means estimates) are given in the output file.'+eol;
  ascii.Write(pchar(x)^,length(x));
  x := 'To be valid as approximation, the lag 1 batch means autocorrelations printed below should not exceed 0.05 much.'+eol;
  ascii.Write(pchar(x)^,length(x));
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  x := 'Observe: the lag 1 batch means autocorrelations are printed in the order:'+eol;
  ascii.Write(pchar(x)^,length(x));
  i := 0;
  if randomeffects then
    begin
    i := i + 1;
    x := format('%s%d%s',['   ',i,'. actor-dependent random effects'])+eol;
    ascii.Write(pchar(x)^,length(x));
    i := i + 1;
    x := format('%s%d%s',['   ',i,'. positive posterior probability of actor-dependent random effects'])+eol;
    ascii.Write(pchar(x)^,length(x));
    i := i + 1;
    x := format('%s%d%s',['   ',i,'. variance-covariance matrix of actor-dependent random effects'])+eol;
    ascii.Write(pchar(x)^,length(x));
    end;
  i := i + 1;
  x := format('%s%d%s',['   ',i,'. first posterior moment of parameter vector'])+eol;
  ascii.Write(pchar(x)^,length(x));
  i := i + 1;
  x := format('%s%d%s',['   ',i,'. second posterior moment of parameter vector'])+eol;
  ascii.Write(pchar(x)^,length(x));
  ascii.Write(pchar(x)^,length(x));//writeln(ascii);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MC_Error_I(var dim: byte; var vector1: pvec; var ldim1: integer; var matrix1: pmatrix; var matrix3: rrarray; var ldim2: integer; var ascii: TFileStream; var input: shortstring);
(*Evaluate Monte Carlo batch means estimator of the Monte Carlo s.e. introduced by
estimating the desired integrals by ergodic averages.*)
var
  lag1: byte;
  i, j, h, divisor: integer;
  u1, u2, mean1, mean2: real;
  s: shortstring;
  f: textfile;
  autocov: R_Ar;
  x : string;
begin
(*
Matrix:
- rows: subchains
- columns: means based on subchains.
Vector: means based on full chain.
*)
for h := 1 to dim do vector1[h] := 0;
for h := 1 to dim do
  begin
  for i := 1 to ldim1 do vector1[h] := vector1[h] + matrix3[h,i];
  end;
if (ldim1 > 0) then divisor := ldim1
else divisor := 1;
for h := 1 to dim do vector1[h] := vector1[h] / divisor;
SetMatrixValue(dim,matrix1,0);
for h := 1 to ldim1 do (*Subchain <h>.*)
  begin
  for i := 1 to dim do (*Evaluate (co)variance(s) corresponding to subchain <h> and add them to the corresponding sums.*)
    begin
    for j := 1 to dim do
      begin
      u1 := matrix3[i,h]; (*Mean <i> based on subchain.*)
      u2 := matrix3[j,h]; (*Mean <j> based on subchain.*)
      mean1 := vector1[i]; (*Mean <i> based on full chain.*)
      mean2 := vector1[j]; (*Mean <j> based on full chain.*)
      matrix1[i,j] := matrix1[i,j] + ((u1 - mean1) * (u2 - mean2));
      end;
    end;
  end;
if (ldim1 > 1) then divisor := (ldim1 - 1)
else divisor := 1;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x := 'Batch means estimates of Monte Carlo standard error'+eol;
ascii.Write(pchar(x)^,length(x));
x := '---------------------------------------------------'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol)); //writeln(ascii);
for i := 1 to dim do
  begin
  for j := 1 to dim do
    begin
    matrix1[i,j] := matrix1[i,j]  / divisor; (*Batch means estimator of Monte Carlo variance, see Gilks et al. (1996), p. 50.*)
    if (matrix1[i,j] >= 0) then begin
      x := format('%10.6f',[sqrt(matrix1[i,j])]);
      ascii.Write(pchar(x)^,length(x));
    end
    else begin
    x :=format('%10.6f',[(999.999999)]);
    ascii.Write(pchar(x)^,length(x));
    end;
    end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
(*In the sister procedure <MC_Error_II> and <MC_Error_III>, the square root of the Monte Carlo variance is taken coordinate-wise to obtain the Monte Carlo s.e.:
it is not done in the present procedure for the simple reason that is is done by standard <Siena> in the output procedures.*)
lag1 := 1;
x := 'Batch means: lag 1 autocorrelation:'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
Autocorrelation(autocov,dim,ldim1,matrix3,lag1,ascii); (*Lag 1 autocorrelation.*)
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
s := input + '.m';
OpenFile(f,s);
for i := 1 to ldim1 do
  begin
  for h := 1 to dim do write(f,matrix3[h,i]:11:6);
  writeln(f);
  end;
closefile(f);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MC_Error_II(var dim: byte; var matrix2: pmatrix; var ldim1: integer; var mc_covre: rrrarray; var matrix1: pmatrix; var ldim2: integer; var ascii: TFileStream; var input: shortstring);
(*Evaluate Monte Carlo batch means estimator of the Monte Carlo s.e. introduced by
estimating the desired integrals by ergodic averages.*)
var
  lag1: byte;
  i, j, h, divisor: integer;
  u: real;
  s: shortstring;
  f: textfile;
  autocov: rrarray;
  x : string;
begin
(*
Matrix:
- rows: subchains
- columns: means based on subchains.
Threedimensional array: means based on full chain.
*)
u := 0;
SetMatrixValue(dim,matrix2,u);
for i := 1 to ldim1 do
  begin
  for j := 1 to dim do
    begin
    for h := 1 to dim do matrix2[j,h] := matrix2[j,h] + mc_covre[j,h,i];
    end;
  end;
if (ldim1 > 0) then ScalarDiv(dim,matrix2,ldim1);
u := 0;
SetMatrixValue(dim,matrix1,u);
for h := 1 to ldim1 do (*Subchain <h>*)
  begin
  for i := 1 to dim do (*Evaluate (co)variance(s) corresponding to subchain <h> and add them to the corresponding sums.*)
    begin
    for j := 1 to dim do matrix1[i,j] := matrix1[i,j] + sqr(mc_covre[i,j,h] - matrix2[i,j]);
    end;
  end;
if (ldim1 > 1) then divisor := (ldim1 - 1)
else divisor := 1;
ScalarDiv(dim,matrix1,divisor); (*Batch means estimator of Monte Carlo variance, see Gilks et al. (1996), p. 50, and in particular Geyer (1992), Statistical Science.*)
MatrixSqrt(dim,matrix1); (*Batch means estimates of Monte Carlo standard devation.*)
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x := 'Batch means estimates of Monte Carlo standard error'+eol;
ascii.Write(pchar(x)^,length(x));
x := '---------------------------------------------------';
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
for i := 1 to dim do
  begin
  for j := 1 to i do begin
    x := format('%10.6f',[matrix1[i,j]]); (*Other than in the sister procedure <MC_ErrorI>, here the square root has already been computed.*)
    ascii.Write(pchar(x)^,length(x));
  end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
lag1 := 1;
i := dim;
x := 'Batch means: lag 1 autocorrelation:'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));// writeln(ascii);
Autocorrelation(autocov,dim,i,ldim1,mc_covre,lag1,ascii); (*Lag 1 autocorrelation.*)
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
s := input + '.m';
OpenFile(f,s);
for i := 1 to ldim1 do
  begin
  for j := 1 to dim do
    begin
    for h := 1 to dim do write(f,mc_covre[j,h,i]:11:6);
    end;
  writeln(f);
  end;
closefile(f);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MC_Error_III(var eb: rrarray; var actors: integer; var dim: byte; var ldim1: integer; var mc_re: rrrarray; var matrix3: rrarray;
var ldim2: integer; var ascii: TFileStream; var input: shortstring);
(*Evaluate Monte Carlo batch means estimator of the Monte Carlo s.e. introduced by
estimating the desired integrals by ergodic averages.*)
var
  lag1: byte;
  i, j, h, divisor: integer;
  s: shortstring;
  f: textfile;
  autocov: rrarray;
  x : string;
begin
(*
Matrix:
- rows: subchains
- columns: means based on subchains.
Threedimensional array: means based on full chain.
*)
eb := nil;
SetLength(eb,actors+1,dim+1);
for i := 1 to ldim1 do
  begin
  for j := 1 to actors do
    begin
    for h := 1 to dim do eb[j,h] := eb[j,h] + mc_re[j,h,i];
    end;
  end;
if (ldim1 > 0) then ScalarDiv(actors,dim,eb,ldim1);
matrix3 := nil;
SetLength(matrix3,actors+1,dim+1);
for h := 1 to ldim1 do (*Subchain <h>*)
  begin
  for i := 1 to actors do (*Evaluate (co)variance(s) corresponding to subchain <h> and add them to the corresponding sums.*)
    begin
    for j := 1 to dim do matrix3[i,j] := matrix3[i,j] + sqr(mc_re[i,j,h] - eb[i,j]);
    end;
  end;
if (ldim1 > 1) then divisor := (ldim1 - 1)
else divisor := 1;
ScalarDiv(actors,dim,matrix3,divisor); (*Batch means estimator of Monte Carlo variance, see Gilks et al. (1996), p. 50, and in particular Geyer (1992), Statistical Science.*)
MatrixSqrt(actors,dim,matrix3); (*Batch means estimates of Monte Carlo standard devation.*)
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x := 'Batch means estimates of Monte Carlo standard error'+eol;
ascii.Write(pchar(x)^,length(x));
x := '---------------------------------------------------'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
for i := 1 to actors do
  begin
  for j := 1 to dim do begin
    x := format('%10.6f',[matrix3[i,j]]); (*Other than in the sister procedure <MC_ErrorI>, here the square root has already been computed.*)
    ascii.Write(pchar(x)^,length(x));
  end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
lag1 := 1;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x := 'Batch means: lag 1 autocorrelation:'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
Autocorrelation(autocov,dim,actors,ldim1,mc_re,lag1,ascii); (*Lag 1 autocorrelation.*)
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
s := input + '.m';
OpenFile(f,s);
for i := 1 to ldim1 do
  begin
  for j := 1 to dim do
    begin
    for h := 1 to dim do write(f,mc_re[j,h,i]:11:6);
    end;
  writeln(f);
  end;
closefile(f);
end;

(*--------------------------------------------------------------------------------------------------*)
 procedure Autocorrelation(var autocov1: R_Ar; var dim: byte; var ldim: integer; var autovector: rrarray; var lag: byte; var file1: textfile); overload;
(*Evaluate the lag <lag> sample auto-covariance/correlation function. If the acceptance rate of the Metropolis-Hastings algorithm is pushed towards
some specified value, then a fair comparison between autocorrelation and iid sampling would be based on every 1/optimal rate-th observation.*)
const u = 1.0000;
var
  l: integer;
  h: byte;
  average, denominator, numerator: extended;
begin
SetArray(autocov1,dim);
for h := 1 to dim do
  begin
  average := 0;
  numerator := 0;
  denominator := 0;
  for l := 1 to ldim do average := average + autovector[h,l]; (*Average over the sample.*)
  if (ldim > 0) then average := average / ldim
  else average := 0;
  if (average > 1e+300) then average := 0;
  for l := 1 to (ldim - lag) do
    begin
    numerator := numerator + ((autovector[h,l] - average) * (autovector[h,l+lag] - average));
    denominator := denominator + sqr(autovector[h,l] - average); (*Average squared deviation from mean over the sample...*)
    end;
  for l := (ldim - lag + 1) to ldim do denominator := denominator + sqr(autovector[h,l] - average); (*...and complete denominator.*)
  autocov1[h] := numerator / ldim; (*Store autocovariance in array (the divisor cancels out in the autocorrelations case)...*)
  if (denominator > 0) then write(file1,' ',(numerator / denominator):10:6)  (*...and write autocorrelation to file.*)
  else write(file1,' ',u:10:6);
  end;
writeln(file1);
end;

procedure Autocorrelation(var autocov1: R_Ar; var dim: byte; var ldim: integer; var autovector: rrarray; var lag: byte; var file1, file2: textfile); overload;
(*Evaluate the lag <lag> sample auto-covariance/correlation function. If the acceptance rate of the Metropolis-Hastings algorithm is pushed towards
some specified value, then a fair comparison between autocorrelation and iid sampling would be based on every 1/optimal rate-th observation.*)
const u = 1.0000;
var
  l: integer;
  h: byte;
  average, denominator, numerator: extended;
begin
SetArray(autocov1,dim);
for h := 1 to dim do
  begin
  average := 0;
  numerator := 0;
  denominator := 0;
  for l := 1 to ldim do average := average + autovector[h,l]; (*Average over the sample.*)
  if (ldim > 0) then average := average / ldim
  else average := 0;
  if (average > 1e+300) then average := 0;
  for l := 1 to (ldim - lag) do
    begin
    numerator := numerator + ((autovector[h,l] - average) * (autovector[h,l+lag] - average));
    denominator := denominator + sqr(autovector[h,l] - average); (*Average squared deviation from mean over the sample...*)
    end;
  for l := (ldim - lag + 1) to ldim do denominator := denominator + sqr(autovector[h,l] - average); (*...and complete denominator.*)
  autocov1[h] := numerator / ldim; (*Store autocovariance in array (the divisor cancels out in the autocorrelations case)...*)
  if (denominator > 0) then
    begin
    write(file1,' ',(numerator / denominator):10:6);  (*...and write autocorrelation to file.*)
    write(file2,' ',(numerator / denominator):10:6);  (*...and write autocorrelation to file.*)
    end
  else
    begin
    write(file1,' ',u:10:6);
    write(file2,' ',u:10:6);
    end;
  end;
writeln(file1);
writeln(file2);
end;

procedure Autocorrelation(var autocov2: rrarray; var dim: byte; var actors, ldim: integer; var automatrix: rrrarray; var lag: byte; var file1: textfile); overload;
(*Evaluate the lag <lag> sample auto-covariance/correlation function. If the acceptance rate of the Metropolis-Hastings algorithm is pushed towards
some specified value, then a fair comparison between autocorrelation and iid sampling would be based on every 1/optimal rate-th observation.*)
const u = 1.0000;
var
  i, l: integer;
  h: byte;
  average, denominator, numerator: extended;
begin
SetArray(autocov2,actors,dim);
for h := 1 to dim do
  begin
  for i := 1 to actors do
    begin
    average := 0;
    numerator := 0;
    denominator := 0;
    for l := 1 to ldim do average := average + automatrix[i,h,l]; (*Average over the sample.*)
    if (ldim > 0) then average := average / ldim
    else average := 0;
    if (average > 1e+300) then average := 0;
    for l := 1 to (ldim - lag) do
      begin
      numerator := numerator + ((automatrix[i,h,l] - average) * (automatrix[i,h,l+lag] - average));
      denominator := denominator + sqr(automatrix[i,h,l] - average); (*Average squared deviation from mean over the sample...*)
      end;
    for l := (ldim - lag + 1) to ldim do denominator := denominator + sqr(automatrix[i,h,l] - average); (*...and complete denominator.*)
    autocov2[i,h] := numerator / ldim; (*Store autocovariance in array (the divisor cancels out in the autocorrelations case)...*)
    if (denominator > 0) then write(file1,' ',(numerator / denominator):10:6)  (*...and write autocorrelation to file.*)
    else write(file1,' ',u:10:6);
    end;
  end;
writeln(file1);
end;


procedure Autocorrelation(var autocov1: R_Ar; var dim: byte; var ldim: integer; var autovector: rrarray; var lag: byte; var file1: TFileStream); overload;
(*Evaluate the lag <lag> sample auto-covariance/correlation function. If the acceptance rate of the Metropolis-Hastings algorithm is pushed towards
some specified value, then a fair comparison between autocorrelation and iid sampling would be based on every 1/optimal rate-th observation.*)
const u = 1.0000;
var
  l: integer;
  h: byte;
  average, denominator, numerator: extended;
  x: string;
begin
SetArray(autocov1,dim);
for h := 1 to dim do
  begin
  average := 0;
  numerator := 0;
  denominator := 0;
  for l := 1 to ldim do average := average + autovector[h,l]; (*Average over the sample.*)
  if (ldim > 0) then average := average / ldim
  else average := 0;
  if (average > 1e+300) then average := 0;
  for l := 1 to (ldim - lag) do
    begin
    numerator := numerator + ((autovector[h,l] - average) * (autovector[h,l+lag] - average));
    denominator := denominator + sqr(autovector[h,l] - average); (*Average squared deviation from mean over the sample...*)
    end;
  for l := (ldim - lag + 1) to ldim do denominator := denominator + sqr(autovector[h,l] - average); (*...and complete denominator.*)
  autocov1[h] := numerator / ldim; (*Store autocovariance in array (the divisor cancels out in the autocorrelations case)...*)
  if (denominator > 0) then begin
    x := format('%s%10.6f',[' ',(numerator / denominator)]); (*...and write autocorrelation to file.*)
    file1.Write(pchar(x)^,length(x));
  end
  else begin
    x := format('%s%10.6f',[' ',u]);
    file1.Write(pchar(x)^,length(x));
  end;
  end;
file1.Write(pchar(eol)^,length(eol));//writeln(file1);
end;

procedure Autocorrelation(var autocov1: R_Ar; var dim: byte; var ldim: integer; var autovector: rrarray; var lag: byte; var file1: TFileStream; var file2: textfile); overload;
(*Evaluate the lag <lag> sample auto-covariance/correlation function. If the acceptance rate of the Metropolis-Hastings algorithm is pushed towards
some specified value, then a fair comparison between autocorrelation and iid sampling would be based on every 1/optimal rate-th observation.*)
const u = 1.0000;
var
  l: integer;
  h: byte;
  average, denominator, numerator: extended;
  x: string;
begin
SetArray(autocov1,dim);
for h := 1 to dim do
  begin
  average := 0;
  numerator := 0;
  denominator := 0;
  for l := 1 to ldim do average := average + autovector[h,l]; (*Average over the sample.*)
  if (ldim > 0) then average := average / ldim
  else average := 0;
  if (average > 1e+300) then average := 0;
  for l := 1 to (ldim - lag) do
    begin
    numerator := numerator + ((autovector[h,l] - average) * (autovector[h,l+lag] - average));
    denominator := denominator + sqr(autovector[h,l] - average); (*Average squared deviation from mean over the sample...*)
    end;
  for l := (ldim - lag + 1) to ldim do denominator := denominator + sqr(autovector[h,l] - average); (*...and complete denominator.*)
  autocov1[h] := numerator / ldim; (*Store autocovariance in array (the divisor cancels out in the autocorrelations case)...*)
  if (denominator > 0) then
    begin
    x := format('%s%10.6f',[' ',(numerator / denominator)]);  (*...and write autocorrelation to file.*)
    file1.Write(pchar(x)^,length(x));
    write(file2,' ',(numerator / denominator):10:6);  (*...and write autocorrelation to file.*)
    end
  else
    begin
    x := format('%s%10.6f',[' ',u]);
    file1.Write(pchar(x)^,length(x));
    write(file2,' ',u:10:6);
    end;
  end;
file1.Write(pchar(eol)^,length(eol));//writeln(file1);
writeln(file2);
end;

procedure Autocorrelation(var autocov2: rrarray; var dim: byte; var actors, ldim: integer; var automatrix: rrrarray; var lag: byte; var file1: TFileStream); overload;
(*Evaluate the lag <lag> sample auto-covariance/correlation function. If the acceptance rate of the Metropolis-Hastings algorithm is pushed towards
some specified value, then a fair comparison between autocorrelation and iid sampling would be based on every 1/optimal rate-th observation.*)
const u = 1.0000;
var
  i, l: integer;
  h: byte;
  average, denominator, numerator: extended;
  x : string;
begin
SetArray(autocov2,actors,dim);
for h := 1 to dim do
  begin
  for i := 1 to actors do
    begin
    average := 0;
    numerator := 0;
    denominator := 0;
    for l := 1 to ldim do average := average + automatrix[i,h,l]; (*Average over the sample.*)
    if (ldim > 0) then average := average / ldim
    else average := 0;
    if (average > 1e+300) then average := 0;
    for l := 1 to (ldim - lag) do
      begin
      numerator := numerator + ((automatrix[i,h,l] - average) * (automatrix[i,h,l+lag] - average));
      denominator := denominator + sqr(automatrix[i,h,l] - average); (*Average squared deviation from mean over the sample...*)
      end;
    for l := (ldim - lag + 1) to ldim do denominator := denominator + sqr(automatrix[i,h,l] - average); (*...and complete denominator.*)
    autocov2[i,h] := numerator / ldim; (*Store autocovariance in array (the divisor cancels out in the autocorrelations case)...*)
    if (denominator > 0) then begin
    x := format('%s%10.6f',[' ',(numerator / denominator)]);  (*...and write autocorrelation to file.*)
    file1.Write(pchar(x)^,length(x));
    end
    else begin
      x:= format('%s%10.6f',[' ',u]);
      file1.Write(pchar(x)^,length(x));
    end;
    end;
  end;
file1.Write(pchar(eol)^,length(eol));//writeln(file1);
end;

(*--------------------------------------------------------------------------------------------------*)

function ISE(const autocov: array of real; const dim: byte): real;
(*Evaluate Initial Sequence Estimates of the Monte Carlo variance,
in particular the Initial Monotone Sequence Estimate;
see Geyer (1992) and the Gilks et al. (1996), p. 50, which contains some minor errors.*)
var
  m: integer;
  lastsum, sum, se: real;
  violation: boolean;
begin
(*Remark:
note that the Initial Convex Sequence Estimate of the Monte Carlo variance
can readily be obtained, but the added value is probably outweighted by the additional effort.*)
se := autocov[1]; (*Must be added here, since it is not added in the loop below.*)
lastsum := autocov[2] + autocov[3]; (*If <m = 1>, then <sum = lastsum> and <sucess> is true if and only if <sum < 0>.*)
m := 0;
repeat
Inc(m);
sum := autocov[2*m] + autocov[(2*m)+1]; (*The estimated function <Gamma_m> of Geyer (1992).*)
violation := (sum <= 0) or (sum > lastsum); (*The estimated function <Gamma_m> must be positive and monotone as a function of <m>.*)
if not(violation) then
  begin
  se := se + sum;
  lastsum := sum; (*Store last value of the estimated function <Gamma_m>.*)
  end;
until (violation or (m = dim)); (*<dim> is the maximum lag t autocovariance available here.*)
ISE := autocov[0] + (2 * se); (*Monte Carlo variance: Initial Monotone Sequence Estimate.*)
end;

(*--------------------------------------------------------------------------------------------------*)

procedure IEFactor(var ascii: TFileStream; var ldim: integer; var dim: byte; var autovector: rrarray; var ISE_var: R_Ar); overload;
(*Evaluate
- lag <t> autocorrelations, <t = 0, 1, 2, ..., m>.
- Monte Carlo standard errors as estimated by Initial Sequence Estimates (Geyer, 1992)
- inefficiency factors (Chib, 2004)
which is all written to the file <ascii>.
*)
const m = 100;
var
  i, h: byte;
  u: real;
  s: shortstring;
  f: textfile;
  autocov: R_Ar;
  auto_covariance: rrarray;
  x: string;
begin
SetArray(auto_covariance,dim,m);
SetArray(ISE_var,dim);
x := format('%s%d',['Lag t autocorrelations, t = 0, ..., ',m])+eol;
ascii.Write(pchar(x)^,length(x));
x := '---------------------------------------'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x := 'Parameter vector.'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));// writeln(ascii);
x := format('%s%d%s',['Autocorrelations based on ',ldim,' iterations.'])+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
s := CurrentProjectName + 'AATheta.m';
OpenFile(f,s);
for i := 0 to m do (*Sample lag <i> auto-covariance/correlation function, <i = 0, 1, 2, ..., m>.*)
  begin
  h := i;
  Autocorrelation(autocov,dim,ldim,autovector,h,ascii,f); (*Evaluate the sample lag <i> auto-covariance/correlation function.*)
  for h := 1 to dim do auto_covariance[h,i] := autocov[h]; (*Store the autocovariances.*)
  end;
closefile(f);
ascii.Write(pchar(eol)^,length(eol));ascii.Write(pchar(eol)^,length(eol));//writeln(ascii); writeln(ascii);
x := 'Efficiency of MCMC sampling scheme:'+eol;
ascii.Write(pchar(x)^,length(x));
x := '-----------------------------------'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x:= format('%s%d%s',['Monte Carlo standard error (Initial Monotone Sequence Estimate: Geyer, 1992) based on ',ldim,' iterations: '])+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
for h := 1 to dim do
  begin
  ISE_var[h] := ISE(auto_covariance[h],m); (*Initial Monotone Sequence Estimate of the Monte Carlo variance; see Geyer (1992).*)
  x := format('%10.6f',[sqrt(ISE_var[h])]);
  ascii.Write(pchar(x)^,length(x));
  end;
ascii.Write(pchar(eol)^,length(eol));ascii.Write(pchar(eol)^,length(eol));//writeln(ascii); writeln(ascii);
x := format('%s%d%s',['Inefficiency factors based on ',ldim,' iterations '])+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));// writeln(ascii);
for h := 1 to dim do
  begin
  if (auto_covariance[h,0] > 0) then u := ISE_var[h] / auto_covariance[h,0]
  else u := 999.999999;
  x := format('%10.6f',[u]); (*Initial Monotone Sequence Estimate of the Monte Carlo variance divided by lag <0> autocovariance.*)
  ascii.Write(pchar(x)^,length(x));
  end;
ascii.Write(pchar(eol)^,length(eol));ascii.Write(pchar(eol)^,length(eol));//writeln(ascii); writeln(ascii);
autocov := nil;
auto_covariance := nil;
end;

procedure IEFactor(var ascii: TFileStream; const indicator: byte; var iterations: integer; var dimension1: integer; var dimension2: byte; var automatrix: rrrarray; var ISE_var: rrarray); overload;
(*Evaluate
- lag <t> autocorrelations, <t = 0, 1, 2, ..., m>.
- Monte Carlo standard errors as estimated by Initial Sequence Estimates (Geyer, 1992)
- inefficiency factors (Chib, 2004)
which is all written to the file <ascii>.
*)
const m = 100;
var
  i, h: byte;
  j: integer;
  u: real;
  auto: R_Ar;
  autocov: rrarray;
  auto_covariance: rrrarray;
  x: string;
begin
(*Remark:
If the matrix in question is the matrix of actor-dependent random effects, then
- <dimension1> is the number of actors
- <dimension2> is the number of actor-dependent random effects.
If the matrix in questions is the variance-covariance matrix of actor-dependent random effects,
then <dimension1 = dimension2> is the number of actor-dependent random effects.
*)
SetArray(auto_covariance,dimension1,dimension2,m);
SetArray(ISE_var,dimension1,dimension2);
x := format('%s%d',['Lag t autocorrelations, t = 0, ..., ',m])+eol;
ascii.Write(pchar(x)^,length(x));
x := '---------------------------------------'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
case indicator of
  1: begin
      x := 'Actor-dependent random effects.'+eol;
      ascii.Write(pchar(x)^,length(x));
  end;
  2: begin
    x := 'Variance-covariance matrix of actor-dependent random effects.'+eol;
    ascii.Write(pchar(x)^,length(x));
  end;
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x := format('%s%d%s',['Autocorrelations based on ',iterations,' iterations.'])+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
for i := 0 to m do (*Sample lag <i> auto-covariance/correlation function, <i = 0, 1, 2, ..., m>.*)
  begin
  h := i;
  Autocorrelation(autocov,dimension2,dimension1,iterations,automatrix,h,ascii); (*Evaluate the sample lag <i> auto-covariance/correlation function.*)
  for j := 1 to dimension1 do
    begin
    for h := 1 to dimension2 do auto_covariance[j,h,i] := autocov[j,h]; (*Store the autocovariances.*)
    end;
  end;
ascii.Write(pchar(eol)^,length(eol));ascii.Write(pchar(eol)^,length(eol));//writeln(ascii); writeln(ascii);
x := 'Efficiency of MCMC sampling scheme:'+eol;
ascii.Write(pchar(x)^,length(x));
x := '-----------------------------------'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));// writeln(ascii);
x := format('%s%d%s',['Monte Carlo standard error (Initial Monotone Sequence Estimate: Geyer, 1992) based on ',iterations,' iterations: '])+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
for h := 1 to dimension2 do (*To write the vectors <1..dimension2> as rows to the file is preferable, since <dimension1> is commonly much larger than <dimension2>.*)
  begin
  for j := 1 to dimension1 do
    begin
    SetArray(auto,m);
    for i := 0 to m do auto[i] := auto_covariance[j,h,i];
    ISE_var[j,h] := ISE(auto,m); (*Initial Monotone Sequence Estimate of the Monte Carlo variance; see Geyer (1992).*)
    x := format('%10.6f',[sqrt(ISE_var[j,h])]);
    ascii.Write(pchar(x)^,length(x));
    end;
   ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  end;
ascii.Write(pchar(eol)^,length(eol));ascii.Write(pchar(eol)^,length(eol));//writeln(ascii); writeln(ascii);
x := format('%s%d%s',['Inefficiency factors based on ',iterations,' iterations '])+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
for h := 1 to dimension2 do (*To write the vectors <1..dimension2> as rows to the file is preferable, since <dimension1> is commonly much larger than <dimension2>.*)
  begin
  for j := 1 to dimension1 do
    begin
    if (auto_covariance[j,h,0] > 0) then u := ISE_var[j,h] / auto_covariance[j,h,0]
    else u := 999.999999;
    x := format('%410.6f',[u]); (*Initial Monotone Sequence Estimate of the Monte Carlo variance divided by lag <0> autocovariance.*)
    ascii.Write(pchar(x)^,length(x));
    end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
auto := nil;
autocov := nil;
auto_covariance := nil;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure SampleMVN(var mvndensities: R_Ar; const actors: integer; const dim: byte; const cholfac: pmatrix;
const scaleRE_i: R_Ar; var mvn: rrarray); overload;
(*Algorithm:
- sample <z ~ N(0, I)>, where <I> is the identity matrix.
- set <mvn = C z>, where <C> is the Cholesky factor of variance-covariance matrix <Sigma>,
  so that <mvn ~ N(0, Sigma)>.
Note that <C> and hence <Sigma> are rescaled.*)
var
  i, j, h: integer;
  constant, sum, density: real;
  z: pvec;
  c: pmatrix;
begin
sum := 0;
SetArray(mvndensities,actors);
for i := 1 to actors do
  begin
  constant := scaleRE_i[i];
  ScalarProd(dim,c,cholfac,constant); (*Rescale Cholesky factor <CholFac> to obtain <C>; note that <CholFac> remains unchanged.*)
  for h := 1 to dim do z[h] := RandG(0,1); (*Sample <z ~ N(0, I) using Delphi function RandG: fast Marsaglia-Bray (1964) method to sample from N(0,1): Devroye (1986, p. 381).*)
  for j := 1 to dim do (*Set <mvn = C z>.*)
    begin
    mvn[i,j] := 0;
    for h := 1 to dim do
      begin
      mvn[i,j] := mvn[i,j] + (c[j,h] * z[h]);
      end;
    end;
  density := 0;
  for h := 1 to dim do density := density + sqr(z[h]); (*Squared length of <z>.*)
  mvndensities[i] := - (density / 2); (*Logarithm of the kernel of <N(0, I)>.*)
  sum := sum + mvndensities[i]; (*Logarithm of the kernel of joint <N(0, I)>.*)
  end;
mvndensities[0] := sum; (*Logarithm of the kernel of joint <N(0, I)>.*)
end;

(*--------------------------------------------------------------------------------------------------*)

function SampleMVN(const dim: byte; const cholfac: pmatrix; const scale: real; var vector1: pvec): real; overload;
(*Algorithm:
- sample <z ~ N(0, I)>, where <I> is the identity matrix.
- set <mvn = C z>, where <C> is the Cholesky factor of variance-covariance matrix <Sigma>,
  so that <mvn ~ N(0, Sigma)>.
Note that <C> and hence <Sigma> are rescaled.*)
var
  j, h: integer;
  density: real;
  z: pvec;
  c: pmatrix;
begin
ScalarProd(dim,c,cholfac,scale); (*Rescale Cholesky factor <CholFac> to obtain <C>; note that <CholFac> remains unchanged.*)
for h := 1 to dim do z[h] := RandG(0,1); (*Sample <z ~ N(0, I) using Delphi function RandG: fast Marsaglia-Bray (1964) method to sample from N(0,1): Devroye (1986, p. 381).*)
for j := 1 to dim do (*Set <vector1 = C z>.*)
  begin
  vector1[j] := 0;
  for h := 1 to dim do
    begin
    vector1[j] := vector1[j] + (c[j,h] * z[h]);
    end;
  end;
density := 0;
for h := 1 to dim do density := density + sqr(z[h]); (*Squared length of <z>.*)
SampleMVN := - (density / 2); (*Logarithm of the kernel of <N(0, I)>.*)
end;

(*--------------------------------------------------------------------------------------------------*)

function SampleGamma(var ldim: integer; var real1: real): real;
(*Generate one draw from the Gamma distribution with parameters <ldim> and <real1>: see Robert & Casella, 1999, p. 45.*)
var
  i: integer;
  sum: real;
begin
sum := 0;
for i := 1 to ldim do sum := sum + ln(random);
SampleGamma := - (sum / real1);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure SampleW(var dim: byte; var ldim: integer; var cholfac, ssigma: pmatrix);
(*Generate one draw from the Wishart distributionl, parametrized such that $E(\Sigma^{-1}) = m \Omega$, where $m$ is the degree of freedoms parameter, and $\Omega$ the scale matrix.*)
var
  i, j, h: byte;
  sum: real;
  dmatrix, inter, tcholfac, z: pmatrix;
begin
for i := 1 to dim do z[i,i] := Random_ChiSq(ldim-i+1);(*Chi-square variate with <#ldim - i> degrees of freedom.*)
for i := 2 to dim do (*Standard Gaussian variates.*)
  begin
  for j := 1 to (i - 1) do
    begin
    z[i,j] := RandG(0,1); (*RandG: Delphi function: fast Marsaglia-Bray (1964) method to sample from N(0,1): Devroye (1986, p. 381).*)
    z[j,i] := z[i,j]; (*Symmetry*)
    end;
  end;
dmatrix[1,1] := z[1,1];
for i := 2 to dim do (*Generate draws from N(0,1) and chi(1) and compute the diagonal elements.*)
  begin
  sum := 0;
  for j := 1 to (i - 1) do sum := sum + sqr(z[j,i]);
  dmatrix[i,i] := z[i,i] + sum; (*Draw from chi-distribution.*)
  end;
for j := 2 to dim do dmatrix[1,j] := z[1,j] * sqrt(z[1,1]);
for i := 1 to dim do (*Compute the off-diagonal elements if there are any.*)
  begin
  for j := (i + 1) to dim do (*Above diagonal elements.*)
    begin
    sum := 0;
    for h := 1 to (i - 1) do sum := sum + (z[h,i] * z[h,j]); (*These above-diagonal elements are already at disposal.*)
    dmatrix[i,j] := (z[i,j] * sqrt(dmatrix[i,i])) + sum; (*Off-diagonal element.*)
    dmatrix[j,i] := dmatrix[i,j]; (*Complete matrix.*)
    end;
  end;
MatrixT(dim,tcholfac,cholfac); (*Transpose Cholesky factor.*)
MatrixProd(dim,dim,inter,cholfac,dmatrix); (*Compute <Cholesky factor times sample matrix>.*)
MatrixProd(dim,dim,ssigma,inter,tcholfac); (*Compute <Cholesky factor times sample matrix times transposed Cholesky factor>.*)
(*Do not (!) divide by the number of degrees of freedom.*)
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CandidateMH(var mvn0: rrarray; const actors: integer; const dim: byte; var mvndensities: R_Ar; const cholfac: pmatrix;
const scaleRE_i: R_Ar; const cdensity: byte; var mvn: rrarray; const re, intermediate: rrarray); overload;
(*Generates candidates for Metropolis-Hastings algorithms with multivariate normal candidate-generating densities: one or more vectors: one vector.*)
var i, h: integer;
begin
(*
1 = random walk density
2 = autoregressive density: Markov chain reflected about <intermediate> (= Empirical Bayes estimate approximated by the ergodic average based on the last Markov chain)
3 = independent density, centered at origin (independence sampler).
*)
SetArray(mvn0,actors,dim);
SampleMVN(mvndensities,actors,dim,cholfac,scaleRE_i,mvn0);
if (cdensity = 1) then (*Random-walk candidate-generating density for Metropolis-Hastings algorithm: Metropolis, Rosenbluth, Rosenbluth, Teller and Teller (1953).*)
  begin
  for i := 1 to actors do
    begin
    for h := 1 to dim do mvn[i,h] := re[i,h] + mvn0[i,h];
    end;
  end
else if (cdensity = 2) then (*Autoregressive candidate-generating density for Metropolis-Hastings algorithm: Tierney (1994).*)
  begin
  for i := 1 to actors do
    begin
    for h := 1 to dim do mvn[i,h] := (2 * intermediate[i,h]) - re[i,h] + mvn0[i,h]; (*Markov chain is reflected about the Empirical Bayes estimate approximated by the ergodic average based on the last Markov chain.*)
    end;
  end
else if (cdensity = 3) then (*Independence sampler.*)
  begin
  for i := 1 to actors do
    begin
    for h := 1 to dim do mvn[i,h] := intermediate[i,h] + mvn0[i,h];
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

function CandidateMH(const dim: byte; const cholfac: pmatrix; const scale: real;
var vector1: pvec; const cdensity: byte; var vector10: pvec; const vector2, vector3: pvec): real; overload;
(*Generates candidates for Metropolis-Hastings algorithms with multivariate normal candidate-generating densities.*)
var
  h: integer;
  u: real;
begin
(*
1 = random walk density
2 = autoregressive density: Markov chain reflected about <vector3>
3 = independent density, centered at point <vector3> (independence sampler).
*)
u := SampleMVN(dim,cholfac,scale,vector1);
SetVector(dim,vector10,vector1);
CandidateMH := u;
if (cdensity = 1) then (*Random-walk candidate-generating density for Metropolis-Hastings algorithm: Metropolis, Rosenbluth, Rosenbluth, Teller and Teller (1953).*)
  begin
  for h := 1 to dim do vector1[h] := vector2[h] + vector1[h]; (*<vector1> stores the multivariate normal variate, <vector2> the present vector.*)
  end
else if (cdensity = 2) then (*Autoregressive candidate-generating density for Metropolis-Hastings algorithm: Tierney (1994).*)
  begin
  for h := 1 to dim do vector1[h] := (2 * vector3[h]) - vector2[h] + vector1[h]; (*Markov chain is reflected about <vector3>.*)
  end
else if (cdensity = 3) then (*Independence chain with origin at point <vector3>.*)
  begin
  for h := 1 to dim do vector1[h] := vector3[h] + vector1[h]; (*<vector3> must be constant across iterations.*)
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

function EvaluatePoisson(var ldim1: integer; var real1: real; const real2: real; const ldim2: integer): real;
(*Evaluate Poisson kernel.*)
var u: real;
begin
(*
<ldim1> number of actors
<real1> value of rate parameter: present period
<real2> length: present period
<ldim2> number of transitions: present period.
*)
u := ldim1 * real1 * real2; (*In the notation of the paper: $n \alpha (t_1 - t_0)$.*)
EvaluatePoisson := - u + (ldim2 * ln(u));
end;

(*--------------------------------------------------------------------------------------------------*)

function EvaluateGamma(var dim: byte; var vector1: pvec; var gamma1: real; var vector2: pvec; var gamma2: real): real;
(*Evaluate Gamma kernel with parameters <gamma1> and <gamma2> for each coordinate of some vector valued random variate.*)
var
  i: byte;
  sum: real;
begin
sum := 0;
for i := 1 to dim do (*On natural logarithmic scale.*)
  begin
  vector1[i] := (gamma1 - 1) * ln(vector2[i]);
  vector1[i] := vector1[i] - (gamma2 * vector2[i]);
  sum := sum + vector1[i];
  end;
EvaluateGamma := sum;
end;

(*--------------------------------------------------------------------------------------------------*)

function EvaluateMVN(var mvndensities: R_Ar; const actors: integer; const dim: byte; const re: rrarray; const isigma: pmatrix): real; overload;
(*Evaluate multivariate Gaussian kernel: one or more vectors.*)
var
  i, j, h: integer;
  density, sum: real;
begin
sum := 0;
SetArray(mvndensities,actors);
for i := 1 to actors do
  begin
  density := 0;
  for j := 1 to dim do (*For each actor, compute <v_i' inverse(Sigma) v_i>.*)
    begin
    for h := 1 to dim do
      begin
      density := density + (re[i,j] * isigma[j,h] * re[i,h]);
      end;
    end;
  mvndensities[i] := - (density / 2);
  sum := sum + mvndensities[i];
  end;
mvndensities[0] := sum;
EvaluateMVN := sum;
end;

(*--------------------------------------------------------------------------------------------------*)

function EvaluateMVN(const dim: byte; const vector1: pvec; const isigma: pmatrix): real; overload;
(*Evaluate multivariate Gaussian kernel: one vector.*)
var
  j, h: integer;
  density: real;
begin
density := 0;
for j := 1 to dim do
  begin
  for h := 1 to dim do
    begin
    density := density + (vector1[j] * isigma[j,h] * vector1[h]);
    end;
  end;
EvaluateMVN := - (density / 2);
end;

(*--------------------------------------------------------------------------------------------------*)

function EvaluateFunctionMVN1(var mvndensities: R_Ar; var ldim: integer; var dim: byte; var re: rrarray; var scaleRE_i: R_Ar): real;
(*Evaluate multivariate Gaussian kernel: one or more actors, variance-covariance matrix equal to identity matrix.*)
var
  i: integer;
  h: byte;
  sum, density: real;
begin
sum := 0;
mvndensities := nil;
SetLength(mvndensities,ldim+1);
for i := 1 to ldim do
  begin
  density := 0;
  for h := 1 to dim do density := density + sqr(re[i,h]); (*Evaluate length of actor-dependent random effects vector.*)
  mvndensities[i] := - (density / 2); (*Evaluate multivariate normal probability density of candidate for the specified actor.*)
  sum := sum + mvndensities[i]; (*Evaluate joint multivariate normal probability density of the candidate.*)
  end;
mvndensities[0] := sum; (*The natural logarithm of the joint multivariate normal probability density of the candidate.*)
EvaluateFunctionMVN1 := sum; (*The natural logarithm of the joint multivariate normal probability density of the candidate.*)
end;

(*--------------------------------------------------------------------------------------------------*)

function EvaluateFunctionMVN2(var mvndensities: R_Ar; const actors: integer; const dim: byte; const re: rrarray; const isigma: pmatrix; const scaleRE_i: R_Ar): real; overload;
(*Evaluate multivariate Gaussian: one or more vectors, variance-covariance matrix rescaled.*)
var
  i, j, h: integer;
  density, sum: real;
begin
sum := 0;
SetArray(mvndensities,actors);
for i := 1 to actors do
  begin
  density := 0;
  for j := 1 to dim do (*For each actor, compute <v_i' inverse(Sigma) v_i>.*)
    begin
    for h := 1 to dim do
      begin
      density := density + (re[i,j] * isigma[j,h] * re[i,h]);
      end;
    end;
  density := density / sqr(scaleRE_i[i]);
  mvndensities[i] := - (density / 2);
  sum := sum + mvndensities[i];
  end;
mvndensities[0] := sum;
EvaluateFunctionMVN2 := sum;
end;

(*--------------------------------------------------------------------------------------------------*)

function EvaluateFunctionMVN2(const dim: byte; const vector1: pvec; const isigma: pmatrix; const scale: real): real; overload;
(*Evaluate multivariate Gaussian: one vector, variance-covariance matrix rescaled.*)
var
  j, h: integer;
  density: real;
begin
density := 0;
for j := 1 to dim do 
  begin
  for h := 1 to dim do
    begin
    density := density + (vector1[j] * isigma[j,h] * vector1[h]);
    end;
  end;
density := density / sqr(scale);
EvaluateFunctionMVN2 := - (density / 2);
end;

(*--------------------------------------------------------------------------------------------------*)

function EvaluateMVNactor(const dim: byte; const re: rrarray; const ldim: integer; const isigma: pmatrix): real;
(*Evaluate multivariate Gaussian: one specified vector of the two or more vectors.*)
var
  h, j: byte;
  density: real;
begin
density := 0;
for j := 1 to dim do (*Compute <v_i' Sigma v_i>, where <i = ldim> is some specified actor.*)
  begin
  for h := 1 to dim do
    begin
    density := density + (re[ldim,j] * isigma[j,h] * re[ldim,h]);
    end;
  end;
EvaluateMVNactor := - (density / 2); (*Evaluate multivariate normal probability density of candidate for the specified actor.*)
end;

(*--------------------------------------------------------------------------------------------------*)

function EvaluateW(var omega: pmatrix; var degrees: integer; var dim: byte; var sigma: pmatrix): real;
(*Evaluate Wishart kernel, which corresponds to the conjugate prior probability density of
the variance-covariance matrix of the random effects.*)
var
  det, u: real;
  cholfactor, outcome: pmatrix;
begin
(*First term:*)
Cholesky(dim,cholfactor,omega,det); (*<Omega> is the inverse (!) scale parameter of the Wishart density with <df = degrees>.*)
u := (degrees / 2) * ln(det);
(*Second term:*)
MatrixProd(dim,dim,outcome,omega,sigma); (*Product of the inverse scale matrix times the sample variance-covariance matrix.*)
u := u - MatrixTrace(dim,outcome) / 2; (*Function of the trace of the product calculated earlier.*)
(*Third term:*)
Cholesky(dim,cholfactor,sigma,det); (*<Sigma> is the sample variance-covariance matrix.*)
u := u + (((degrees - dim - 1) / 2) * ln(det));
EvaluateW := u;
end;

(*--------------------------------------------------------------------------------------------------*)

function RescaleCGD(const observed, desired, iterations, tolerance: integer; var number: byte; var success: boolean): real;
(*Rescale candidate-generating density to achieve desired acceptance rate:
method I in Browne's (1998) thesis <Applying MCMC methods to multi-level models>.*)
var
  u: real;
  tolerable: boolean;
begin
(*Comment:
<observed> = observed acceptance rate
<desired> = desired acceptance rate
<iterations> = number of iterations.*)
if (observed > desired) then
  begin
  u := 2 - ((iterations - observed) / (iterations - desired));
  RescaleCGD := u;
  end
else
  begin
  u := 2 - (observed / desired);
  RescaleCGD := 1 / u;
  end;
(*Comment:
<tolerance> = tolerance
<number> = count: number of times passed.*)
tolerable := (abs(observed - desired) <= tolerance);
if tolerable then
  begin
  Inc(number);
  if (number = 2) then success := true; (*Two successive times tolerable.*)
  end
else number := 0;
end;

(*--------------------------------------------------------------------------------------------------*)

function DecisionMH(var acceptancepr, real1, real2: real; var blackbox: blackarray): boolean;
(*Decide whether or not the candidate is to be accepted.*)
var
  u, temperature: real;
  accept: boolean;
begin
acceptancepr := real1 + real2; (*On natural logarithmic scale.*)
temperature := blackbox[20];
acceptancepr := acceptancepr * temperature; (*Modify the stationary distribution by temperatures > 1 to improve mixing: let the distribution melt (remember: importance sampling needed).*)
if (acceptancepr >= 0) then
  begin
  acceptancepr := 1; (*To avoid overflow...*)
  accept := true;
  end
else if (acceptancepr < -700) then acceptancepr := 0 (*...underflow.*)
else acceptancepr := exp(acceptancepr); (*Acceptance probability was calculated on natural logarithmic scale.*)
if (acceptancepr < 1) then
  begin
  u := random; (*Delphi/Pascal pseudo-random number generator to sample from U(0,1].*)
  if (acceptancepr >= u) then accept := true
  else accept := false;
  end;
DecisionMH := accept;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MCMC_Output_Files_I(const input, input1, input2, input3: shortstring);
(*Monte Carlo output procedures.*)
var
  batch, c1, c2, i, h, m, max_no_row_elements: integer;
  u: extended;
  condition: boolean;
  s: shortstring;
  ascii, g: textfile;
  mcmc: rrarray;
begin
s := CurrentProjectName + input + '.m'; (*Matlab file.*)
ReadFile(ascii,s);
mcmc := nil;
max_no_row_elements := 0;
c1 := 0;
while not(eof(ascii)) do
  begin
  inc(c1);
  SetLength(mcmc,c1+1,max_no_row_elements+1);
  c2 := 0;
  while not(eoln(ascii)) do
    begin
    inc(c2);
    read(ascii,u);
    if (c2 > max_no_row_elements) then
      begin
      max_no_row_elements := c2;
      SetLength(mcmc,c1+1,max_no_row_elements+1);
      end;
    mcmc[c1,c2] := u; (*Matrix to store full MCMC output.*)
    end;
  readln(ascii);
  end;
closefile(ascii);
OpenFile(ascii,s); (*Delete contents of Matlab file.*)
s := CurrentProjectName + input + '.g'; (*Gnuplot data file.*)
OpenFile(g,s);
m := 0;
if (c1  > 1000) then batch := trunc(c1 / 1000) (*A subsample is written to the Matlab file whose contents were deleted.*)
else batch := 1;
for i := 1 to c1 do (*Go from the first row to the last row.*)
  begin
  inc(m);
  if (m = batch) then (*Boolean: true = selected row, false = non-selected row.*)
    begin
    m := 0;
    condition := true;
    end
  else condition := false;
  write(g,' 1 '); (*Unit column vector is convenient in Gnuplot.*)
  for h := 1 to max_no_row_elements do (*In the given row, go from the first column to the last one.*)
    begin
    write(g,mcmc[i,h]:14:10); (*Gnuplot data file.*)
    if condition then write(ascii,mcmc[i,h]:14:10); (*Matlab file.*)
    end;
  writeln(g);
  if condition then writeln(ascii);
  end;
closefile(g);
closefile(ascii);
s := CurrentProjectName + input;
GenerateGnu(s,input1,input2,input3); (*Generate Gnuplot source code.*)
mcmc := nil;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MCMC_Output_Files_II(const input: shortstring; const no_columns: integer; const input1, input2, input3: shortstring);
(*Monte Carlo output procedures.*)
var
  batch, c1, c2, count, i, h, m, max_no_row_elements: integer;
  u: extended;
  condition: boolean;
  s: shortstring;
  ascii, g: textfile;
  mcmc: rrarray;
begin
s := CurrentProjectName + input + '.m'; (*Matlab file.*)
ReadFile(ascii,s);
mcmc := nil;
max_no_row_elements := 0;
c1 := 0;
while not(eof(ascii)) do
  begin
  inc(c1);
  SetLength(mcmc,c1+1,max_no_row_elements+1);
  c2 := 0;
  while (c2 < no_columns) do
    begin
    inc(c2);
    read(ascii,u);
    if (c2 > max_no_row_elements) then
      begin
      max_no_row_elements := c2;
      SetLength(mcmc,c1+1,max_no_row_elements+1);
      end;
    mcmc[c1,c2] := u; (*Matrix to store full MCMC output.*)
    end;
  readln(ascii);
  end;
closefile(ascii);
OpenFile(ascii,s); (*Delete contents of Matlab file.*)
s := CurrentProjectName + input + '.g'; (*Gnuplot data file.*)
OpenFile(g,s);
m := 0;
if (c1  > 1000) then batch := trunc(c1 / 1000) (*A subsample is written to the Matlab file whose contents were deleted.*)
else batch := 1;
for i := 1 to c1 do (*Go from the first row to the last row.*)
  begin
  inc(m);
  count := 0;
  if (m = batch) then (*Boolean: true = selected row, false = non-selected row.*)
    begin
    m := 0;
    condition := true;
    end
  else condition := false;
  write(g,' 1 '); (*Unit column vector is convenient in Gnuplot.*)
  for h := 1 to no_columns do (*In the given row, go from the first column to the last one.*)
    begin
    (*Comment:
    ASCII files store up to 1024 characters on one line; strings exceeding 1024 characters are cut after the
    1024th character and continued on the following line; this leads to problems when reading the files later.*)
    if (count >= 1000) then
      begin
      writeln(g);
      if condition then writeln(ascii);
      count := 0;
      end
    else count := count + 10;
    write(g,mcmc[i,h]:10:6); (*Gnuplot data file.*)
    if condition then write(ascii,mcmc[i,h]:10:6); (*Matlab file.*)
    end;
  writeln(g);
  if condition then writeln(ascii);
  end;
closefile(g);
closefile(ascii);
s := CurrentProjectName + input;
GenerateGnu(s,input1,input2,input3); (*Generate Gnuplot source code.*)
mcmc := nil;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure Get_MCMC_Output(const Bayesian: boolean; const no_columns: integer);
(*Obtain (and transform) MCMC output.*)
var s, s1, s2, s3: shortstring;
begin
if Bayesian then
  begin
  s := 'Theta'; (*Parameter vectors (rate and non-rate parameter vector).*)
  s1 := char(39) + 'Trace plot of parameter estimates' + char(39);
  s2 := char(39) + 'iteration number' + char(39);
  s3 := char(39) + 'estimates' + char(39);
  MCMC_Output_Files_I(s,s1,s2,s3);
  end;
if randomeffects then
  begin
  s := 'RE'; (*Actor-dependent random effects.*)
  s1 := char(39) + 'Trace plot of actor-dependent random effects' + char(39);
  s2 := char(39) + 'iteration number' + char(39);
  s3 := char(39) + 'random effects' + char(39);
  MCMC_Output_Files_II(s,no_columns,s1,s2,s3);
  if Bayesian then
    begin
    s := 'CovRE'; (*Variance-covariance matrix of the actor-dependent random effects.*)
    s1 := char(39) + 'Trace plot of (co)variances' + char(39);
    s2 := char(39) + 'iteration number' + char(39);
    s3 := char(39) + '(co)variances' + char(39);
    MCMC_Output_Files_I(s,s1,s2,s3);
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatlabPlot(var ascii: textfile);
(*Adds some convenient Matlab code that enables columnwise plots.*)
begin
writeln(ascii,']');
writeln(ascii,'%j = 1'); (*The dimension of the columns.*)
writeln(ascii,'%for i = 1:j');
writeln(ascii,'%  y(i) = x(i,1)');
writeln(ascii,'%end');
writeln(ascii,'plot(x)');
MatlabPlotExtra(ascii);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatlabPlotExtra(var ascii: textfile);
(*Some Matlab code.*)
begin
writeln(ascii,'z = sort(x)');
writeln(ascii,'mean(x)');
writeln(ascii,'std(x)');
writeln(ascii,'%i = round(j * 0.025)');
writeln(ascii,'%z(i)');
writeln(ascii,'%i = round(j * 0.975)');
writeln(ascii,'%z(i)');
end;

(*--------------------------------------------------------------------------------------------------*)

procedure GenerateGnu(const input, input1, input2, input3: shortstring);
(*Generates Gnuplot source code to make simple trace plots and so on.*)
var
  s: shortstring;
  f: textfile;
begin
s := input + '.gnu';
OpenFile(f,s);
writeln(f,'#set terminal latex');
writeln(f,'#set output ' + input + '.fig');
writeln(f,'set nokey');
writeln(f,'set nogrid');
writeln(f,'set noautoscale');
writeln(f,'#no zeroaxis');
writeln(f,'#set xrange [xmin: xmax]');
writeln(f,'#set yrange [0: ymax]');
writeln(f,'#set title ' + input1);
writeln(f,'set xlabel ' + input2);
writeln(f,'set ylabel ' + input3);
writeln(f,'#set xtics autofreq');
writeln(f,'#set ytics autofreq');
writeln(f,'#set format x "%.1f"');
writeln(f,'#set format y "%.1f"');
writeln(f,'plot '+ char(39) + input + '.g' + char(39) + ' using 2 with lines');
closefile(f);
end;

(*--------------------------------------------------------------------------------------------------*)

end.

