                             (* S_Start.PAS *)

(* SIENA - Simulation Investigation for Empirical Network Analysis *)

(* TS: For signed graphs, the ones codes are the absolute values of non-missing
       non-zero codes, their sign indicating whether it is interpreted as -1 or +1. *)
(* TS: only structural 0 and 1 values allowed, no structural -1 values. *)

unit s_start;

(* This unit starts a SIENA project
   and writes it to the internally used files. *)

Interface

Uses SysUtils, slib, s_constants, s_matrices, rangen, eight, digraph, s_dat, s_base,
     dialogs, Forms, classes;

const Struct0 = 10; Struct1 = 11;
     (* datafile code values for structural zeros and structural ones *)
     CarryForward = 7; (* Datafile code values indicating the same value as in the previous wave;
                          special feature, used only if AllowCarryForward is set true in ReadWriteData.  *)
     namelength = 12;

Type
  Readgraph = object(ModelSpecification)
(* The following variables are used in view of the <multi> option.
   The *1 variables are the numbers in the first subproject;
   the *2 variables are those in the currently read subproject
   (which has index number thissubproject). *)
    multi, multiplus: boolean;                   (* multiple .in files *)
                           (* multiplus indicates that the .min file contains a line with a +,
                             which will add subproject dummies to the multiple project. . *)
    firstperiod, nxa2, nza2, nz1, nz2, numz1, numz2, numzc1, numzc2, nzc1,
       nzz1, nzz2, nzzc1, nzzc2, nzx1, nzx2, multitot, thissubproject: integer;
    firstobs: array of integer; (* for multi-group projects:
                                  firstobs[k] = first observation for subproject k *)

    procedure InitialStatistics(var outf,lf : TFileStream);
    (* This procedure gives some basic data description in the output file. *)
    (* TS 18-jun-05: transferred declaration of this procedure to this object *)
    procedure ReadWriteData(var fn      : shortstring;     (* name of .IN file                *)
                            nmx,totper  : integer;
                            lastfile    : boolean;      (* indicates that this is the last sub-project *)
                            var halterr : boolean;    (* error flag                      *)
                            var outf,lf : TFileStream);  (* names of .OUT and .LOG files    *)
    (* CS: This procedure handles the extended file protocol of the SIENA 2 versions.    *)

    procedure Preparations(pn1         : shortstring;
                           var halterr : boolean;
                           var outf,lf : TFileStream);
    (* reads data, writes internal data files, initialises parameters. *)

    procedure ChangeModel;
    (* Adapts model before the .mo file is written; only for use
       when running special versions of Siena01. *)

    procedure ConvertExoToStruct(firper: integer; var outf: TFileStream);
    (* Converts composition change to structural zeros *)

  end; (* Readgraph *)


implementation

uses siena_1;

var siena1files: boolean;  (* for compatibility with stocnet 1.4 *)
    fffmat: array[1..300,1..300] of boolean;

procedure Readgraph.InitialStatistics(var outf,lf : TFileStream);
  (* This procedure gives some basic data description in the output file,
     and calculates the initial value for the density parameter estimate. *)
var i,j,k, ma, mi  : integer;
      nmisd, sumup, sumdown, nconst, nup, ndown: cardinal;
      counts: array of array of cardinal;
      cmis: array of cardinal;
      anymis: boolean;
      x : string;
begin (* InitialStatistics *)
   Heading(1,outf,'Initial data description.');
   if (observations >= 2) then
     Heading(2,outf,'Change in networks:')
    else
     Heading(2,outf,'Tie and dyad counts:');
   InitialAlphas(alpa_l,alpa_f,1,outf);
   //writeln(outf);
   outf.Write(pchar(eol)^,length(eol));
   if (nza > 0)
   then begin
     Heading(2,outf,'Dependent actor variables:');
     if (nza = 1) then
       begin
       x := format('%s',[ZNames[nz+1]])+eol;
       outf.Write(pchar(x)^,length(x));
      // writeln(outf);
       outf.Write(pchar(eol)^,length(eol));
       end;
     if (nza = 1) then Heading(3,outf,'Marginal distribution')
       else Heading(3,outf,'Marginal distributions');
     for i:=1 to nza do
       begin
       mi := round(zcmin[i]);
       ma := round(zcmax[i]);
       SetLength(counts,observations + 1, ma - mi + 2);
       SetLength(cmis,observations+1);
       for k:=1 to observations do
         begin
         n := nni[k];
         for j:= 0 to ma-mi do counts[k,j] := 0;
         cmis[k] := 0;
         for j := 1 to n do if (not miszc(j,i,k)) then
           inc(counts[k,round(zc[j,i,k])-mi])
          else inc(cmis[k]);
         end;
       if (nza >= 2) then begin
       x := format('%s%d%s%s',['Dependent actor variable ',i,': ',ZNames[nz+i]])+eol;
       outf.Write(pchar(x)^,length(x));
       end;
       //writeln(outf);
       outf.Write(pchar(eol)^,length(eol));
       for k := 1 to observations do outf.write(pchar(string('   '))^,3);
       x := '       Observations        '+eol;
       outf.Write(pchar(x)^,length(x));
       x :='values          ';
       outf.Write(pchar(x)^,length(x));
       for k:=1 to observations do begin
        x := format('%s%5d',[' ',k]);
        outf.Write(pchar(x)^,length(x));
       end;
       //writeln(outf);
       outf.Write(pchar(eol)^,length(eol));
       x := '                ';
       outf.Write(pchar(x)^,length(x));
       for k:=1 to observations do outf.write(pchar(string('------'))^,6);
       //writeln(outf);
       outf.Write(pchar(eol)^,length(eol));
       for j:=mi to ma do
         begin
         x := format('%3d%s',[j,'             ']);
         outf.Write(pchar(x)^,length(x));
         for k:=1 to observations do begin
          x:= format('%s%5d',[' ',counts[k,j-mi]]);
         outf.Write(pchar(x)^,length(x));
         end;
         //writeln(outf);
         outf.Write(pchar(eol)^,length(eol));
         end;
       anymis := false;
       for k:=1 to observations do if (cmis[k] > 0) then anymis := true;
       if anymis then
         begin
         x := 'missing         ';
         outf.Write(pchar(x)^,length(x));
         for k:=1 to observations do begin
          x := format('%s%5d',[' ',cmis[k]]);
          outf.Write(pchar(x)^,length(x));
         end;
         //writeln(outf);
         //writeln(outf);
         outf.Write(pchar(eol)^,length(eol));
         outf.Write(pchar(eol)^,length(eol));
         x := '(fraction missing ';
         outf.Write(pchar(x)^,length(x));
         for k:=1 to observations do begin
          n := nni[k];
          x := format('%4.2f%s',[cmis[k]/n,'  ']);
          outf.Write(pchar(x)^,length(x));
         end;
         x := ')'+eol;
         outf.Write(pchar(x)^,length(x));
         end
        else begin
          x := 'No missings'+eol;
          outf.Write(pchar(x)^,length(x));
        end;
        if (i < nza) then outf.Write(pchar(eol)^,length(eol));//writeln(outf);
       end;
     //writeln(outf);
     outf.Write(pchar(eol)^,length(eol));
     if (observations >= 2) then
       begin
       Heading(3,outf,'Changes');
       for i:=1 to nza
       do begin
         if (nza >= 2) then begin
          x := format('%s%d%s%s',['Dependent actor variable ',i,': ',ZNames[nz+i]])+eol;
          outf.Write(pchar(x)^,length(x));
         end;
         x := ' periods    actors:  down   up   constant  missing  ;';
         outf.Write(pchar(x)^,length(x));
         x := '   steps:   down    up  total'+eol;
         outf.Write(pchar(x)^,length(x));
         for k:=1 to observations-1
         do if (not skipi[k]) then
           begin
           n := nni[k];
           x := format('%3d%s%3d%s',[k,'  =>',k+1,'        ']);
           outf.Write(pchar(x)^,length(x));
           sumup := 0; sumdown := 0; nconst := 0; nup  := 0; ndown := 0;
//         for j:=1 to n
//         do if ((not (miszc(j,i,k) or miszc(j,i,k+1))) and (zc[j,i,k] > zc[j,i,k+1]))
//         then sum := sum + round(zc[j,i,k]) - round(zc[j,i,k+1]);
//         write(outf,round(sum):6);
//         sum:=0.0;
           nmisd := 0;
           for j:=1 to n
           do if (not (miszc(j,i,k) or miszc(j,i,k+1))) then begin
             case sgn(round(zc[j,i,k+1] - round(zc[j,i,k]))) of
               -1: begin
                   sumdown := sumdown + round(zc[j,i,k]) - round(zc[j,i,k+1]);
                   inc(ndown);
                   end;
                0: inc(nconst);
                1: begin
                   sumup := sumup + round(zc[j,i,k+1]) - round(zc[j,i,k]);
                   inc(nup);
                   end;

              end; (* case *)
           end
          else
           begin
           inc(nmisd);
           end;
           x := format('%6d%6d%8d%10d%s%9d%6d%6d',[ndown,nup,nconst,nmisd,'            ',
                        sumdown,sumup,ObservedActionDistance[i,k]])+eol;
           outf.Write(pchar(x)^,length(x));
          end (* not skipi *)
         else
          begin
          wwriteln(outf,'');
          end;
        end; (* if observations >= 2 *)
       if (i < nza) then outf.Write(pchar(eol)^,length(eol));//writeln(outf);
     end; (* i *)
     //writeln(outf);
     outf.Write(pchar(eol)^,length(eol));
     if (nza >= 2)
     then begin
      x := 'For these variables,';
      outf.Write(pchar(x)^,length(x));
     end
     else begin
      x := 'For this variable,';
      outf.Write(pchar(x)^,length(x));
     end;
     x := ' the standard initial behavioral tendency parameter';
     outf.Write(pchar(x)^,length(x));
     if (nza = 1)
     then begin
      x := format('%s%8.4f%s',[' is ',alpa_f[NetworkEffects_f + 1],' .'])+eol;
      outf.Write(pchar(x)^,length(x));
     end
     else begin
       x := 's are'+eol;
       outf.Write(pchar(x)^,length(x));
       for i:=1 to nza
       do  begin
        x := format('%d%s%s%8.4f',[i,': ',AlignString(UpperCase(ZNames[nz+i]),namelength),
                  alpa_f[NetworkEffects_f + (i-1)*ActionEffects_f + 1]])+eol;
        outf.Write(pchar(x)^,length(x));
       end;
     end;
     //writeln(outf)
     outf.Write(pchar(eol)^,length(eol));
   end; (* if nza > 0 *)

  if (observations = 1) then begin
    x := 'These standard initial parameters are calculated'+eol;
    outf.Write(pchar(x)^,length(x));
    if ExistStruct[1] then begin
      x := 'using only dyads with complete and non-structurally determined data.'+eol;
      outf.Write(pchar(x)^,length(x));
    end
      else begin
        x := 'using only dyads with complete data.'+eol;
        outf.Write(pchar(x)^,length(x));
      end;
    x := 'The parameters are:'+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%8.4f',['density      ',alpa_f[1]])+eol;
    outf.Write(pchar(x)^,length(x));
    if ((not symmetric) or asymmetric) then begin
    x := format('%s%8.4f',['reciprocity  ',alpa_f[2]])+eol;
    outf.Write(pchar(x)^,length(x));
    end;
    x := 'and further zero values.' + eol;
    outf.Write(pchar(x)^,length(x));
   end;

  if (observations = 1) then
    begin
     // writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      Heading(2,outf,'Clique analysis');
      describecliques(outf,0,4,20);
    end;

end; (* InitialStatistics *)

procedure Readgraph.Preparations(pn1         : shortstring;
                                 var halterr : boolean;
                                 var outf,lf : TFileStream);
var minfn, infn, proname, inname : shortstring;
  atf, lastfile  : Boolean;
  Correspondence  : SBYarray; (* for 'pro forma' use only *)
  vertpar         : BBarray;  (* for 'pro forma' use only *)
  param           : SSIArray; (* for 'pro forma' use only *)
  obs1, n1, obstot, nmaxx   : integer;
  fm,f            : textfile;
  x               : string;
begin (* Preparations *)
(* results of Preparations after ReadWriteData
   is written to model specification files *)
  halterr := false;
  initialise;
  if (pn1 > '') then proname := pn1;
  atf := false;
  AttachProjectName(pn1,outf,lf,atf,true);
  x := format('%s%s',['SIENA version ',SienaVersion])+eol;
  outf.Write(pchar(x)^,length(x));
  //writeln(lf);
  lf.Write(pchar(eol)^,length(eol));
  x := 'SIENA01'+eol;
  lf.Write(pchar(x)^,length(x));
  //writeln(lf);
  lf.Write(pchar(eol)^,length(eol));
  Heading(1,lf,'Preparations');

  (* Delete potentially incompatible earlier files with same project name: *)
  DeleteFile(pname+'.MO');
  DeleteFile(pname+'.SI');

  minfn := pname+'.min';
  multi := fileexists(minfn);
  if not multi then infn := pname+'.in';

  thisperiod := 0;
  multiplus := false;
  if multi then
    begin
    wwriteln(outf,'');
    x := 'Multi-group input file detected: <'+minfn+'> .'+eol;
    outf.Write(pchar(x)^,length(x));
    assignfile(fm,minfn);
    multitot := 0;
    obstot := 0;
    nmaxx := 0;
    reset(fm);
    multiplus := false;
    while not eof(fm) do
      begin
      readln(fm,inname);
      if (inname = '+') then multiplus := true;
      infn := parselo(inname,45);
      if (infn > '') then
        begin
        infn := infn+'.in';
        if fileexists(infn) then
          begin
          inc(multitot);
          wwriteln(outf,'Subproject '+inttostr(multitot)+': <'+infn+'> ');
          assignfile(f,infn);
          reset(f);
          readln(f);
          readln(f,obs1);  (* number of waves  *)
          readln(f,n1);             (* number of actors *)
          obstot := obstot + obs1;
          if (nmaxx < n1) then nmaxx := n1;
          closefile(f);
          end
         else
          begin
          x := 'Warning: There is no .in file for subproject <'+infn+'> .'+eol;
          outf.Write(pchar(x)^,length(x));
          end;
        end;
      end; (* while *)
    x := 'Multi-group command file '+minfn+' contains '
          +inttostr(multitot)+' subprojects.'+eol;
    outf.Write(pchar(x)^,length(x));
    outf.Write(pchar(eol)^,length(eol));
    if (multitot <= 1) then multiplus := false;
    setlength(firstobs,multitot+1);
    thissubproject := 0;
    reset(fm);
    while not eof(fm) do
      begin
      readln(fm,inname);
      infn := parselo(inname,45);
      if (inname <> infn) and (infn > '') then
        begin
        x := 'Subproject read as <'+inname+'>; this is converted to <'+infn+'>'+eol;
        outf.Write(pchar(x)^,length(x));
        end;
      if (infn > '') then
        begin
        infn := infn+'.in';
        if fileexists(infn) then
          begin
          inc(thissubproject);
          Heading(1,outf,'Subproject '+inttostr(thissubproject)+': <'+infn+'> ');   // changed 3.17v
          lastfile := (thissubproject >= multitot);
          ReadWriteData(infn,nmaxx,obstot,lastfile,halterr,outf,lf);
//            ReadWriteData(infn,32,5,lastfile,halterr,outf,lf);
          end;
        end;
      end; (* while *)
    if (thissubproject = 0) then begin
      Halterror('No valid subprojects in multi-group information file.',outf,lf);
      halterr := true;
      exit;
      end;
    end (* if multi *)
   else
    begin
    infn := pname+'.in';
    if fileexists(infn) then
      begin
      assignfile(f,infn);
      reset(f);
      readln(f);
      readln(f,obs1);  (* number of waves  *)
      readln(f,n1);             (* number of actors *)
      closefile(f);
      obstot := obs1;
      end
     else
      begin
      x := 'Warning: There is no .in file for subproject <'+infn+'> .'+eol;
      outf.Write(pchar(x)^,length(x));
      halterror('No .in file',outfile,lf);
      halterr := true;
      exit;
      end;
    ReadWriteData(infn,n1,obs1,true,halterr,outf,lf);
    end;
  if halterr then exit;
  ChangeDefaults;
  if halterr then exit;
  (* check for consistency and set other defaults like basic rate parameters: *)
  InitForm.MessageLabel.caption := 'initial model preparation';
  Application.ProcessMessages;
{$if defined(batch)}
  Writeln(Initform.MessageLabel.caption);
{$ifend}
  InitialiseBase1(outf,lf);
 if (observations = 1) and symmetric then Gibbs := 16; 
 ReviseDefinition(halterr,outf,lf);
  (* Define effect labels according to the latest model revision: *)
  DefineModel_lnames(labels_l,Correspondence,Param);
  Definemodel_fnames(labels_f,Correspondence,VertPar,Param);

  InitForm.MessageLabel.caption := 'basic data description';
  Application.ProcessMessages;
{$if defined(batch)}
  Writeln(Initform.MessageLabel.caption);
{$ifend}

(* AA start *)
  Checkmaxima(halterr,outf,lf);
  if halterr then exit;
(* AA end *)

 observations := obstot;
 InitialStatistics(outf,lf);
  (* Define statistics labels: *)
  DefineFunctionNames(lf);
  (* Write model definition file: *)
   InitForm.MessageLabel.caption := 'writing internal files';
  Application.ProcessMessages;
{$if defined(batch)}
       Writeln(Initform.MessageLabel.caption);
{$ifend}
  Heading(2,lf,'Write model definition files.');
  (* the WriteDefinition Procedures are located in S_Dat. *)
  ChangeModel;
  WriteDefinitionModel(lf);
  (* Write simulation definition file: *)


  WriteDefinitionStats(lf);
end; (* Preparations *)

procedure ReadGraph.ConvertExoToStruct(firper: integer; var outf: TFileStream);
var numbernonactive,h,i,j,k: integer;
  StructExists, Anystructure: boolean;
  structinactive: array of boolean;
  x: string;
begin
  if (not exo) then exit;
  Anystructure := false;
  SetLength(structinactive,nmax+1);
  for h:=1 to observations do
    begin
    n := nni[firper-1+h];
    StructExists := false;
    for i:=1 to n do
    if (h < zx[i,1]) or (zx[i,2] < h) then
      begin
      for j:=1 to n do if (j <>i) then
        begin
        setstruct(i,j,h+firper-1,true);
        setstruct(j,i,h+firper-1,true);
        end;
      StructExists := true;
      Anystructure := true;
      structinactive[i] := true;
      end (* for i if *) else structinactive[i] := false;
    if StructExists then
      begin
      x:= format('%s%d%s',['For observation ',h,', absent actors are treated as structural zeros.'])+eol;
      outf.Write(pchar(x)^,length(x));
      numbernonactive := 0;
      for i:=1 to n do if (structinactive[i]) then
        begin
        inc(numbernonactive);
        for k:=1 to nza do setmiszc(i,k,h+firper-1,true);
        end;
      if (numbernonactive = 1) then
        begin
        x := 'This led to 1 absent actor. This was actor '+eol;
        outf.Write(pchar(x)^,length(x));
        for i:=1 to n do if structinactive[i] then begin
          x:= format('%d%s',[i,'.'])+eol;
          outf.Write(pchar(x)^,length(x));
        end;
        end
       else
      if (numbernonactive >= 2) then
        begin
        x := format('%s%d%s',['This led to ',numbernonactive ,' absent actors.'])+eol;
        outf.Write(pchar(x)^,length(x));
        x:= 'These are actors';
        outf.Write(pchar(x)^,length(x));
        for i:=1 to n do if structinactive[i] then begin
          x:= format('%s%d',[' ',i]);
          outf.Write(pchar(x)^,length(x));
        end;
        x := '.'+eol;
        outf.Write(pchar(x)^,length(x));
        end;
      if (numbernonactive >= 1) then  begin
       x:= format('%s%d%s',
       ['The remaining number of active actors is ',n-numbernonactive,'.'])+eol;
       outf.Write(pchar(x)^,length(x));
      end;
     // writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      end;
    end; (* h *)
 // writeln(outf);
  outf.Write(pchar(eol)^,length(eol));
  exo := false;
  nzx := 0;
  if Anystructure then
    begin
    x:= '============================================================='+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'In this project, changing composition is treated '+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'in the form of structural zeros.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := '============================================================='+eol;
    outf.Write(pchar(x)^,length(x));
    end;
end; (* ConvertExoToStruct *)


procedure ReadGraph.ChangeModel;
    (* Adapts model before the .mo file is written; only for use
       when running special versions of Siena01. *)
begin
  effects_f[2] := true;
// voor andrea:
{effects_f[3] := true;
  effects_f[21] := true;
  effects_f[25] := true;
  effects_f[30] := true;
  effects_f[43] := true;
  effects_f[44] := true;
  effects_f[47] := true;
  effects_f[48] := true;
  effects_f[49] := true;
  effects_f[52] := true;
  effects_f[53] := true;
  effects_f[54] := true;
  effects_f[55] := true;
  effects_f[63] := true;
  effects_f[73] := true;
  effects_f[79] := true;
  effects_f[80] := true;
  effects_f[82] := true;
  effects_f[92] := true;
  effects_f[98] := true;
  effects_f[99] := true;
  effects_f[101] := true;
  effects_f[102] := true;
  effects_f[117] := true;
  effects_f[118] := true;}
//  effects_f[Networkeffects_f+2] := true;
//  effects_f[Networkeffects_f+11] := true;
  if (observations >= 2) then
    begin
//    maxlike := true; // special version
//    rstep0 := 15.0;  // special version
    firstgain := 0.2;
//    modelseed := 1234;
//    nrun3 := 2000;  // special version
//    InitialStandard := true; // special version
    end;
end; (* ChangeModel *)


procedure Readgraph.ReadWriteData(var fn: shortstring; nmx,totper  : integer;
                                  lastfile: boolean;
                                  var halterr: boolean; var outf,lf: TFileStream);
(*CS: This procedure handles the extended file protocol of the SIENA 2 versions. *)
const onemax = 10;
var f,df,fff,fm : textfile;
    dfn,fname,
    dumstring   : string;
    i,j,h,k,io,
    numz,
    numzc,nnz,
    nr,               (* CS: nr for section number in .IN files of SIENA 2 protocol *)
    dumint      : integer;
    vmin,vmax   : shortint;
    cmisz       : array[1..nzmax] of integer; (* CS: code for missing data on attributes *)
    dd, slength,
    tempinteger,
    t_exo       : integer; (* MH: t_exo for events *)
    texo        : real;
    zmin,zmax   : array of real;
    zzmin, zzmax: array of array of real;   (* for range checks on covariates *)
    zdum: iiarray;
    tempreal,
    f_exo       : real;    (* MH: f_exo for events *)
    errors, trypajek, tpajek     : boolean;
    CarryMissingForward, extra: boolean;
    AllowCarryForward, UseCarryForward: boolean; (* related to carryforward, not to missings *)
    anymissings, nothing, cf: boolean;
    freq: array of integer;
    fmax, jmax: integer;
    notfor, ert: shortstring;
    ReservedCodes: set of byte;
    zwarning: array of array of boolean;
    zwarnings: array of integer;
    x : string;
    dumnames: array of shortstring;
    missi: boolean;
    per, constvar: integer;

(* The following arrays are only present in unit S_Start because they are subtracted internally,
   reported, and then forgotten. *)
    zmean       : array[1..nzmax] of real;      (* means of constant covariates                *)
    nzmean      : array[1..nzmax] of integer;   (* # valid cases of constant covariates        *)
    zcmeani     : array[1..nzmax] of real;      (* global means of changing covariates         *)
    nzcmeani    : array[1..nzmax] of integer;   (* total # valid cases for changing covariates *)
    zcmaxi,                                     (* maximal value of behavioural variables      *)
    zcmini      : array[1..nzmax] of real;      (* minimal value of behavioural variables      *)
    zcmean      : array of array of real;       (* period means of changing covariates         *)
    nzcmean     : array of array of integer;    (* period # valid cases of changing covariates *)

    z_exo       : array of array of integer; (* MH: for exogenous events *)


(* Some local procedures for frequent sub-operations in ReadWrite Data *)
  procedure wri(a,i: integer);
  var x : string;
  begin (* wri *)
    if (n <= 999) then x := format('%3d',[a]) else x := format('%4d',[a]);
    outf.Write(pchar(x)^,length(x));
    if ((i mod 20) = 0) then outf.Write(pchar(eol)^,length(eol));
  end; (* wri *)

  procedure Startdfile(var outf,lf: TFileStream; var paj: boolean);
(* if paj is true on input,
   then startdfile checks if the extension name of dfn is 'net',
   and if so, returns paj = true and does not open the file with name dfn *)
   var x: string;
  begin
  (* Read data file *)
    if not FileExists(dfn)
    then begin
      if (dfn = '')
      then begin
        x := 'There is no data file name in the basic information file.'+eol;
        outf.Write(pchar(x)^,length(x));
      end
      else begin
      x := 'There is no file in the current directory with name '+dfn+'.'+eol;
      outf.Write(pchar(x)^,length(x));
      end;
      Halterror('Data file '+dfn+' is not found.',outf,lf);
      halterr := true;
      exit
    end;
   if paj then paj := UpperCase(ExtractFileExt(dfn)) = '.NET';
// 1234   writeln(outf,'extension = ',ExtractFileExt(dfn));
   if paj then exit;
  {$I-}
    assign(df,dfn);
    reset(df);
  {$I+}
    if (IOResult > 0)
    then begin
      x := format('%s%d',[ 'IOResult = ',IOResult])+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'Data file '+dfn+' cannot be opened.'+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('File '+dfn+' cannot be opened.',outf,lf);
      halterr := true;
      exit;
    end
    else begin
    x := 'Reading data file '+dfn+' '+eol;
    lf.Write(pchar(x)^,length(x));
    end;
  end; (* Startdfile *)

  procedure readonefile(h : integer; pass: integer);
(* reads network matrix file; h is the number of the file *)
(* for pass = 1, signed indicates whether there is at least one negative number in the file; *)
(* for pass = 2, the data in the file are read and processed. *)
  type codestype     = array[1..onemax] of integer;
  var i,j,
      nposcode, nposcodeplus,
      nmiscode,
      nmissing,numbernonactive,
      ierr,jerr      : integer;
      readingerror, pajek, oneone,
      StructExists   : boolean;
      ones, missings : codestype;
      structoutdegree,
      inpdegree, inmdegree,
      outpdegree, outmdegree      : array of integer;
      structinactive: array of boolean;
(* degrees of plus (p) or of minus (m) values *)
      NumberStruct0, NumberStruct1: integer; (* numbers of structural zeros and ones,
                                              not counting the diagonal           *)
    comps       : intvec;
    nc, ni, nis : integer;
    ert: shortstring;
    a,b: array of shortint;
(* AA change *)    sva: array of integer;
    x : string;

    function StructValue(x: shortint): shortint;
    begin
      Structvalue := x;
    end;

    function yes(a: integer): boolean;
    var i  : integer;
        yy : Boolean;
    begin (* yes *)
      i := nposcode; yy := false;
      repeat
        if (a = ones[i]) then yy := true;
        dec(i);
      until yy or (i <= 0);
      yes := yy;
    end; (* yes *)

    function missing(a: integer): boolean;
    var i  : integer;
        mm : Boolean;
    begin (* missing *)
      if (nmiscode = 0)
      then missing := false
      else begin
        i := nmiscode; mm := false;
        repeat
          if (a = missings[i]) then mm := true;
          dec(i);
        until mm or (i <= 0);
        missing := mm;
      end;
    end; (* missing *)

    function numberpart(a: string): boolean;
    begin
      numberpart := ((a = '-') or (a = '0') or (a = '1')
                    or (a = '2') or (a = '3') or (a = '4')
                    or (a = '5') or (a = '6') or (a = '7')
                    or (a = '8') or (a = '9') or (a = '0'));
    end;

    procedure readcodes(var codes: codestype; var ncode: integer; codekind: string);
    var code,
        codestring : string;
        k,i,lc     : integer;
        a1,a2      : string[1];
        x          : string;
    begin (* Read codeline  *)
      i := 0; errors := false;
      readln(f,codestring);
      lc := length(codestring);
      if (pass = 1) then begin
        x := 'For codes of kind '+codekind+', code string was read as <<'+
        codestring+'>> .'+eol;
         lf.Write(pchar(x)^,length(x));
         end;
      k := 0;
      code := '';
      repeat
      (* extract a code number *)
        repeat
          inc(k);
          a1 := copy(codestring,k,1);
        until (numberpart(a1) or (k >= lc));
        code := a1;
        repeat
          inc(k);
          a2 := copy(codestring,k,1);
          if numberpart(a2) then code := code+a2;
        until ((not numberpart(a2)) or (k >= lc));
        (* decode the number *)
        if (code > '')
        then if (i < onemax)
             then begin
               inc(i);
               val(code,codes[i],ierr);
               if (ierr <> 0) then dec(i)
                else
                begin
                if (codes[i] in ReservedCodes) then
                  begin
                  if (pass = 1) then
                    begin
                    //writeln(outf); writeln(outf);
                    outf.Write(pchar(eol)^,length(eol));
                    outf.Write(pchar(eol)^,length(eol));
                    x := '*** Warning ***'+eol;
                    outf.Write(pchar(x)^,length(x));
                    x := 'While reading the section in the basic information file '+pname+'.in for this data file,'+eol;
                    outf.Write(pchar(x)^,length(x));
                    x := format('%s%s%s%d%s',['among the codes of type <',codekind,'> the number ',
                    codes[i],' was read.'])+eol;
                    outf.Write(pchar(x)^,length(x));
                    x := 'This is a reserved code, and therefore not allowed.'+eol;
                    outf.Write(pchar(x)^,length(x));
                    x := 'See the SIENA manual.' +eol;
                    outf.Write(pchar(x)^,length(x));
                    x := 'Data input progresses as if this code was not encountered here.'+eol;
                    outf.Write(pchar(x)^,length(x));
                    //writeln(outf);
                    outf.Write(pchar(eol)^,length(eol));
                    end;
                  dec(i);
                  end;
                end;
             end
             else errors := true;
      until (k >= lc);
      if errors
      then begin
        x := format('%s%s%s%d%s',[ 'Too many codes of kind ',codekind,' in this file (maximum is onemax = ',
          onemax,').'])+eol;
        outf.Write(pchar(x)^,length(x));
        Halterror('Too many codes in .IN file.',outf,lf);
        halterr := true;
        exit
      end;
      ncode := i;
      if (pass = 1) then begin
        if (ncode = 1)
        then begin
          x := codekind+' code is ' ;
          outf.Write(pchar(x)^,length(x));
        end
        else begin
          x := codekind+' codes are ';
          outf.Write(pchar(x)^,length(x));
        end;
        for i:=1 to ncode
        do begin
          x := format('%d',[codes[i]]);
          outf.Write(pchar(x)^,length(x));
          if (i < ncode)
          then outf.write(pchar(string(','))^,1);
        end;
        if (ncode = 0) then begin
          x := '.... (none)';
          outf.Write(pchar(x)^,length(x));
        end;
       end; (* if *)
    end; (* readcodes *)

  begin (* readonefile *)
    SetLength(structoutdegree,n+1);
    SetLength(inpdegree,n+1);
    SetLength(inmdegree,n+1);
    SetLength(outpdegree,n+1);
    SetLength(outmdegree,n+1);
    SetLength(structinactive,n+1);
    StructExists := false;
    readingerror := false;

    (* Read file name *)
    readln(f,dfn);
    if (pass = 1) then begin
      x :=  'File '+dfn+' contains ';
      outf.Write(pchar(x)^,length(x));
      if (observations >= 2)
      then begin
        x := format('%s%d%s',['observation moment ',h,'.'])+eol;
        outf.Write(pchar(x)^,length(x));
      end
      else begin
        x := 'network data.'+eol;
        outf.Write(pchar(x)^,length(x));
      end;
     end; (* if *)
    (* Read code lines for nonzero values *)
    readcodes(ones,nposcode,'nonzero');
    if (nposcode <= 0)
    then begin
      nposcode := 1;
      ones[1] := 1; (* default dichotomization 1=1, rest=0 *)
      if (pass = 1) then begin
        x := 'No valid codes for present ties for file '+dfn+'.'+eol;
        outf.Write(pchar(x)^,length(x));
        x := '! Attention ! - Default dichotomization is used:'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'Positive code is 1.'+eol;
        outf.Write(pchar(x)^,length(x));
       end; (* if *)
    end; (* if nposcode<=0 *)
    (* Read code lines for missings *)
    if (pass = 1) then outf.write(pchar(string('; '))^,2);
    readcodes(missings,nmiscode,'missing');
    if (pass = 1) then outf.write(pchar('.'+eol)^,1+length(eol));
    if halterr then exit;
    for i:=0 to 9
    do if (yes(i) and missing(i) and (pass = 1))
       then begin
         x := format('%s%d%s',['Code ',i,' is given as a nonzero and also as a missing code.'])+eol;    (*[s]*)
         outf.Write(pchar(x)^,length(x));
         x := 'This is meaningless.'+eol;
         outf.Write(pchar(x)^,length(x));
         HaltError('Duplicate use of codes.',outf,lf);
         halterr := true;
         exit;
       end;
    (* Read data file *)
    pajek := true;
    Startdfile(outf,lf,pajek);
    if halterr then exit;
    if pajek then
      begin
      ert := '';
      if (pass = 1) then
        begin
        ReadPajekDataFile(dfn,ert,extra,h,valmin,valplus);
        x := 'Reading pajek format file '+dfn+'.'+eol;
        outf.Write(pchar(x)^,length(x));
        x := format('%s%d%s%d%s',['Tie variables range from ',valmin,' to ',valplus,'.'])+eol;
        outf.Write(pchar(x)^,length(x));
(* The data has not yet been recoded *)
        valplus := 1;
        if (valmin < 0) then valmin := -1;
        if (ert > '')  then
          begin
          x := 'Error when reading';
          outf.Write(pchar(x)^,length(x));
          if (observations >= 2) then begin
            x := format('%s%d%s',[', for observation ',h,', ']);
            outf.Write(pchar(x)^,length(x));
          end;
          x := ' Pajek network file '+dfn+'.'+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'Error message: '+ert+eol;
          outf.Write(pchar(x)^,length(x));
          Halterror('Error when reading Pajek input file.',outf,lf);
          halterr := true;
          exit;
          end;
        end (* pass = 1 *)
       else
        begin (* pass = 2 *)
(* Recode structurally determined values *)
        SetLength(a,2);
        SetLength(b,2);
        a[0] := Struct0;
        a[1] := Struct1;
        for i:=0 to 1 do b[i] :=StructValue(i);
        ChangeDataValues(a,b,1,h,1);
(* Recode tie values *)
        Setlength(a,nposcode+1);
        Setlength(b,nposcode+1);
        oneone := false;
        for i:=1 to nposcode do
          begin
          a[i-1] := ones[i];
          if (ones[i] = 1) then oneone := true;
          b[i-1] := 1;
          end;
        nposcodeplus := nposcode;
        if (not oneone) then
          begin
          a[nposcode] := 1;
          b[nposcode] := 0;
          nposcodeplus := nposcode+1;
          end
         else
          nposcodeplus := nposcode;
        ChangeDataValues(a,b,nposcodeplus,h,3);
(* Recode missing values *)
        if (nmiscode >= 1) then
          begin
          SetLength(a,nmiscode);
          for i:=1 to nmiscode do
            begin
            a[i-1] := missings[i];
            b[i-1] := 0;
            end;
          j := nmiscode-1;
          ChangeDataValues(a,b,j,h,2);
          end;
        for i:=1 to n do
          begin
          outpdegree[i] := PosOutdm(i,h);
          outmdegree[i] := NegOutdm(i,h);
          inpdegree[i]  := PosIndm(i,h);
          inmdegree[i]  := NegIndm(i,h);
          StructOutDegree[i] := StructRowElements(i,h);
          end;
(* Carry forward for pajek option *)
        if (nmiscode >= 1) then
          begin
          if CarryMissingForward and (h >= 2) then ImputePrevious(h);
(* This comes after registering the degrees, because the recorded degrees
   are what is actually observed. *)
          end;
        NumberStruct0 := 0;
        NumberStruct1 := 0;
        Setlength(sva,2);
        sva[0] := 0;
        sva[1] := 0;
        StructuralDataValues(h,sva);
        (* this presupposes structural tie variable values being only 0 and 1 *)
        NumberStruct0 := sva[0];
        NumberStruct1 := sva[1];
        StructExists := ((NumberStruct0 + NumberStruct1) > 0);
        end;  (* pass = 2 *)
      end (* Pajek *)
     else
      begin (* not Pajek *)
(* The pajek treatment utilizes new procedures, that also can be used
   to simplify the non-pajek treatment *)
      if halterr then exit;
      if (pass = 1) then
        begin
        for i:=1 to n do
          begin
          for j:=1 to n do
            begin
            read(df,dd);
            if (dd < 0) then signed := true;
            end; (* j *)
// readln(df); // !!! this is if you wish to use data files with superfluous stuff in the lines
          end; (* i *)
        closefile(df);
        if signed then
          begin
          valmin := -1; valplus := 1;
          end
         else
          begin
          valmin := 0; valplus := 1;
          end;
        end
       else (* pass = 2 *)
        begin
        ierr := 0; jerr := 0;
        for i:=1 to n do
          begin
          outpdegree[i] := 0;
          inpdegree[i] := 0;
          outmdegree[i] := 0;
          inmdegree[i] := 0;
          end;

(* TS: following parts about structural zeros copied to ConvertExoToStruct *)
        NumberStruct0 := 0;
        NumberStruct1 := 0;
        for i:=1 to n do
          begin
          structoutdegree[i] := 0;
          for j:=1 to n do
            begin
            read(df,dd);
            if ((i <> j) and yes(dd)) then
              begin
              if (dd >= 1) then dd := 1;
              if (dd <= -1) then dd := -1;
              setyvm(i,j,h,dd);
              if (dd >0) then
                begin
                Inc(outpdegree[i]); Inc(inpdegree[j]);
                end
               else
                if (dd <0) then
                  begin
                  Inc(outmdegree[i]); Inc(inmdegree[j]);
                  end;
              end (* if yes *)
             else
              begin
          (* check for structurally determined values *)
              if ((i <> j) and (dd = struct0)) then
                begin
                setstruct(i,j,h,true);
                inc(structoutdegree[i]);
                Inc(NumberStruct0);
                StructExists := true;
                end;
              if ((i <> j) and (dd = struct1)) then
                begin
                setstruct(i,j,h,true);
                inc(structoutdegree[i]);
                Inc(NumberStruct1);
                StructExists := true;
                setyvm(i,j,h,1);
                Inc(outpdegree[i]); Inc(inpdegree[j]);
                end;
              if ((i <> j) and (dd = CarryForward) and AllowCarryForward and (h >= 2)) then
                begin
(* Carry forward the previous value of this tie variable *)
                usecarryforward := true;
                setyvm(i,j,h,ynvm(i,j,h-1));
                if (ynvm(i,j,h-1) >0) then
                  begin
                  Inc(outpdegree[i]); Inc(inpdegree[j]);
                  end
                 else
                  if (ynvm(i,j,h-1) <0) then
                    begin
                    Inc(outmdegree[i]); Inc(inmdegree[j]);
                    end;
                end;
              if CarryMissingForward and (h >= 2) and missing(dd) then
                begin
                if (ynvm(i,j,h-1) <> 0) then setyvm(i,j,h,ynvm(i,j,h-1));
                end;
              end; (* if not yes *)
          (* Note that all diagonal elements remain "false" *)
            if (((IOResult > 0) or (eof(df) and (i< n)and (j < n))) and (ierr + jerr <= 0)) then
              begin
              readingerror := true;
              ierr := i; jerr := j
              end;
            end; (* for j *)
      (* skipped in version 1.831, not 1.84, again from 1.85 onward:   readln(df); *)
// readln(df); // !!! this is if you wish to use data files with superfluous stuff in the lines
          end; (* for i *)

        if readingerror then
          begin
          x := format('%s%d',['IOResult = ',IOResult])+eol;
          outf.Write(pchar(x)^,length(x));
          x:= format('%s%s%s%d%s%d%s',[ 'Problems with reading file ',dfn,' starting at element (',
                 ierr,', ',jerr,') .'])+eol;
          outf.Write(pchar(x)^,length(x));
          x := format('%s%d%s',
            ['(This may have to do with the number of actors, given as ',n,'.)'])+eol;
            outf.Write(pchar(x)^,length(x));
          Halterror('Problems reading file '+dfn+'.',outf,lf);
          halterr := true;
          exit;
          end;
{$I+}
        end; (* pass = 2 *)
      end; (* not pajek *)

    if (pass = 2) then
      begin

      x := 'For file '+dfn+', degree distributions are as follows:'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'Nodes'+eol;
      outf.Write(pchar(x)^,length(x));
      for i:=1 to n do wri(i,i);   outf.Write(pchar(eol)^,length(eol)); //        writeln(outf);
      if signed then
        begin
        x := 'out-degrees positive'+eol;
        outf.Write(pchar(x)^,length(x));
        for i:=1 to n do wri(outpdegree[i],i); outf.Write(pchar(eol)^,length(eol));// writeln(outf);
        x := 'out-degrees negative'+eol;
        outf.Write(pchar(x)^,length(x));
        for i:=1 to n do wri(outmdegree[i],i); outf.Write(pchar(eol)^,length(eol));//writeln(outf);
        x := 'in-degrees positive'+eol;
        outf.Write(pchar(x)^,length(x));
        for i:=1 to n do wri(inpdegree[i],i);  outf.Write(pchar(eol)^,length(eol));//writeln(outf);
        x := 'in-degrees negative'+eol;
        outf.Write(pchar(x)^,length(x));
        for i:=1 to n do wri(inmdegree[i],i);  outf.Write(pchar(eol)^,length(eol));//writeln(outf);
        end
       else
        begin
        x := 'out-degrees '+eol;
        outf.Write(pchar(x)^,length(x));
        for i:=1 to n do wri(outpdegree[i],i); outf.Write(pchar(eol)^,length(eol));//writeln(outf);
        x := 'in-degrees '+eol;
        outf.Write(pchar(x)^,length(x));
        for i:=1 to n do wri(inpdegree[i],i); outf.Write(pchar(eol)^,length(eol));// writeln(outf);
        end;

      if StructExists then
        begin
        //writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
        x :='The input file contains codes for structurally determined values:'+eol;
        outf.Write(pchar(x)^,length(x));
        x := format('%s%4d%s',['  ',NumberStruct0,' structural zero']);
        outf.Write(pchar(x)^,length(x));
        if (NumberStruct0 >= 2) then outf.write(pchar(string('s were'))^,6)
         else outf.write(pchar(string(' was'))^,4);
        x := format('%s%d%s',[' found (code ',struct0,').'])+eol;
        outf.Write(pchar(x)^,length(x));
        x := format('%s%4d%s',['  ',NumberStruct1,' structural one']);
        outf.Write(pchar(x)^,length(x));
        if (NumberStruct1 >= 2) then outf.write(pchar(string('s were'))^,6)
         else outf.write(pchar(string(' was'))^,4);
        x := format('%s%d%s',[' found (code ',struct1,').'])+eol;
        outf.Write(pchar(x)^,length(x));
        numbernonactive := 0;
        for i:=1 to n do if (structoutdegree[i] = n-1) then
          begin
          inc(numbernonactive);
          structinactive[i] := true;
          end else structinactive[i] := false;
        if (numbernonactive = 1) then
          begin
          x := 'Actor ';
          outf.Write(pchar(x)^,length(x));
          for i:=1 to n do if structinactive[i] then begin
            x:= format('%d',[i]);
            outf.Write(pchar(x)^,length(x));
          end ;
          if (observations = 1) then begin
            x := ' is inactive.'+eol;
            outf.Write(pchar(x)^,length(x));
            end
           else begin
             x := ' is inactive at this observation.'+eol;
             outf.Write(pchar(x)^,length(x));
            end;
          end
         else
        if (numbernonactive >= 2) then
          begin
          x := 'Actors ';
          outf.Write(pchar(x)^,length(x));
          j := 0;
          for i:=1 to n do if structinactive[i] then begin
            x := format('%d%s',[i,', ']);
            outf.Write(pchar(x)^,length(x));
            inc(j);
            if (j mod 10) = 0 then wwriteln(outf,'');
          end ;
          if (observations = 1) then  begin
            x :=' are inactive.'+eol;
            outf.Write(pchar(x)^,length(x));
            end
           else begin
            x := ' are inactive at this observation.'+eol;
            outf.Write(pchar(x)^,length(x));
            end;
          end;

        SetLength(comps,n+1);
        Component(nc,true,comps,h);
        Setlength(ComponentNumber,nc+1,n+1);
        Component(nc,false,comps,h);
        if (nc >= 1) then
          begin
          x := 'The structurally free (non-determined) values separate'+eol;
          outf.Write(pchar(x)^,length(x));
          x := format('%s%d%s',['the actors into ',nc,' component']);
          outf.Write(pchar(x)^,length(x));
          if (nc = 1) then begin
            x := '.'+eol;
            outf.Write(pchar(x)^,length(x));
            end
           else begin
            x := 's:'+eol;
            outf.Write(pchar(x)^,length(x));
           end;
          if (nc >= 2) then
            begin
            x := 'component       actors'+eol;
            outf.Write(pchar(x)^,length(x));
            i := 1;
            nis := 0;
            repeat
              x := format('%6d%s',[i,'        ']);
              outf.Write(pchar(x)^,length(x));
              ni := 0;
              for j:=1 to n do if (comps[j]=i) then
                begin
                inc(ni);
                inc(nis);
                x := format('%s%5d',[' ',j]);  
                outf.Write(pchar(x)^,length(x));
                if (ni mod 10 = 0) then
                  begin
                 // writeln(outf);
                  outf.Write(pchar(eol)^,length(eol));
                  x :=  '              ';
                  outf.Write(pchar(x)^,length(x));
                  end;
                end;
              if (ni = 1) then begin
              x := ' (inactive)'+eol;
              outf.Write(pchar(x)^,length(x));
              end
               else outf.Write(pchar(eol)^,length(eol));//writeln(outf);
              inc(i);
             until ((nis >= n) or (i >= n));
            end; (* if nc >= 2 *)
          end; (* if nc >= 1 *)

        //writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
       end; (* if StructExists  *)

(* Identify missing data  *)
      if pajek then
        begin
        nmissing := TotalMissings(h);
        for i:=1 to n do
          begin
          outpdegree[i] := MissingOut(i,h);
          inpdegree[i]  := MissingIn(i,h);
          SetMis(i,i,h,true);
(* In this way, the missings at the diagonal are not counted. *)
          end;
        end (* if pajek *)
       else
(*Also here the pajek treatment utilizes new procedures, that also can be used
   to simplify the non-pajek treatment *)
        begin
        reset(df);
        for i:=1 to n do
          begin
          outpdegree[i] := 0;
          inpdegree[i] := 0
          end;
        nmissing := 0;
        for i:=1 to n do
          begin
          for j:=1 to n do if (i = j) then
            begin
            read(df,dd);
            setmis(i,j,h,true);
            end
           else
            begin
            read(df,dd);
            if missing(dd) then
              begin
              setmis(i,j,h,true);
              inc(nmissing);
              Inc(outpdegree[i]);
              Inc(inpdegree[j])
              end;(* if missing *)
            end; (* for j if (i <>j) *)
        (* skipped in version 1.831, not 1.84, again from 1.85 onward:   readln(df); *)
          end; (* for i *)
        closefile(df);
        end; (* if not pajek *)

      if (nmissing = 0)
      then begin
        if (observations = 1)
        then begin
        x := 'No missing data.'+eol;
        outf.Write(pchar(x)^,length(x));
        end
        else begin
        x := format('%s%d%s%s%s',['No missing data for observation ',h,' (file ',dfn,').'])+eol;
        outf.Write(pchar(x)^,length(x));
        end;
       // writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
      end
      else begin
        anymissings := true;
        //writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
        x := 'For data file '+dfn+', number of missing values are '+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'Nodes  '+eol;
        outf.Write(pchar(x)^,length(x));
        for i:=1 to n do wri(i,i);        outf.Write(pchar(eol)^,length(eol));//    writeln(outf);
        x := 'missing in rows    '+eol;
        outf.Write(pchar(x)^,length(x));
        for i:=1 to n do wri(outpdegree[i],i); outf.Write(pchar(eol)^,length(eol));//writeln(outf);
        x := 'missing in columns '+eol;
        outf.Write(pchar(x)^,length(x));
        for i:=1 to n do wri(inpdegree[i],i);  outf.Write(pchar(eol)^,length(eol));//writeln(outf);
        x := format('%s%d%s%6.3f%s',['total number of missing data: ',nmissing,
                     ', corresponding to a fraction of ',nmissing/(n*(n-1.0)),'.'])+eol;
        outf.Write(pchar(x)^,length(x));
        if (h >= 2) then begin
          x := 'In reported in- and outdegrees, missings are not counted.'+eol;
        outf.Write(pchar(x)^,length(x));
        end;
        //writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
      end (* if nmissing > 0 *)
    end; (* if pass = 2 *)
  end; (* readonefile *)


  procedure nextsection(var n: integer);
  (* CS: searches for next section of .IN file in SIENA 2 format *)
  var ch   : char;
      code : integer;
      x : string;
  begin
    repeat
      read(f,ch);
      if ch='['
      then repeat
        read(f,ch)
      until ch=']'
    until ((ch='@') or eof(f));
    if ch='@'
    then begin
      readln(f,ch);
      val(ch,n,code);
    end
    else begin
      x := 'A SIENA 2 section cannot be read from basic information file.'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'Basic information file does not have SIENA 2 format.'+eol;
      lf.Write(pchar(x)^,length(x));
      halterr := true
    end
  end; (* nextsection *)


begin (* ReadWriteData *)
  AllowCarryForward := false;
  ReservedCodes := [Struct0, Struct1];
  if AllowCarryForward then ReservedCodes := ReservedCodes + [CarryForward];
  halterr := false;
  signed := false;
  Heading(1,outf,'Data input.');

  (* First read basic information file. *)
  x := 'Basic information file.'+eol;
  lf.Write(pchar(x)^,length(x));
  if not FileExists(fn)
  then begin
    x := 'Basic information file with name <<'+fn+'>> cannot be opened.'+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('No basic information file.',outf,lf);
    halterr := true;
    exit;
  end;
  assignfile(f,fn);
  reset(f);
 // writeln(lf);
  lf.Write(pchar(eol)^,length(eol));
  x := 'Read basic information file '+fn+'.'+eol;
  lf.Write(pchar(x)^,length(x));
  x := 'Read basic information file '+fn+'.'+eol;
  outf.Write(pchar(x)^,length(x));
{$I-}
  nextsection(nr);
  if halterr
  then begin
    x := 'This file does not have SIENA 2 format.'+eol;
    outf.Write(pchar(x)^,length(x));
    CloseFile(f);
    exit
  end;
  if nr<>1
  then begin
    x := format('%s%d%s',['Unexpected section encountered: @',nr,' instead of @1 (general information).'])+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Unexpected section.',outf,lf);
    halterr := true;
    exit
  end;
  readln(f,observations);  (* number of waves                                             *)
  readln(f,n);             (* number of actors                                            *)
  readln(f,nxa2);           (* number of dependent network variables                       *)
  readln(f,nza2);           (* number of dependent actor variables                         *)
  readln(f,numz2);          (* number of files containing constant actor covariates        *)
  readln(f,numzc2);         (* number of exogenous changing actor covariates               *)
  readln(f,nzz2);           (* number of constant dyadic covariates                        *)
  readln(f,nzzc2);          (* number of exogenous changing dyadic covariates              *)
  readln(f,nzx2);           (* option nzx for presence and treatment of composition change *)

  if (thisperiod > 0) then
    begin
    x := '';
    if (nxa2 <> nxa) then
      begin
      x :=
        'Error: subproject file '+fn+' deviates in number of dependent network variables ('
          +inttostr(nxa2)+' versus '+inttostr(nxa)+').'+eol;
      halterr := true;
      end;
    if (nza2 <> nza) then
      begin
      x :=
        'Error: subproject file '+fn+' deviates in number of dependent actor variables ('
          +inttostr(nza2)+' versus '+inttostr(nza)+').'+eol;
      halterr := true;
      end;
    if (numz2 <> numz1) then
      begin
      x :=
        'Error: subproject file '+fn+' deviates in number of constant actor covariate files ('
          +inttostr(numz2)+' versus '+inttostr(numz1)+').'+eol;
      halterr := true;
      end;
    if (numzc2 <> numzc1) then
      begin
      x :=
        'Error: subproject file '+fn+' deviates in number of changing actor covariates ('
          +inttostr(numzc2)+' versus '+inttostr(numzc1)+').'+eol;
      halterr := true;
      end;
    if (nzz2 <> nzz1) then
      begin
      x :=
        'Error: subproject file '+fn+' deviates in number of constant dyadic covariates ('
          +inttostr(nzz2)+' versus '+inttostr(nzz1)+').'+eol;
      halterr := true;
      end;
    if (nzzc2 <> nzzc1) then
      begin
      x :=
        'Error: subproject file '+fn+' deviates in number of changing dyadic covariates ('
          +inttostr(nzzc2)+' versus '+inttostr(nzzc1)+').'+eol;
      halterr := true;
      end;
    if (nzx2 <> nzx1) then
      begin
      x :=
        'Error: subproject file '+fn+' deviates in changing composition directive ('
          +inttostr(nzx2)+' versus '+inttostr(nzx1)+').'+eol;
      halterr := true;
      end;
      if halterr then
        begin
        outf.Write(pchar(x)^,length(x));
        Halterror('Misspecification of subproject file '+fn,outf,lf);
        exit;
        end;
    end (* if thisperiod > 0 *)
   else
    begin
    nxa  := nxa2;
    nza  := nza2;
    numz1 := numz2;
    numzc := numzc2;
    numzc1 := numzc2;
    nzz1 := nzz2;
    nzzc1 := nzzc2;
    nzx1 := nzx2;
    nzx  := nzx2;
    end;  (* if thisperiod = 0 *)
  if multi then
    begin
    numz := 0;
    numzc := numz1 + numzc1;
    end
   else
    begin
    numz := numz1;
    numzc := numzc1;
    end;

{$I+}
  if (IOResult > 0)
  then begin
    x := format('%s%d',[ 'IOResult = ', IOResult])+eol;
    outf.Write(pchar(x)^,length(x));
    x :=
      'Error in reading basic model codes (section @1 of '+fn+' file).'+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Misspecification of basic model codes',outf,lf);
    halterr := true;
    exit;
  end;

  (* write protocol to .log-file *)
  x := format('%d%s',[ observations,' observation']);
  lf.Write(pchar(x)^,length(x));
  if (observations = 1) then begin
    X := ',' + eol;
    lf.Write(pchar(x)^,length(x));
  end
   else begin
    x := 's,'+eol;
    lf.Write(pchar(x)^,length(x));
   end;
  x := format('%d%s',[ n,' actors,'])+eol;
  lf.Write(pchar(x)^,length(x));
  x := format('%d%s',[nxa,' dependent network variable']);
  lf.Write(pchar(x)^,length(x));
  if (nxa = 1) then begin
    x := ','+eol;
    lf.Write(pchar(x)^,length(x));
  end
  else begin
    x := 's,'+eol;
    lf.Write(pchar(x)^,length(x));
  end;
  x := format('%d%s',[ nza,' dependent actor variable']);
  lf.Write(pchar(x)^,length(x));
  if (nza = 1) then begin
    x := ','+eol;
    lf.Write(pchar(x)^,length(x));
  end
   else begin
   x := 's,'+eol;
   lf.Write(pchar(x)^,length(x));
   end;
  x := format('%d%s',[ numz1,' file']);
  lf.Write(pchar(x)^,length(x));
  if (numz1 = 1) then lf.write(pchar(string(' '))^,1) else lf.write(pchar(string('s '))^,2);
  x := 'with constant actor covariates,'+eol;
  lf.Write(pchar(x)^,length(x));
  x := format('%d%s',[ numzc1,' exogenous changing actor covariate']);
  lf.Write(pchar(x)^,length(x));
  if (numzc1 = 1) then begin
    x := ','+eol;
    lf.Write(pchar(x)^,length(x));
  end
   else begin
    x := 's,'+eol;
    lf.Write(pchar(x)^,length(x));
   end;
  x := format('%d%s',[ nzz1,' constant dyadic covariate']);
  lf.Write(pchar(x)^,length(x));
  if (nzz1 = 1) then begin
    x := ','+eol;
    lf.Write(pchar(x)^,length(x));
  end
   else begin
    x := 's,'+eol;
    lf.Write(pchar(x)^,length(x));
   end;
  x := format('%d%s',[nzzc1,' exogenous changing dyadic covariate']);
  lf.Write(pchar(x)^,length(x));
  if (nzzc1 = 1) then begin
    x := ','+eol;
    lf.Write(pchar(x)^,length(x));
  end
   else begin
    x :=  's,' +eol;
   lf.Write(pchar(x)^,length(x));
   end;

  (* write same protocol to .out-file *)
  x := format('%d%s',[ observations,' observation']);
  outf.Write(pchar(x)^,length(x));
  if (observations = 1) then begin
    x := ','+eol;
    outf.Write(pchar(x)^,length(x));
  end
  else begin
    x := 's,'+eol;
    outf.Write(pchar(x)^,length(x));
  end;
  x := format('%d%s',[ n,' actors,'])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%d%s',[ nxa,' dependent network variable']);
  outf.Write(pchar(x)^,length(x));
  if (nxa = 1) then begin  //writeln(outf,',') else writeln(outf,'s,');
    x := ','+eol;
    outf.Write(pchar(x)^,length(x));
  end
  else begin
    x := 's,'+eol;
    outf.Write(pchar(x)^,length(x));
  end;
  x := format('%d%s',[nza, ' dependent actor variable']);
  outf.Write(pchar(x)^,length(x));
  if (nza = 1) then begin //writeln(outf,',') else writeln(outf,'s,');
    x := ','+eol;
    outf.Write(pchar(x)^,length(x));
  end
  else begin
    x := 's,'+eol;
    outf.Write(pchar(x)^,length(x));
  end;
  x := format('%d%s',[numz1,' file']);
  outf.Write(pchar(x)^,length(x));
  if (numz1 = 1) then outf.write(pchar(string(' '))^,1) else outf.write(pchar(string('s '))^,2);
  x := 'with constant actor covariates,'+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%d%s',[numzc1,' exogenous changing actor covariate']);
  outf.Write(pchar(x)^,length(x));
  if (numzc1 = 1) then begin //writeln(outf,',') else writeln(outf,'s,');
   x := ','+eol;
    outf.Write(pchar(x)^,length(x));
  end
  else begin
    x := 's,'+eol;
    outf.Write(pchar(x)^,length(x));
  end;
  x := format('%d%s',[ nzz1,' constant dyadic covariate']);
  outf.Write(pchar(x)^,length(x));
  if (nzz1 = 1) then begin //writeln(outf, ',') else writeln(outf, 's,');
   x := ','+eol;
    outf.Write(pchar(x)^,length(x));
  end
  else begin
    x := 's,'+eol;
    outf.Write(pchar(x)^,length(x));
  end;
  x := format('%d%s',[ nzzc1,' exogenous changing dyadic covariate']);
  outf.Write(pchar(x)^,length(x));
  if (nzzc1 = 1) then begin //writeln(outf, ',') else writeln(outf, 's,');
   x := ','+eol;
    outf.Write(pchar(x)^,length(x));
  end
  else begin
    x := 's,'+eol;
    outf.Write(pchar(x)^,length(x));
  end;
  if (observations = 1) then Continuous := true;

  if (observations<0) or (nxa<0) or (nza<0) or (numz1<0) or (numzc1<0) or (nzz1<0) or (nzzc1<0)
  then begin
    x := 'One of these numbers is negative, this is not allowed.'+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Misspecification of basic model codes',outf,lf);
    halterr := true;
    exit;
  end;

  if (numz1 + nzz1 > 0) and multi then
    begin
    x := 'For multi-group projects, constant covariates are treated as changing covariates.'+eol;
    outf.Write(pchar(x)^,length(x));
    if (nzz1 > 0) then
      begin
      x := 'Note that missings in changing dyadic covariates are not (yet) supported!'+eol;
      outf.Write(pchar(x)^,length(x));
      end;
    end;

  {
  if (observations = 1) and (nza >= 1) then
    begin
    writeln(outf,'Having dependent behavior variables for ERG modeling is not implemented.');
    Halterror('Misspecification of basic model codes',outf,lf);
    halterr := true;
    exit;
  end;
  }

  (* MH: write exogenous events information in log and outfile *)
  (* If nzx=2 then joiners get value 0 and leavers are treated as missing,
     if nzx=3 then joiners and leavers are treated as missing, but
     contrary to regular missings they are only partly present in the
     simulation runs: only before/after the time they left/joined. *)
  (* TS: added 07-02-2006:
     option nzx=4 treats absent actors as structural zeros *)
  if (thisperiod = 0) then
  begin   
    if (nzx in [1,2,3]) and multi then
      begin
      x :=  'Joiners/leavers option '+inttostr(nzx)+
            ' is superseded for multi-group Siena by option 4.'+eol;
      outf.Write(pchar(x)^,length(x));
      lf.Write(pchar(x)^,length(x));
      nzx := 4;
      end;
  if (nzx >= 1)
  then begin
    exo := true;
    CConditional := false; (* unconditional estimation for networks of changing composition *)
    x := '1 file with times of composition change';
     lf.Write(pchar(x)^,length(x));
    x :=  '1 file with times of composition change';
    outf.Write(pchar(x)^,length(x));

//    nzx := 4; // supersede: special version

    if (nzx = 1)
    then begin
      x :=  '.'+eol;
      lf.Write(pchar(x)^,length(x));
      x := '.'+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    if (nzx > 1)
    then begin
      x := ':';
      lf.Write(pchar(x)^,length(x));
      x := format('%s%d%s',['  joiners/leavers option ',nzx,'.'])+eol;
      lf.Write(pchar(x)^,length(x));
      x := ':';
      outf.Write(pchar(x)^,length(x));
      x := format('%s%d%s',['  joiners/leavers option ',nzx,'.'])+eol;
      outf.Write(pchar(x)^,length(x));
      //writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      x := '***************************************************************************'+eol;
      outf.Write(pchar(x)^,length(x));
      x := '  Attention !!'+eol;
      outf.Write(pchar(x)^,length(x));
      x := format('%s%d%s',['  Option ',nzx,' for composition change was chosen.'])+eol;
      if (nzx <= 3) then begin
        x := '  In this specification, the values of the joiners before joining'+eol;
        outf.Write(pchar(x)^,length(x));
        if (nzx = 2)
        then begin
          x := '  are replaced by the value 0 (no ties), and the values of the'+eol;
          outf.Write(pchar(x)^,length(x));
          x := '  leavers after leaving are treated as regular missing data.'+eol;
          outf.Write(pchar(x)^,length(x));
        end;
        if (nzx = 3)
        then begin
          x := '  and the values of the leavers after leaving are treated as regular'+eol;
          outf.Write(pchar(x)^,length(x));
          x := '  missing data.'+eol;
          outf.Write(pchar(x)^,length(x));
        end;
        x := '  Missing values are not used in the estimation procedure, but during'+eol;
        outf.Write(pchar(x)^,length(x));
        x := '  simulation the missing actors are allowed to make changes, thereby'+eol;
        outf.Write(pchar(x)^,length(x));
        x := '  influencing the evolutionary process. Contrary to regular missings,'+eol;
        outf.Write(pchar(x)^,length(x));
        x := '  joiners and leavers are only partly present: only after/before they'+eol;
        outf.Write(pchar(x)^,length(x));
        x := '  joined/left, respectively.'+eol;
        outf.Write(pchar(x)^,length(x));
       end (* if nzx <= 3 *) else if (nzx = 4) then begin
        x := '  Before joining and after leaving, actors are treated as structural zeros.'+eol;
        outf.Write(pchar(x)^,length(x));
       end; (* if nzx = 4 *)
        x := '***************************************************************************'+eol;
        outf.Write(pchar(x)^,length(x));
    end;
  end
  else begin
    exo := false;
    x := 'no file with times of composition change.'+eol;
    lf.Write(pchar(x)^,length(x));
    x := 'no file with times of composition change.'+eol;
    outf.Write(pchar(x)^,length(x));
  end;
  end; (* if thisperiod = 0 *)

  //writeln(lf);
  lf.Write(pchar(eol)^,length(eol));
  //writeln(outf);
  outf.Write(pchar(eol)^,length(eol));

  (* now some tests of data format compatibility with the capabilities of SIENA: *)

  (* number of waves *)
  if (thisperiod + observations >= 100)
  then begin
    x := format('%d%s',[ thisperiod + observations ,' observations is not implemented in ReadPDefinition.'])+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Too many observations.',outf,lf);
    halterr := true;
  end;
  (* number of actors *)
  if (n <= 2)
  then begin
    x := format('%s%d%s',[ 'Only ',n,' actors; this makes no sense - try another tool for data analysis.'])+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Not enough actors.',outf,lf);
    halterr := true;
  end;
  (* number of network variables *)
  if (nxa = 0)
  then begin
    x := 'no dependent network variable; this makes no sense - try another tool for data analysis.'+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('No network variable.',outf,lf);
    halterr := true;
  end;
  if (nxa > 1)
  then begin
    x := format('%d%s%s%s',[nxa,' dependent network variables; SIENA ',SienaVersion,
      ' can only analyze one.'])+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Too many network variables.',outf,lf);
    halterr := true;
  end;
  (* number of exogenous changing actor covariates *)
{  if (observations <= 2) and (numzc1 >= 1)
  then begin
    x := 'Error reading basic information file:'+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%d%s%d%s',['Only ',observations,' observations, but ',numzc1,
      ' exogenous changing actor covariates.'])+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Exogenous changing covariates are meaningless for less than 3 observation moments.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Change this data specification to constant actor covariates.'+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Error in basic information file.',outf,lf);
    halterr := true;
    exit
  end;
}
  (* total number of dyadic covariates *)
  if (nzz1+nzzc1 > nzzmax)
  then begin
    x := 'The input file specifies too many dyadic covariates:'+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%d%s%d%s',['The numbers are indicated as nzz1 = ',nzz1,' and nzzc1 = ',nzzc1,','])+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%d%s',[ 'but the maximum total number is nzzmax = ',nzzmax,'.'])+eol;
    outf.Write(pchar(x)^,length(x));
    x :=  'Change your input file or recompile SIENA'+eol;
    outf.Write(pchar(x)^,length(x));
    x :=  'with nzzmax in S_Dat.PAS at a higher value.'+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Too many dyadic covariates',outf,lf);
    halterr := true;
    exit;
  end;
  (* number of exogenous changing dyadic covariates *)
{
  if (observations <= 2) and (nzzc1 >=1)
  then begin
    x := 'Error reading basic information file:'+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%d%s%d%s',['Only ',observations,' observations, but ',nzzc1,
      ' exogenous changing dyadic covariates.'])+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Exogenous changing covariates are meaningless for less than 3 observation moments.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Change this data specification to constant dyadic covariates.'+eol;
    outf.Write(pchar(x)^,length(x));
    Halterror('Error in basic information file.',outf,lf);
    halterr := true;
    exit
  end;
}
  (* MH: exogenous events file not used if observations <= 1 *)
  if ((observations <= 1) and exo)
  then begin
    x := format('%s%d%s',['With only ',observations,' observation']);
    outf.Write(pchar(x)^,length(x));
    if (observations = 1) then outf.write(pchar(string(', '))^,2) else outf.write(pchar(string('s, '))^,3);
    x := 'the file with times of composition change is ignored.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Increase the number of observations in order to use this file.'+eol;
    outf.Write(pchar(x)^,length(x));
    exo:=false;
  end;

  if halterr then exit;
  CarryMissingForward := (observations >= 2);
  nzc1 := numzc1+nza;   (* the code understands nza to be AMONG the nzc changing actor cov's *)


  if (thisperiod = 0) then
    begin
    nmax := n;
    SetNumberOfPeriods(nmx,nzc1,nzz1,nzzc1,totper);
    end
   else
    begin
    if (nmax < n) then nmax := n;
    end;
  firstperiod := thisperiod + 1;
 if multi then firstobs[thissubproject] := firstperiod;
  thisperiod := thisperiod + observations;
  for i:= firstperiod to thisperiod do
    begin
    nni[i] := n;
    skipi[i] := false;
    end;
  skipi[thisperiod] := true;

  (* READ NETWORK VARIABLES *)
  InitForm.MessageLabel.caption := 'reading network variables';
  Application.ProcessMessages;
{$if defined(batch)}
       Writeln(Initform.MessageLabel.caption);
{$ifend}
  Heading(2,lf,'Reading data.');
  Heading(2,outf,'Reading network variables.');
  nextsection(nr);
  if halterr then exit;
  if nr<>2
  then begin
    x := format('%s%d%s',['Unexpected section encountered: @',nr,' instead of @2 (network files).'])+eol;
    Halterror('Unexpected section.',outf,lf);
    halterr := true;
    exit
  end;
  for i := 1 to nxa
  do begin
    anymissings := false;
    x := 'Reading digraph file';
    outf.Write(pchar(x)^,length(x));
    if (observations = 1)
    then begin
      x := ' for the ';
      outf.Write(pchar(x)^,length(x));
    end
    else begin
      x := format('%s%d',['s for the ',i]);
      outf.Write(pchar(x)^,length(x));
      case i of
        1,21,31 : outf.write(pchar(string('st'))^,2);
        2,22,32 : outf.write(pchar(string('nd'))^,2);
        3,23,33 : outf.write(pchar(string('rd'))^,2);
        else outf.write(pchar(string('th'))^,2)
      end
    end;
    x := ' network variable:'+eol;
    outf.Write(pchar(x)^,length(x));
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    x := 'Network variable:'+eol;
    lf.Write(pchar(x)^,length(x));
    usecarryforward := false;
 (* First check if the files contain to signed or valued graph data: *)
    for k:=firstperiod to thisperiod
    do begin (* Digraph time k *)
      readonefile(k,1);
      if halterr then exit;
    end; (* for k *)
    if signed
    then begin
      x := 'The data files contain negative tie values'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'and are interpreted as signed graphs.'+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    if valued (* CS: currently, this is mutually exclusive with signed *)
    then begin
      x := 'The data files contain tie values larger than one'+eol;
      outf.Write(pchar(x)^,length(x));
      x:= 'and are interpreted as valued graphs.'+eol;
      outf.Write(pchar(x)^,length(x));
    end;
   //writeln(outf);
   outf.Write(pchar(eol)^,length(eol));

  (* Now really read and process the data. *)
    reset(f);
    nextsection(nr);  (* this is just to get past section 1 *)
    nextsection(nr);  (* this is to get at the start of section 2 *)
    if halterr then exit;
    if nr<>2
    then begin
      x := format('%s%d%s',['At second pass, unexpected section encountered: @',nr,' instead of @2 (network files).'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Unexpected section.',outf,lf);
      halterr := true;
      exit
     end;
    for k:=firstperiod to thisperiod
    do begin (* Digraph time k *)
      if (nxa <= 1) then
        InitForm.MessageLabel.caption := 'reading network variables '+inttostr(k)
       else InitForm.MessageLabel.caption :=
         'reading network variables '+inttostr(i)+'.'+inttostr(k);
      Application.ProcessMessages;
{$if defined(batch)}
       Writeln(Initform.MessageLabel.caption);
{$ifend}
      readonefile(k,2);
      if halterr then exit;
    end; (* for k *)
(* CS: name of network variable not yet passed on anywhere else *)
    readln(f,dumstring); (* name of network variable i *)
    x := 'Name of this ';
    outf.Write(pchar(x)^,length(x));
    if (i >= 2) then begin
      x := format('%d',[i]);
      outf.Write(pchar(x)^,length(x));
      case i of
        1,21,31 : outf.write(pchar(string('st '))^,3);
        2,22,32 : outf.write(pchar(string('nd '))^,3);
        3,23,33 : outf.write(pchar(string('rd '))^,3);
        else outf.write(pchar(string('th '))^,3)
      end;
     end; (* if *)
    x := 'network variable: ';
    outf.Write(pchar(x)^,length(x));
    if dumstring=''
    then begin
      x := '....(none given).'+eol;
      outf.Write(pchar(x)^,length(x));
    end
    else begin
      x := dumstring+'.'+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    if usecarryforward then
      begin
      x := format('%s%d%s',['Note that the carry forward code ',carryforward, 'was encountered;'])+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'this is interpreted as carrying forward the previous observed value for this tie variable. '+eol;
      outf.Write(pchar(x)^,length(x));
      end;
    if anymissings then
      begin
      if (observations = 1) then
        begin
        x := 'There are missings in this data set.'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'The current implementation of SIENA does not allow'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'missing data for the ERGM (non-longitudinal) case.'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'Therefore, the analysis will proceed'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'as if the missings indicate absence of ties.'+eol;
        outf.Write(pchar(x)^,length(x));
        end;
      if CarryMissingForward then
        begin
        x := 'There are missing data for the network variable, '+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'and the <<carry missings forward>> option is active.'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'This means that for each tie variable,'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'the last previous nonmissing value (if any) is imputed.'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'If there is no previous nonmissing value, the value 0 is imputed.'+eol;
        outf.Write(pchar(x)^,length(x));
        end;
      end;
  end; (* for i *)

  
  (* initialise some variable-specific arrays: *)
  setlength(zcmean,nzc1+nza+1,observations+1);
  setlength(nzcmean,nzc1+nza+1,observations+1);

  (* READ DEPENDENT ACTOR VARIABLES *)
  if  (nza > 0)
  then begin
    Heading(2,outf,'Reading dependent actor variables.');
    InitForm.MessageLabel.caption := 'reading dependent actor variables';
    Application.ProcessMessages;
{$if defined(batch)}
       Writeln(Initform.MessageLabel.caption);
{$ifend}
    nextsection(nr);
    if halterr then exit;
    if nr<>3
    then begin
      x := format('%s%d%s',['Unexpected section encountered: @',nr,' instead of @3 (dependent actor variables).'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Unexpected section.',outf,lf);
      halterr := true;
      exit;
      //writeln(lf);
      lf.Write(pchar(eol)^,length(eol));
    end;
    for i:=1 to nza
    do begin
      if (firstperiod = 1) then zcmini[i] := 255.0;
      if (firstperiod = 1) then zcmaxi[i] := 0.0;
      for j:= 1 to observations
      do begin
        zcmean[i,j] := 0.0;
        nzcmean[i,j] :=0
      end
    end;
    (* read data files: *)
    for i := 1 to nza do begin
      x := format('%d',[i]);
      outf.Write(pchar(x)^,length(x));
      case i of
        1,21,31 : outf.write(pchar(string('st'))^,2);
        2,22,32 : outf.write(pchar(string('nd'))^,2);
        3,23,33 : outf.write(pchar(string('rd'))^,2);
        else outf.write(pchar(string('th'))^,2)
      end;
      x := ' dependent actor variable named ';
      outf.Write(pchar(x)^,length(x));
      readln(f,dfn);

      (* missing data identification: *)
      readln(f,dumstring);  (* row with code for missing data on this variable *)
      (* transform string with missing code into a number *)
      val(dumstring,cmisz[1],dumint);
      if dumint<>0
      then begin
        dumstring := copy(dumstring,1,dumint-1);
        val(dumstring,cmisz[1],dumint);
      end;
      if dumint<>0 then cmisz[1] := -9; (* in case no missing code is given, take this one *)

      readln(f,ZNames[i]);      (* CS: note that index i is not the 'final' index of the actor variable, *)
                                (*     constant actor covariates will be read that are inserted before!  *)
      slength := length(ZNames[i]);
      if slength=0
      then begin
        str(i,ZNames[i]);
        ZNames[i]:='behavior-'+ZNames[i]
      end;
      x := UpperCase(ZNames[i])+' read from file '+dfn+'.'+eol;
      outf.Write(pchar(x)^,length(x));

      if dumint=0
      then begin
        x := format('%s%d%s',['Value ',cmisz[1],' is the code for missing data.']) +eol;
        outf.Write(pchar(x)^,length(x));
      end
      else begin
        x := 'No code for missing data is specified; SIENA will treat scores of -9 as missing data.'+eol;
        outf.Write(pchar(x)^,length(x));
      end;

      trypajek := false; (* reading the next files does not have a special option
                            for extension name 'net'. *)

    (* read dependent actor variable file *)
      Startdfile(outf,lf,trypajek);
      anymissings := false;
      if halterr then exit;
      for j:=1 to n
      do begin
        for h:=firstperiod to thisperiod
        do begin
          read(df,zc[j,i,h]);
          setmiszc(j,i,h,(zc[j,i,h]=cmisz[1]));
          if not miszc(j,i,h) then
          begin
            if (zc[j,i,h] < zcmini[i]) then zcmini[i] := zc[j,i,h];
            if (zc[j,i,h] > zcmaxi[i]) then zcmaxi[i] := zc[j,i,h];
            zcmean[i,h-firstperiod+1] := zcmean[i,h-firstperiod+1] + zc[j,i,h];     // hier
            inc(nzcmean[i,h-firstperiod+1]) (* count on how many observations the mean calculation is based *)
          end
          else
            begin
            zc[j,i,h] := 0.0; (* for missing cases, set variable value to zero;
                                 superseded below *)
            anymissings := true;
            end;
          if (eof(df) and (j < n))
          then begin
           // writeln(outf);
            outf.Write(pchar(eol)^,length(eol));
            x := format('%s%s%s%d%s%d%s',['Unexpected end of file ',dfn,' after line ',j,', observation ',h,':'])+eol;
            outf.Write(pchar(x)^,length(x));
            x := 'either the number of actors is too small, or'+eol;
            outf.Write(pchar(x)^,length(x));
            x := format('%s%s%s%d%s',['the number of observations for each actor on ',ZNames[i],
                         ' does not equal ',observations, ','])+eol;
            outf.Write(pchar(x)^,length(x));
            x := format('%s%s%s%d%s',[ 'i.e., the number of columns in ',dfn,
              ' is not equal to ',observations,'.'])+eol;
            outf.Write(pchar(x)^,length(x));
            Halterror('Unexpected end of file '+dfn,outf,lf);
            halterr := true;
            exit;
          end
        end; (* h *)
        readln(df);
      end; (* for j *)
      closefile(df);


      x := format('%s%d%s%d%s',['Minimum and maximum rounded values are ',round(zcmini[i]),
        ' and ',round(zcmaxi[i]),'.'])+eol;
      outf.Write(pchar(x)^,length(x));
      if (zcmini[i] < 0)
      then begin
        x := 'Negative minima not allowed for dependent actor variables.'+eol;
        outf.Write(pchar(x)^,length(x));
        halterr := true;
      end;
      if (zcmaxi[i] > 255)
      then begin
        x := 'Maxima more than 255 not allowed for dependent actor variables.'+eol;
        outf.Write(pchar(x)^,length(x));
        halterr := true;
      end;
      if (zcmini[i] >= zcmaxi[i])
      then begin
        x := 'Dependent actor variables must not be constant.'+ eol;
        outf.Write(pchar(x)^,length(x));
        halterr := true;
      end;
      if halterr
      then begin
        Halterror('Incorrect dependent actor variable.',outf,lf);
        exit;
      end;

      if anymissings then
(* impute *)
        begin
        x :=
        'Missing values in this actor variable are imputed by the mode per observation.'+eol;
        outf.Write(pchar(x)^,length(x));
        if CarryMissingForward then
          begin
          x := 'But if there is a previous nonmissing value, ';
          outf.Write(pchar(x)^,length(x));
          x := 'this is used as the imputed value.'+eol;
          outf.Write(pchar(x)^,length(x));
          end;
        x := 'Modal values:'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'Observation  ';
        outf.Write(pchar(x)^,length(x));
        for h:=1 to observations do begin
          x := format('%4d',[h]);
          outf.Write(pchar(x)^,length(x));
        end;
        // writeln(outf);
         outf.Write(pchar(eol)^,length(eol));
        x := 'mode         ';
        outf.Write(pchar(x)^,length(x));
        SetLength(freq,round(zcmaxi[i]-zcmini[i])+1);
        for h:=firstperiod to thisperiod do
(* Determine periodwise modal value jmax *)
          begin
          for j:= round(zcmaxi[i]-zcmini[i]) downto 0 do freq[j] := 0;
          for j:=1 to n do if not miszc(j,i,h)
            then inc(freq[round(zc[j,i,h]-zcmini[i])]);
          fmax := freq[0];
          jmax := round(zcmini[i]);
          for j:= 1 to round(zcmaxi[i]-zcmini[i]) do
            begin
            if (freq[j] > fmax) then
              begin
              jmax := j+round(zcmini[i]);
              fmax := freq[j];
              end;
            end;
          x := format('%4d',[jmax]);
          outf.Write(pchar(x)^,length(x));
(* impute last observed or mode: *)
          for j:=1 to n do if miszc(j,i,h) then
            begin
            cf := CarryMissingForward and (h >= 2);
            if cf then
              begin
(* now let cf = (there is any nonmissing value previous to h) *)
              cf := false;
              for k:=h-1 downto firstperiod do
                if (not miszc(j,i,k)) then cf := true;  // 1 changed to firstperiod, 3.17u
              end;
            if cf then zc[j,i,h] := zc[j,i,h-1] else zc[j,i,h] := jmax;
            end; (* for j *)
          end; (* for h *)
        outf.Write(pchar(eol)^,length(eol));
        end; (* if anymissings *)
      outf.Write(pchar(eol)^,length(eol));
    end; (* for i =1 to nza *)

    x := format('%s%d%s',['A total of ',nza,' dependent actor variable']);
    outf.Write(pchar(x)^,length(x));
    if (nza = 1) then begin
      x := '.'+eol;
      outf.Write(pchar(x)^,length(x));
    end
     else begin
      x := 's.'+eol;
      outf.Write(pchar(x)^,length(x));
     end;
  //  writeln(outf);
    outf.Write(pchar(eol)^,length(eol));

    (* report missings *)
    x := 'Number of missing cases per observation:'+eol;
    outf.Write(pchar(x)^,length(x));
    x := ' observation ';
    outf.Write(pchar(x)^,length(x));
    for h:=firstperiod to thisperiod do begin
      x := format('%9d',[h]);
      outf.Write(pchar(x)^,length(x));
    end;
    x := '       overall'+eol;
    outf.Write(pchar(x)^,length(x));
    anymissings := false;
    nothing := false;
    for i:=1 to nza
    do begin
      x := AlignString(Znames[i],namelength+4);
      outf.Write(pchar(x)^,length(x));
      nzcmeani[i] := 0;
      for h:= 1 to observations
      do begin
        inc(nzcmeani[i],nzcmean[i,h]);
        x := format('%6d%s',[n-nzcmean[i,h],'   ']);
        outf.Write(pchar(x)^,length(x));
        if (nzcmean[i,h] <= 0) then
          begin
          nothing := true;
          notfor := Znames[i]+', observation '+inttostr(h);
          end;
      end;
      if ((n-nzcmeani[i]) >= 1) then anymissings := true;
      x := format('%7d%s%4.1f%s',[n*(observations)-nzcmeani[i],'      (',
                   (100*(n*(observations)-nzcmeani[i])/
                        (n*(observations))),' %)'])+eol;
      outf.Write(pchar(x)^,length(x));
    end; (* for i *)
   // writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    if nothing then
      begin
      x := 'There are no valid data for '+notfor+'.'+eol;
      outf.Write(pchar(x)^,length(x));
      halterr := true;
      Halterror('Not enough valid data.',outf,lf);
      exit;
      end;

    (* report averages *)
    x := 'Means per observation';
    outf.Write(pchar(x)^,length(x));
    if anymissings then begin
      x := ' (missings not counted):'+eol;
      outf.Write(pchar(x)^,length(x));
    end
    else  begin
      x := ':'+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    x := ' observation ';
    outf.Write(pchar(x)^,length(x));
    for h:=firstperiod to thisperiod do begin
      x := format('%9d',[h]);
      outf.Write(pchar(x)^,length(x));
    end;
    x := '       overall'+eol;
    outf.Write(pchar(x)^,length(x));
    for i:=1 to nza
    do begin
      x := AlignString(Znames[i],namelength+4);
      outf.Write(pchar(x)^,length(x));
      zcmeani[i] := 0.0;
      for h:= 1 to observations
      do begin
        zcmeani[i] := zcmeani[i] + zcmean[i,h];
        zcmean[i,h] := divi(zcmean[i,h],nzcmean[i,h]);
        if (nzcmean[i,h] >= 1) then begin
          x := format('%8.3f%s',[zcmean[i,h],' ']);
          outf.Write(pchar(x)^,length(x));
          end
              else begin
                x := '    --- ';
                outf.Write(pchar(x)^,length(x));
              end;;
      end;
      zcmeani[i] := divi(zcmeani[i],nzcmeani[i]);
        if (nzcmeani[i] >= 1) then begin
          x := format('%8.3f%s',[zcmeani[i],' ']);
          outf.Write(pchar(x)^,length(x));
        end
              else begin
                x := '    --- ';
                outf.Write(pchar(x)^,length(x));
              end;;
     // writeln(outf)
      outf.Write(pchar(eol)^,length(eol));
    end; (* for i *)
    //writeln(outf)
    outf.Write(pchar(eol)^,length(eol));
  end; (* if nza > 0 *)

  (* READ CONSTANT ACTOR COVARIATES *)
  if (numz1 > 0)
  then begin
    Heading(2,outf,'Reading constant actor covariates.');
    InitForm.MessageLabel.caption := 'reading constant actor covariates';
    Application.ProcessMessages;
{$if defined(batch)}
       Writeln(Initform.MessageLabel.caption);
{$ifend}
    nextsection(nr);
    if halterr then exit;
    if nr<>4
    then begin
      x := format('%s%d%s',['Unexpected section encountered: @',nr,' instead of @4 (constant actor covariates).'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Unexpected section.',outf,lf);
      halterr := true;
      //writeln(lf);
      lf.Write(pchar(eol)^,length(eol));
      exit
    end;
    (* initialise variable-specific arrays *)
    for i:=1 to nzmax
    do zmean[i] := 0.0;
    k := 0; (* will be the cumulative number of individual covariates, is
               used for determining nz = total number of individual covariates  *)
    (* read data files *)
    for i:=1 to numz1
    do begin
      x := 'Constant actor covariate:'+eol;
     lf.Write(pchar(x)^,length(x));
      readln(f,dfn); (* name of covariate data file *)

{CS: bypass StOCNET error with one-variable-covariate files: }

readln(f,dumstring);
if copy(dumstring,length(dumstring)-17,18)='[code for missing]' (* this happens when line with  #cov. is omitted *)
then begin
  x := 'StOCNET error bypassed when reading covariate data file named "'+dfn+'"'+eol;
  outf.Write(pchar(x)^,length(x));
  nnz:=1;
  nr:=-1 {code for error having occurred}
end
else begin
{TS: changed so that only the number part of dumstring is read}
  nnz := firstnumber(dumstring);
  if (nnz < 0) then
    begin
    x := 'Unexpected line encountered: '+dumstring+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'This data file is treated as containing 0 variables.'+eol;
    outf.Write(pchar(x)^,length(x));
    nnz := 0;
    end;
  nr:=0 {code for error not having occurred}
end;

{$I-}
//      readln(f,nnz); (* number of actor covariates in this file *)
{$I+}
{      if (IOResult <> 0)
      then begin
        writeln(outf,'Error reading basic information file:');
        writeln(outf,'number of variables in covariate data file named "',dfn,'"');
        writeln(outf,'is not an integer number.');
        Halterror('Error in basic information file.',outf,lf);
        halterr := true;
        exit;
      end;
}
      if (nnz > nzmax)
      then begin
        x := 'Basic information file says that '+eol;
        outf.Write(pchar(x)^,length(x));
        x := format('%s%s%s%d%s',['covariate data file ',dfn,' has ',nnz,' variables.'])+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'This is too many.'+eol;
        outf.Write(pchar(x)^,length(x));
        Halterror('Too many individual covariates.',outf,lf);
        halterr := true;
        exit;
      end;
      (* read covariate labels and missing-codes *)
      x := format('%s%s%s%d%s',['Covariate data file ',dfn,' with ',nnz,' variables, named:'])+eol;
      outf.Write(pchar(x)^,length(x));
      for j:=k+1 to k+nnz
      do begin
{CS: following IF-THEN-ELSE conditional only to bypass StOCNET error with one-variable-covariate
     files -- after that bug is fixed, replace it with what is written in the THEN row
     -- this reads the code for missing line (which already IS read when the StOCNET error occurs). }
        if (nr<>-1)
        then readln(f,dumstring)  (* row with code for missing data on this variable *)
        else nr := 0;

        (* missing data identification: *)
     {   readln(f,dumstring); ALREADY IS READ NOW.. } (* row with code for missing data on this variable *)
        (* transform string with missing code into a number *)
        val(dumstring,cmisz[j],dumint);
        if dumint<>0
        then begin
          dumstring := copy(dumstring,1,dumint-1);
          val(dumstring,cmisz[j],dumint);
        end;
        if dumint<>0 then cmisz[j] := -9; (* in case no missing code is given, take this one *)

        for h:=nza downto 1         (* shift labels of dependent actor variables one up, making *)
        do ZNames[j+h] := ZNames[j+h-1]; (* room for this new constant actor covariate's label. *)
        readln(f,ZNames[j]);
        slength := length(ZNames[j]);
        if slength=0
        then begin
          str(j,ZNames[j]);
          ZNames[j]:='attribute-'+ZNames[j]
        end;
//        for h:= slength + 1 to namelength do ZNames[j] := ZNames[j]+' ';
      end;
      for j:=k+1 to k+nnz
      do begin
        x := AlignString(UpperCase(ZNames[j]),namelength)+' (code for missing data is ';
        outf.Write(pchar(x)^,length(x));
        if (cmisz[j]=-9)
        then begin
          x := 'not given, SIENA will treat scores of -9 as missing data)';
          outf.Write(pchar(x)^,length(x));
        end
        else begin
        x := format('%d%s',[cmisz[j],')']);
        outf.Write(pchar(x)^,length(x));
        end;
        if (j<(k+nnz))
        then begin
          x := ','+eol ;
          outf.Write(pchar(x)^,length(x));
        end
        else begin
          x := '.'+eol;
          outf.Write(pchar(x)^,length(x));
        end;
      end;

      (* read the constant covariate file *)
      Startdfile(outf,lf,trypajek);
      if halterr then exit;
      for h := 1 to nnz
      do begin
        zmean[k+h] := 0.0;
        nzmean[k+h] := 0
      end;
      for j:=1 to n
      do begin
        for h:=1 to nnz
        do begin
          read(df,z[j,k+h]);
          setmisz(j,k+h,(z[j,k+h]=cmisz[k+h]));
          if not misz(j,k+h)
          then begin
            zmean[k+h] := zmean[k+h] + z[j,k+h];
            inc(nzmean[k+h]);
          end
          else z[j,k+h] := 0.0; (* for missings, define value as zero *)
          if (eof(df) and ((j < n) or (h < nnz)))
          then begin
           x := format('%s%s%s%d%s%d%s%d%s',['Unexpected end of file ',dfn,' after line ',j,', variable ',h,
                         ', each line giving ',nnz,' variables.'])+eol;
            outf.Write(pchar(x)^,length(x));
            Halterror('Unexpected end of file '+dfn,outf,lf);
            halterr := true;
            exit
          end (* if *)
        end; (* for h *)
        readln(df)
      end; (* for j *)
      closefile(df);

      k := k+nnz;
      if (k + nza > nzmax)
      then begin
        x := format('%s5d',['Cumulative number of individual covariates exceeds ',nzmax])+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'in file '+dfn+'.'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'Work with less individual covariates or recompile SIENA'+eol;
        outf.Write(pchar(x)^,length(x));
        x :=  'with nzmax in S_Dat.PAS at a higher value.'+eol;
        outf.Write(pchar(x)^,length(x));
        Halterror('Too many individual covariates',outf,lf);
        halterr := true;
        exit
      end
    end; (* for i  -- that is the file number count *)
    nz := k; (* remember cumulated number of constant actor covariates *)

    if multi then
      begin
      if (firstperiod = 1) then nz1 := nz
       else
        begin
        nz2 := nz;
        if (nz1 <> nz2) then
          begin
          x := 'Error: subproject file '+fn+' deviates in total number of actor covariates ('
          +inttostr(nz2)+' versus '+inttostr(nz1)+')'+eol;
          outf.Write(pchar(x)^,length(x));
          Halterror('Multi-group project error',outf,lf);
          halterr := true;
          exit
          end;
        end;
      end;

    if (nz > 0)
    then begin
     // writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      x := format('%s%d%s',['A total of ',nz,' non-changing individual covariate']);
      outf.Write(pchar(x)^,length(x));
      if (nz = 1) then begin
         x :='.'+eol;
        outf.Write(pchar(x)^,length(x));
        end
       else begin
       x := 's.'+eol;
       outf.Write(pchar(x)^,length(x));
       end;
      //writeln(outf);
      outf.Write(pchar(eol)^,length(eol));

      (* report missings *)
      x := 'Number of missing cases:'+eol;
      outf.Write(pchar(x)^,length(x));
      anymissings := false;
      for h:=1 to nz
      do begin
        x := format('%s%4d%s%4.1f%s',[AlignString(Znames[h],namelength+1),n-nzmean[h],
        '   (',100*(n-nzmean[h])/n,' %)'])+eol;
        outf.Write(pchar(x)^,length(x));
        if ((n - nzmean[h]) >= 1) then anymissings := true;
      end;

      Setlength(zmin,nz+1);
      Setlength(zmax,nz+1);
      (* range checks *)
      for h:=1 to nz
      do begin
        zmin[h]:=1E30; (* CS: assume that no user uses attributes with all scores higher than that.. *)
        zmax[h]:=-zmin[h];  (* ..or lower than that.                                                 *)
        for i:=1 to n
        do if not misz(i,h)
        then begin
          if zmin[h]>z[i,h] then zmin[h]:=z[i,h];
          if zmax[h]<z[i,h] then zmax[h]:=z[i,h];
        end;
        if (zmin[h]=zmax[h]) (* range check *)
        then begin
(* AA start *)
          poszvar[h] := false;
(* AA end *)
          //writeln(outf);
          outf.Write(pchar(eol)^,length(eol));
          x := 'Warning: Valid scores have range zero on variable '+UpperCase(Znames[h])+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'DO NOT INCLUDE THE MAIN EFFECT OF THIS NON-INFORMATIVE COVARIATE IN THE MODEL SPECIFICATION!'+eol;
          if not multi then outf.Write(pchar(x)^,length(x));
(* AA start *)
        end else poszvar[h] := true;
(* AA end *)
        if nzmean[h]=0
        then begin
          //writeln(outf);
          outf.Write(pchar(eol)^,length(eol));
          x := 'WARNING: No valid scores on variable '+UpperCase(Znames[h])+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'DO NOT INCLUDE THIS NON-INFORMATIVE COVARIATE IN THE DATA SPECIFICATION!'+eol;
          outf.Write(pchar(x)^,length(x));
        end
      end;
      //writeln(outf);
      outf.Write(pchar(eol)^,length(eol));

      (* report extrema and means *)
      x := 'Information about covariates';
      outf.Write(pchar(x)^,length(x));
      if anymissings
        then begin
          x := ' (missings not counted):'+eol;
          outf.Write(pchar(x)^,length(x));
        end
        else begin
          x := ':'+ eol;
          outf.Write(pchar(x)^,length(x));
        end;
      x := '             minimum  maximum     mean '+eol;
      outf.Write(pchar(x)^,length(x));
      for h:=1 to nz
      do begin
        if nzmean[h]<>0
        then begin
          zmean[h] := zmean[h]/nzmean[h];
          x := format('%s%6.1f%s%7.1f%s%10.3f%s',[AlignString(Znames[h],namelength+1),
                       zmin[h],' ',zmax[h],' ',zmean[h],' '])+eol;
          outf.Write(pchar(x)^,length(x));
        end
        else begin
          x := AlignString(Znames[h],namelength+4)+'--- '+eol;
          outf.Write(pchar(x)^,length(x));
        end;
      end;

    (* report averages and runtime-centering of constant actor covariates *)
      constvar := 0;
      for h:=1 to nz do
        if poszvar[h] then      // changed 3.17w
          begin
          for i:=1 to n do if not misz(i,h)
            then z[i,h] := z[i,h] - zmean[h]
            else z[i,h] := 0.0; (* missings get covariate score zero, also after subtraction *)
          end
         else          // changed 3.17w
          begin
          inc(constvar);
          wwriteln(outf,'Variable '+ZNames[h]+' is constant, and therefore is not centered.');
          end;
      outf.Write(pchar(eol)^,length(eol));
      if (nz - constvar > 0) then
        begin
        x := 'The mean value';
        if (nz - constvar >= 2) then x := x+'s';
        if (constvar >= 1) then
          begin
          x := x + ' of the other variable';
          if (nz - constvar >=2 ) then x := x+'s';
          end;
        if (nz - constvar >= 2) then x := x+' are '
         else x := x+' is ';
        x := x+'subtracted from the covariate';
        if (nz - constvar >= 2) then x := x+'s.'
         else x := x+'.';
        wwriteln(outf,x);
        wwriteln(outf,'');
        end;
    end (* if nz>0 *)

  end (* Read constant actor covariates *)
  else nz := 0; (* no constant actor covariate FILES -> no constant actor COVARIATES! *)

  (* READ CHANGING ACTOR COVARIATES *)
  if (numzc1 > 0)
  then begin
    Heading(2,outf,'Reading exogenous changing actor covariates.');
    InitForm.MessageLabel.caption := 'reading changing actor covariates';
    Application.ProcessMessages;
{$if defined(batch)}
       Writeln(Initform.MessageLabel.caption);
{$ifend}
    nextsection(nr);
    if halterr then exit;
    if nr<>5
    then begin
      x := format('%s%d%s',['Unexpected section encountered: @',nr,
                   ' instead of @5 (exogenous changing actor covariates).'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Unexpected section.',outf,lf);
      halterr := true;
      exit
    end;

    (* recall that field lengths were already set before reading the dependent actor variables *)
    (* initialize some fields for missings and averaging *)
    for i:=nza+1 to nzc1
    do for j:= 1 to observations
    do begin
      zcmean[i,j] := 0.0;
      nzcmean[i,j] :=0
    end;

    (* read data files *)
    for i:=1 to numzc1
    do begin
      x := 'Changing actor covariate:';
      lf.Write(pchar(x)^,length(x));
      readln(f,dfn);
      Startdfile(outf,lf,trypajek);
      if halterr then exit;
      if (i+nz+nza > nzmax)
      then begin
        x := format('%s%d',['Cumulative number of individual variables exceeds ',nzmax])+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'in file '+dfn+'.'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'Work with less individual variables or recompile SIENA'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'with nzmax in S_Dat.PAS at a higher value.'+eol;
        outf.Write(pchar(x)^,length(x));
        Halterror('Too many individual variables',outf,lf);
        halterr := true;
        exit
      end;

     x := 'Exogenous changing covariate named ';
      outf.Write(pchar(x)^,length(x));
      (* missing data identification: *)
      readln(f,dumstring);  (* row with code for missing data on this variable *)
      (* transform string with missing code into a number *)
      val(dumstring,cmisz[1],dumint);
      if dumint<>0
      then begin
        dumstring := copy(dumstring,1,dumint-1);
        val(dumstring,cmisz[1],dumint);
      end;
      if dumint<>0 then cmisz[1] := -9; (* in case no missing code is given, take this one *)
      readln(f,ZNames[nz+nza+i]);
      slength := length(ZNames[nz+nza+i]);
      if slength=0
      then begin
        str(nz+nza+i,ZNames[nz+nza+i]);
        ZNames[nz+nza+i]:='attribute-'+ZNames[nz+nza+i]
      end;
//      for h:= slength + 1 to namelength do ZNames[nz+nza+i] := ZNames[nz+nza+i]+' ';
      x := UpperCase(ZNames[nz+nza+i])+' read from file '+dfn+'.'+eol;
      outf.Write(pchar(x)^,length(x));
      if dumint=0
      then begin
        x := format('%s%d%s',['Value ',cmisz[1],' is the code for missing data.'])+eol;
        outf.Write(pchar(x)^,length(x));
      end
      else begin
        x := 'No code for missing data is specified; SIENA will treat scores of -9 as missing data.'+eol;
        outf.Write(pchar(x)^,length(x));
      end;
      anymissings := false; // added 3.17u
      zcmini[i+nza] := 255.0; // added 3.17u
      zcmaxi[i+nza] := 0.0; // added 3.17u

      (* now read data *)
      for j:=1 to n
      do begin
        for h:= firstperiod to thisperiod-1
        do begin
          read(df,zc[j,i+nza,h]);
          setmiszc(j,i+nza,h,(zc[j,i+nza,h]=cmisz[1]));
          if not miszc(j,i+nza,h) then
            begin
            if (zc[j,i+nza,h] < zcmini[i+nza]) then zcmini[i+nza] := zc[j,i+nza,h]; // added 3.17u
            if (zc[j,i+nza,h] > zcmaxi[i+nza]) then zcmaxi[i+nza] := zc[j,i+nza,h]; // added 3.17u
            zcmean[i+nza,h+1-firstperiod] := zcmean[i+nza,h+1-firstperiod] + zc[j,i+nza,h];
            inc(nzcmean[i+nza,h+1-firstperiod])
            end
          else
            begin
            zc[j,i+nza,h]:=0.0;
            anymissings := true;
            end;

          if (eof(df) and ((j < n) or (h < thisperiod-1)))
          then begin
            //writeln(outf);
            outf.Write(pchar(eol)^,length(eol));
            x:= format('%s%s%s%d%s%d%s',['Unexpected end of file ',dfn,' after line ',j,', observation ',h,':'])+eol;
            outf.Write(pchar(x)^,length(x));
            x := 'either the number of actors is too small, or'+eol;
            outf.Write(pchar(x)^,length(x));
            x := format('%s%s%s%d%s',['the number of observations for each actor on ',
            ZNames[nz+nza+i], ' does not equal ',observations-1, ','])+eol;
            outf.Write(pchar(x)^,length(x));
            x := format('%s%s%s%d%s',[ 'i.e., the number of columns in ',dfn,' is not equal to ',observations-1,'.'])+eol;
            outf.Write(pchar(x)^,length(x));
            Halterror('Unexpected end of file '+dfn,outf,lf);
            halterr := true;
            exit
          end;
        end; (* h *)
        readln(df) (* ignore potential extra columns in the file *)
      end; (* for j *)
      closefile(df);

      if anymissings then
(* impute: added 3.17u *)
        begin
        x :=
        'Missing values in this actor variable are imputed by the mode per observation.'+eol;
        outf.Write(pchar(x)^,length(x));
        if CarryMissingForward then
          begin
          x := 'But if there is a previous nonmissing value, ';
          outf.Write(pchar(x)^,length(x));
          x := 'this is used as the imputed value.'+eol;
          outf.Write(pchar(x)^,length(x));
          end;
        x := 'Modal values:'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'Observation  ';
        outf.Write(pchar(x)^,length(x));
        for h:=firstperiod to thisperiod-1 do begin
          x := format('%4d',[h]);
          outf.Write(pchar(x)^,length(x));
        end;
        // writeln(outf);
         outf.Write(pchar(eol)^,length(eol));
        x := 'mode         ';
        outf.Write(pchar(x)^,length(x));
        SetLength(freq,round(zcmaxi[i+nza]-zcmini[i+nza])+1);
        for h:=firstperiod to thisperiod-1 do
(* Determine periodwise modal value jmax *)
          begin
          for j:= round(zcmaxi[i+nza]-zcmini[i+nza]) downto 0 do freq[j] := 0;
          for j:=1 to n do if not miszc(j,i+nza,h)
            then inc(freq[round(zc[j,i+nza,h]-zcmini[i+nza])]);
          fmax := freq[0];
          jmax := round(zcmini[i+nza]);
          for j:= 1 to round(zcmaxi[i+nza]-zcmini[i+nza]) do
            begin
            if (freq[j] > fmax) then
              begin
              jmax := j+round(zcmini[i+nza]);
              fmax := freq[j];
              end;
            end;
          x := format('%4d',[jmax]);
          outf.Write(pchar(x)^,length(x));
(* impute last observed or mode: *)
          for j:=1 to n do if miszc(j,i+nza,h) then
            begin
            cf := CarryMissingForward and (h >= 2);
            if cf then
              begin
(* now let cf = (there is any nonmissing value previous to h) *)
              cf := false;
              for k:=h-1 downto firstperiod do
                if (not miszc(j,i+nza,k)) then cf := true;
              end;
            if cf then zc[j,i+nza,h] := zc[j,i+nza,h-1] else zc[j,i+nza,h] := jmax;
            end; (* for j *)
          end; (* for h *)
        outf.Write(pchar(eol)^,length(eol));
        end; (* if anymissings; end imputation added 3.17u *)

    end; (* for i =1 to numzc1 *)

    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    x := format('%s%d%s',['A total of ',numzc1,' exogenous changing actor covariate']);
    outf.Write(pchar(x)^,length(x));
    if (numzc1 = 1) then begin
      x := '.'+eol;
      outf.Write(pchar(x)^,length(x));
    end
     else begin
      x := 's.'+eol;
      outf.Write(pchar(x)^,length(x));
     end;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));

    (* report missings *)
    x := 'Number of missing cases per period:'+eol;
    outf.Write(pchar(x)^,length(x));
    anymissings := false;
    x := ' period   ';
    outf.Write(pchar(x)^,length(x));
    for h:=1 to observations-1 do begin
      x := format('%9d',[h+firstperiod-1]);
      outf.Write(pchar(x)^,length(x));
    end;
    x := '       overall'+eol;
    outf.Write(pchar(x)^,length(x));
    for i:=1 to numzc1
    do begin
      x := AlignString(Znames[nz+nza+i],namelength+1);
      outf.Write(pchar(x)^,length(x));
      nzcmeani[nza+i] := 0;
      for h:= 1 to observations-1
      do begin
        inc(nzcmeani[nza+i],nzcmean[nza+i,h]);
        x := format('%6d%s',[n-nzcmean[nza+i,h],'   ']);
        outf.Write(pchar(x)^,length(x));
      end;
      if ((n - nzcmeani[nza+i]) >= 1) then anymissings := true;
      x := format('%7d%s%4.1f%s',[n*(observations-1)-nzcmeani[nza+i],'      (',
                   (100*(n*(observations-1)-nzcmeani[nza+i])/(n*(observations-1))),' %)']) +eol;
      outf.Write(pchar(x)^,length(x));
    end; (* for i *)
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));

    (* range checks *)
    SetLength(zwarning,numzc1+1,observations);
    SetLength(zwarnings,numzc1+1);
    Setlength(zzmin,numzc1+1,observations);
    Setlength(zzmax,numzc1+1,observations);
    for i:= 0 to numzc1 do
      begin
      zwarnings[i] := 0;
      for h:=0 to observations-1 do zwarning[i,h] := false;
      end;
    for i:=1 to numzc1 do
      begin
(* AA start *)
      poszvar[nz+nza+i] := false;
(* AA end *)
      for h:=1 to observations-1 do
        begin
        zzmin[i,h]:=1E30;
        zzmax[i,h]:=-zzmin[i,h];
        for j:=1 to n do
         if not miszc(j,nza+i,h) then
          begin
          if zzmin[i,h]>zc[j,nza+i,h+firstperiod-1] then zzmin[i,h]:=zc[j,nza+i,h+firstperiod-1];
          if zzmax[i,h]<zc[j,nza+i,h+firstperiod-1] then zzmax[i,h]:=zc[j,nza+i,h+firstperiod-1];
          end;
        if zzmin[i,h]=zzmax[i,h] (* range check *) then
          begin
          zwarning[i,h] := true;
          inc(zwarnings[i]);
(* AA start *)
          end else poszvar[nz+nza+i] := true;
(* AA end *)
        if nzcmean[nza+i,h]=0 then
          begin
          x := format('%s%d%s%s',['WARNING: No valid scores in period ',h+firstperiod-1,
            ' on variable ',UpperCase(Znames[nz+nza+i])])+eol;
          outf.Write(pchar(x)^,length(x));
(* No implication that this variable should not be used. *)
          //writeln(outf);
          outf.Write(pchar(eol)^,length(eol));
          end;
        end; (* h *)
     // writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      end; (* i *)

      //writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      (* report extrema and means *)
(* AA change *)
      x := 'Information about changing covariates';
      outf.Write(pchar(x)^,length(x));
      if anymissings
        then begin
          x := ' (missings not counted):'+eol;
          outf.Write(pchar(x)^,length(x));
        end
        else begin
          x := ':'+eol;
          outf.Write(pchar(x)^,length(x));
        end;
      x := '             minimum  maximum     mean '+eol;
      outf.Write(pchar(x)^,length(x));

    for i:=1 to numzc1
    do begin
      x := AlignString(Znames[nz+nza+i],namelength+1)+eol;
      outf.Write(pchar(x)^,length(x));
      zcmeani[nza+i] := 0.0;
      nzcmeani[nza+i] := 0;
      for h:= 1 to observations-1 do
        begin
        x := format('%s%3d%s',['  period ',h+firstperiod-1,'  ']);
        outf.Write(pchar(x)^,length(x));
        if nzcmean[i+nza,h]<>0 then
          begin
          zcmeani[i+nza] := zcmeani[i+nza] + zcmean[i+nza,h];
          zcmean[i+nza,h] := zcmean[i+nza,h]/nzcmean[i+nza,h];
          inc(nzcmeani[nza+i],nzcmean[nza+i,h]);
          x := format('%6.1f%s%7.1f%s%10.3f%s',[zzmin[i,h],' ',zzmax[i,h],' ',
            zcmean[i+nza,h],' ']) +eol;
          outf.Write(pchar(x)^,length(x));
          end
         else begin
          x := '    ---  '+eol;
          outf.Write(pchar(x)^,length(x));
         end;
        end;
      zcmeani[i+nza] := zcmeani[i+nza]/nzcmeani[nza+i];
      x := format('%s%8.3f',['Overall                        ',zcmeani[i+nza]])+eol;
      outf.Write(pchar(x)^,length(x));
    //  writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
      if (zwarnings[i] > 0) then
        begin
        x := 'WARNING: Valid scores have range zero on variable '+UpperCase(Znames[nz+nza+i]);
        outf.Write(pchar(x)^,length(x));
        if (zwarnings[i] = 1) then outf.write(pchar(string(' in period'))^,10) else
            outf.write(pchar(string(' in periods'))^,11);
        tempinteger := 0;
        for h:=1 to observations - 1 do
          begin
          if zwarning[i,h] then
            begin
            x := format('%s%d',[' ',h+firstperiod-1]);
            outf.Write(pchar(x)^,length(x));
            inc(tempinteger);
            end;
          if (tempinteger < zwarnings[i]) then outf.write(pchar(string(','))^,1) else begin
            x := '.'+eol;
            outf.Write(pchar(x)^,length(x));
          end;
          end;
       // writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
        end;
(*     no implication that this variable should not be used;
       it may be variable in other periods; and it could define the time trend. *)
    end; (* for i *)
   // writeln(outf);
    outf.Write(pchar(eol)^,length(eol));

    (* report averages and runtime-centring of changing actor covariates *)

    for h:=1 to numzc1
    do for i:=1 to n
       do begin
       for j:=firstperiod to thisperiod-1
       do if not miszc(i,h+nza,j)
       then zc[i,h+nza,j] := zc[i,h+nza,j] - zcmeani[h+nza]
       else zc[i,h+nza,j] := 0.0; (* missings get score zero *)
    end; (* i *)
   // writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    x := 'The overall mean value';
    outf.Write(pchar(x)^,length(x));
    if (numzc1 = 1) then outf.write(pchar(string(' is '))^,4) else outf.write(pchar(string('s are '))^,6);
    x := 'subtracted from the covariate';
    outf.Write(pchar(x)^,length(x));
    if (numzc1 = 1) then begin
      x := '.'+eol;
      outf.Write(pchar(x)^,length(x));
    end
      else begin
        x := 's.'+eol;
        outf.Write(pchar(x)^,length(x));
      end;
    //writeln(outf)
    outf.Write(pchar(eol)^,length(eol));
  end; (* Read changing actor covariates *)

  (* READ CONSTANT DYADIC COVARIATES *)
  if (nzz1 > 0)
  then begin
    Heading(2,outf,'Reading constant dyadic covariates.');
    InitForm.MessageLabel.caption := 'reading constant dyadic covariates';
    Application.ProcessMessages;
{$if defined(batch)}
       Writeln(Initform.MessageLabel.caption);
{$ifend}
    nextsection(nr);
    if halterr then exit;
    if nr<>6
    then begin
      x := format('%s%d%s',['Unexpected section encountered: @',nr,' instead of @6 (constant dyadic covariates).'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Unexpected section.',outf,lf);
      halterr := true;
      exit
    end;

    Setlength(zmin,nzz1+1);
    Setlength(zmax,nzz1+1);

    (* read data files *)
    for i:=1 to nzz1
    do begin
      zzmean[i] := 0.0; (* initialise average.. *)
      nzmean[i] := 0;   (* ..and # valid cases  *)
      zmin[i]:= 1E30;
      zmax[i]:=-zmin[i];

      x := 'Constant dyadic covariate:'+eol;
      lf.Write(pchar(x)^,length(x));
      readln(f,dfn);
      tpajek := true;
      Startdfile(outf,lf,tpajek);
      if halterr then exit;

      (* missing data identification: *)
      readln(f,dumstring);  (* row with code for missing data on this variable
                               n.b. StOCNET allows for multiple missing codes here and writes them all
                                    to the .IN-file, but SIENA only accepts the first in the row. *)
      (* transform string with missing code into a number *)
      val(dumstring,cmisz[1],dumint);
      if dumint<>0
      then begin
        dumstring := copy(dumstring,1,dumint-1);
        val(dumstring,cmisz[1],dumint);
      end;
      if dumint<>0 then cmisz[1] := -9; (* in case no missing code is given, take this one *)
      (* read variable label *)
      readln(f,ZZNames[i]);
      slength := length(ZZNames[i]);
      if slength=0
      then begin
        str(i,ZZNames[i]);
        ZZNames[i]:='dyad-cov-'+ZZNames[i]
      end;
//      for h:= slength + 1 to namelength do ZZNames[i] := ZZNames[i]+' ';
      x := 'Dyadic covariate named '+UpperCase(ZZNames[i])+' read from file '+dfn+'.'+eol;
      outf.Write(pchar(x)^,length(x));
      (* report missing code *)
      if dumint=0
      then begin
        x := format('%s%d%s',['Value ',cmisz[1],' is the code for missing data.'])+eol;
        outf.Write(pchar(x)^,length(x));
      end
      else begin
        x := 'No code for missing data is specified; SIENA will treat scores of -9 as missing data.'+eol;
        outf.Write(pchar(x)^,length(x));
      end;

      if tpajek then
        begin
        ert := '';
        ReadPajekDataFileToMatrix(dfn,n,ert,extra,zdum,vmin,vmax);
        x := 'Reading Pajek format file '+dfn+'.'+eol;
        outf.Write(pchar(x)^,length(x));
        x := format('%s%d%s%d%s',['Values range from ',vmin,' to ',vmax,'.'])+eol;
        outf.Write(pchar(x)^,length(x));
        if (ert > '')  then
          begin
          x := 'Error when reading tpajek network file '+dfn+'.'+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'Error message: '+ert+eol;
          outf.Write(pchar(x)^,length(x));
          Halterror('Error when reading tpajek input file.',outf,lf);
          halterr := true;
          exit;
          end;
        if extra then wwriteln(outf,'*** Warning: data files contains ties with vertices larger than '
                                   +inttostr(n)+'. ***');  
        end; (* if tpajek *)

      (* read one data file *)
      for j:=1 to n do
        begin
        for h:=1 to n do
          begin
          if tpajek then
            begin
            tempinteger := zdum[j,h];
            tempreal := zdum[j,h]; // added 3.181
            end else
            begin
            read(df,tempreal);
{            tempinteger := trunc(tempreal);
            if (abs(tempinteger-tempreal)>1e-6)
            then begin
              x := format('%s%d%s%d%s%s%s%20.6f%s',['Value ',h,' on line ',j,' of file ',dfn,
                ' has value ',tempreal,' .'])+eol;
             outf.Write(pchar(x)^,length(x));
             halterr := true;
            end;
            if ((tempinteger < 0) or (tempinteger >= 256))
            then begin
              x := format('%s%d%s%d%s%s%s%d%s',['Value ',h,' on line ',j,' of file ',dfn,' has value ',
                tempinteger,'.'])+eol;
              outf.Write(pchar(x)^,length(x));
              halterr := true;
            end;
            if halterr
            then begin
              x := 'However, for dyadic covariates only integer values from 0 to 255 are allowed.'+eol;
              outf.Write(pchar(x)^,length(x));
              x := 'Please use another dyadic covariate file!'+eol;
              outf.Write(pchar(x)^,length(x));
              Halterror('Wrong value for dyadic covariate',outf,lf);
              exit;
            end;
}
          end; (* if not tpajek *)
//         zz[j,h,i] := tempinteger;
          zz[j,h,i] := tempreal;
          setmiszz(j,h,i,tempreal=cmisz[1]);
          if not miszz(j,h,i)
          then begin
            if (j <> h)
            then begin
              zzmean[i] := zzmean[i] + zz[j,h,i];
              inc(nzmean[i]);
              if zmin[i]>zz[j,h,i] then zmin[i]:=zz[j,h,i];
              if zmax[i]<zz[j,h,i] then zmax[i]:=zz[j,h,i]
            end
          end
          else zz[j,h,i]:=0 (* score zero for missings *)
        end; (* for h *)
      end; (* for j *)
      if (not tpajek) then closefile(df);

    end; (* for i *)

    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    x := format('%s%d%s',['A total of ',nzz1,' dyadic covariate']);
    outf.Write(pchar(x)^,length(x));
    if nzz1=1
    then begin
      x :='.'+eol;
      outf.Write(pchar(x)^,length(x));
    end
    else begin
      x := 's.'+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));

    (* report missings *)
    x := 'Number of tie variables with missing data:'+eol;
    outf.Write(pchar(x)^,length(x));
    anymissings := false;
    for h:=1 to nzz1
    do begin
      x := format('%s%4d%s%4.1f%s',[AlignString(Zznames[h],namelength+1),n*(n-1)-nzmean[h],
          '   (',100*(n*(n-1)-nzmean[h])/(n*(n-1)),' %)'])+eol;
      outf.Write(pchar(x)^,length(x));
      if ((n*(n-1)-nzmean[h]) >= 1) then anymissings := true;
    end;
    //writeln(outf);
    if (multi and anymissings) then
      begin
      x := 'In multi-group project, constant covariates are represented as changing covariates.'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'Therefore, missings are treated as observed data!!!'+eol;
      outf.Write(pchar(x)^,length(x));
      end;
    outf.Write(pchar(eol)^,length(eol));

    (* range checks & calculation of means *)
    for i := 1 to nzz1
    do begin
      if zmin[i]=zmax[i]
      then begin
        x := 'WARNING: Valid scores have range zero on dyadic variable '+UpperCase(ZZnames[i])+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'DO NOT INCLUDE THIS NON-INFORMATIVE COVARIATE IN THE DATA SPECIFICATION!'+eol;
        if (not multi) then outf.Write(pchar(x)^,length(x));
        //writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
      end;
      if nzmean[i]=0
      then begin
        x := 'WARNING: No valid scores on dyadic variable '+UpperCase(ZZnames[i])+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'DO NOT INCLUDE THIS NON-INFORMATIVE COVARIATE IN THE DATA SPECIFICATION!'+eol;
        outf.Write(pchar(x)^,length(x));
        //writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
      end
      else zzmean[i]:=zzmean[i]/nzmean[i]
    end;

    (* report averages and runtime-centring of dyadic covariates *)
    x := 'Means of covariates';
    outf.Write(pchar(x)^,length(x));
    if anymissings
      then begin
        x :=' (missings not counted):'+eol;
        outf.Write(pchar(x)^,length(x));
      end
      else begin
        x := ':'+eol;
        outf.Write(pchar(x)^,length(x));
      end;
    x := '               minimum  maximum      mean '+eol;
    outf.Write(pchar(x)^,length(x));
    for h:=1 to nzz1
    do if nzmean[h]<>0
      then begin
        x := format('%s%8.3f%s%8.3f%s%10.5f',[AlignString(zznames[h],namelength+1),
                       zmin[h],' ',zmax[h],'  ',zzmean[h]])+eol;
        outf.Write(pchar(x)^,length(x));
      end
      else begin
        x := AlignString(zznames[h],namelength+1)+'    ---'+eol;
        outf.Write(pchar(x)^,length(x));
      end;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    if (nzz1 = 1) then begin
      x := 'This mean value is ';
      outf.Write(pchar(x)^,length(x));
    end
     else begin
      x := 'These mean values are ';
      outf.Write(pchar(x)^,length(x));
     end;
    x := 'subtracted from the dyadic covariate';
    outf.Write(pchar(x)^,length(x));
    if (nzz1 = 1) then begin
        x :='.'+eol;
        outf.Write(pchar(x)^,length(x));
      end
     else begin
      x :='s.'+eol;
      outf.Write(pchar(x)^,length(x));
     end;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));

  end; (* if nzz1 > 0; Read constant dyadic covariate files *)

  (* READ EXOGENOUS CHANGING DYADIC COVARIATES;
     for the time being, these do not allow missing values yet. *)

  if (nzzc1 > 0)
  then begin
    Heading(2,outf,'Reading exogenous changing dyadic covariates.');
    InitForm.MessageLabel.caption := 'reading changing dyadic covariates';
    Application.ProcessMessages;
{$if defined(batch)}
       Writeln(Initform.MessageLabel.caption);
{$ifend}
    x := 'Note that no missing values are considered yet for changing dyadic covariates.'+eol;
    outf.Write(pchar(x)^,length(x));
    anymissings := false;
    nextsection(nr);
    if halterr then exit;
    if nr<>7
    then begin
      x := format('%s%d%s',['Unexpected section encountered: @',nr,
                   ' instead of @7 (exogenous changing dyadic covariates).'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Unexpected section.',outf,lf);
      halterr := true;
      exit
    end;
    (* initialise variable-specific arrays *)
    for i:=1 to nzzc1 do
      begin
      zzmean[nzz1+i] := 0;
      nzmean[nzz1+i] := 0;
      end;

    (* read data files *)
    for i:=1 to nzzc1
    do begin
      x := format('%s%d',['Reading the ',i]);
      outf.Write(pchar(x)^,length(x));
      case i of
        1 : outf.write(pchar(string('st'))^,2);
        2 : outf.write(pchar(string('nd'))^,2);
        3 : outf.write(pchar(string('rd'))^,2);
        else outf.write(pchar(string('th'))^,2)
      end;
      x := ' exogenous changing dyadic covariate:'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'Changing dyadic covariate:'+eol;
      lf.Write(pchar(x)^,length(x));
      for io:= firstperiod to thisperiod-1
      do begin
        readln(f,dfn);
        tpajek := true;
        Startdfile(outf,lf,tpajek);
        if halterr then exit;
        x:= format('%s%s%s%d%s',['File ',dfn,' contains covariate for period ',io,'.'])+eol;
        outf.Write(pchar(x)^,length(x));

        if tpajek then
          begin
          ert := '';
          ReadPajekDataFileToMatrix(dfn,n,ert,extra,zdum,valmin,valplus);
          x := 'Reading tpajek format file '+dfn+'.'+eol;
          outf.Write(pchar(x)^,length(x));
          x := format('%s%d%s%d%s',['Values range from ',vmin,' to ',vmax,'.'])+eol;
          outf.Write(pchar(x)^,length(x));
          if (ert > '')  then
            begin
            x := 'Error when reading tpajek network file '+dfn+'.'+eol;
            outf.Write(pchar(x)^,length(x));
            x := 'Error message: '+ert+eol;
            outf.Write(pchar(x)^,length(x));
            Halterror('Error when reading tpajek input file.',outf,lf);
            halterr := true;
            exit;
            end;
          end; (* if tpajek *)

        for j:=1 to n do
          begin
          for h:=1 to n do
            begin
            if tpajek then
              begin
              tempinteger := zdum[j,h];
              tempreal := zdum[j,h];
              end else
              begin
              read(df,tempreal);
              tempinteger := trunc(tempreal);
(* skip 3.181
              if (abs(tempinteger-tempreal)>1e-6)
              then begin
                x := format('%s%d%s%d%s%s%s%20.6f%s',[ 'Value ',h,' on line ',j,' of file ',
                  dfn,' has value ',tempreal,' .'])+eol;
                outf.Write(pchar(x)^,length(x));
                halterr := true;
              end;
              if ((tempinteger < 0) or (tempinteger >= 256))
              then begin
                x := format('%s%d%s%d%s%s%s%d%s',[ 'Value ',h,' on line ',j,' of file ',
                  dfn,' has value ',tempinteger,'.'])+eol;
                outf.Write(pchar(x)^,length(x));
                halterr := true;
              end;
              if halterr
              then begin
                x := 'However, for dyadic covariates only integer values from 0 to 255 are allowed.'+eol;
                outf.Write(pchar(x)^,length(x));
                x := 'Please use another dyadic covariate file!'+eol;
                outf.Write(pchar(x)^,length(x));
                Halterror('Wrong value for dyadic covariate',outf,lf);
                exit;
              end;
*)
            end; (* if not tpajek *)
(*            zzc[j,h,i,io] := tempinteger;
             zzmean[nzz1+i] := zzmean[nzz1+i] + tempinteger;
changed 3.181: *)
            zzc[j,h,i,io] := tempreal;
            zzmean[nzz1+i] := zzmean[nzz1+i] + tempreal;
            inc(nzmean[nzz1+i]);
          end; (* for h *)
        end; (* for j *)
        if (not tpajek) then closefile(df);
        readln(f,dumstring);  (* code for missing data on this variable -- missings handling here not yet functional *)
      end; (* for io -- the period loop *)
      readln(f,ZZNames[nzz1+i]);
      slength := length(ZZNames[nzz1+i]);
      if slength=0
      then begin
        str(nzz1+i,ZZNames[nzz1+i]);
        ZZNames[nzz1+i]:='dyad-cov-'+ZZNames[nzz1+i]
      end;
//      for h:= slength + 1 to namelength do ZZNames[nzz1+i] := ZZNames[nzz1+i]+' ';
      x:= 'Name of this covariate : '+ZZNames[nzz1+i]+'.'+eol;
      outf.Write(pchar(x)^,length(x));
    end; (* for i -- the covariate number loop *)

    (* calculation of means *)
    for i := 1 to nzzc1
    do begin
      if nzmean[nzz1+i]=0
      then begin
        x := 'WARNING: No valid scores on changing dyadic variable '+UpperCase(ZZnames[nzz1+i])+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'DO NOT INCLUDE THIS NON-INFORMATIVE COVARIATE IN THE DATA SPECIFICATION!'+eol;
        if (not multi) then outf.Write(pchar(x)^,length(x));
        //writeln(outf);
        outf.Write(pchar(eol)^,length(eol));
      end
      else zzmean[nzz1+i]:=zzmean[nzz1+i]/nzmean[nzz1+i]
    end;

    (* report averages and runtime-centring of changing dyadic covariates *)
    x := 'Means of covariate';
    outf.Write(pchar(x)^,length(x));
    if (nzzc1 >= 2) then outf.write(pchar(string('s'))^,1);
    if anymissings then begin
      x :=' (missings not counted):'+eol;
      outf.Write(pchar(x)^,length(x));
    end
       else begin
        x := ':'+eol;
        outf.Write(pchar(x)^,length(x));
       end;
    for h:=1 to nzzc1
    do if (nzmean[nzz1+h] > 0)
    then begin
      x := format('%s%8.3f',[AlignString(zznames[nzz1+h],namelength),zzmean[nzz1+h]])+eol;
      outf.Write(pchar(x)^,length(x));
    end
    else begin
      x := AlignString(zznames[nzz1+h],namelength)+'    ---'+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    if (nzzc1 = 1) then begin
      x := 'This global mean value is ';
      outf.Write(pchar(x)^,length(x));
    end
     else begin
      x := 'These global mean values are ';
       outf.Write(pchar(x)^,length(x));
     end;
    x :='subtracted from the changing dyadic covariate';
    outf.Write(pchar(x)^,length(x));
    if (nzzc1 = 1) then begin
      x := '.'+eol;
      outf.Write(pchar(x)^,length(x));
    end
     else begin
      x := 's.'+eol;
      outf.Write(pchar(x)^,length(x));
     end;
   // writeln(outf);
    outf.Write(pchar(eol)^,length(eol));


    //writeln(lf);
   lf.Write(pchar(eol)^,length(eol));
    //writeln(outf)
    outf.Write(pchar(eol)^,length(eol));
  end; (* if nzzc1 > 0; Read exogenous changing dyadic covariate files *)

  (* MH: READ INFORMATION ABOUT EXOGENOUS EVENTS (i.e., network composition change) *)
  if (exo)
  then begin
    Heading(2,outf,'Reading file with times of composition change.');
    nextsection(nr);
    if halterr then exit;
    if nr<>8
    then begin
      x := format('%s%d%s',['Unexpected section encountered: @',nr,' instead of @8 (network composition change).'])+eol;
      outf.Write(pchar(x)^,length(x));
      Halterror('Unexpected section.',outf,lf);
      halterr := true;
      exit
    end;

    (* initialise variable-specific arrays *)
    setlength(z_exo,observations,3);
    for i:=1 to observations-1
    do for j:=1 to 2 do z_exo[i,j] := 0;

    (* read data files *)
    x :=  'File with times of composition change.'+eol;
    lf.Write(pchar(x)^,length(x));
    readln(f,dfn);
    Startdfile(outf, lf,trypajek);
    if halterr then exit;
    x := 'File with times of composition change '+dfn+'.'+eol;
    outf.Write(pchar(x)^,length(x));
    for i:=1 to n
    do begin
      for j:=1 to 2
      do begin
        read(df,texo,f_exo);
        t_exo := round(texo);
        if (f_exo = 0.0) then f_exo := 0.0001; (* event just after begin of period *)
        if (f_exo = 1.0) then f_exo := 0.9999; (* event just before end of period *)
        zx[i,j] := t_exo+f_exo;
         (* Checks on joining and leaving times *)
        if ((t_exo < 0) or (t_exo > observations))
        then begin
          x := format(' %s%d%s%s%s',['Actor ',i,' in file ',dfn,' wants to ']);
          outf.Write(pchar(x)^,length(x));
          if (j=1)
          then begin
            x:= format('%s%d%s',[ 'join in period ', t_exo ,'.'])+eol;
            outf.Write(pchar(x)^,length(x));
          end
          else begin
            x := format('%s%d%s',[ 'leave in period ', t_exo, '.'])+eol;
            outf.Write(pchar(x)^,length(x));
          end;
          x := format('%s%d%s',['This is not an existing period; periods range from 0 to ',
                        observations,'.'])+eol;
          outf.Write(pchar(x)^,length(x));
          halterr := true;
        end;
        if ((f_exo < 0.0) or (f_exo > 1.0))
        then begin
          x := format('%s%d%s%s%s',['Actor ',i,' in file ',dfn,' wants to ']);
          outf.Write(pchar(x)^,length(x));
          if (j=1)
          then begin
            x := format('%s%4.2f%s',['join at time ', f_exo, '.'])+eol;
            outf.Write(pchar(x)^,length(x));
            x := 'This is an incorrect time point; joining times range from 0.0 to 1.0.'+eol;
            outf.Write(pchar(x)^,length(x));
          end
          else begin
            x := format('%s%4.2f%s',['leave at time ', f_exo, '.'])+eol;
            outf.Write(pchar(x)^,length(x));
            x := 'This is an incorrect time point; leaving times range from 0.0 to 1.0.'+eol;
            outf.Write(pchar(x)^,length(x));
          end;
          halterr := true;
        end;
      end; (* for j*)

      if (trunc(zx[i,1]) = trunc(zx[i,2]))
      then begin
        x := format('%s%d%s%d%s',['Actor ',i,' joins and leaves in the same period (period ',
                      trunc(zx[i,1]),').'])+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'This is not allowed.'+eol;
        outf.Write(pchar(x)^,length(x));
        halterr := true;
      end;
      if halterr
      then begin
        x := 'Please use another file with times of composition change.'+eol;
        outf.Write(pchar(x)^,length(x));
        Halterror('Wrong values for times of composition change.',outf,lf);
        exit
      end;

      if (zx[i,1] > 1)
      then begin
        x := format('%s%d%s%4.2f%s',['Actor ',i,' joins network at time ', zx[i,1],'.'])+eol;
        outf.Write(pchar(x)^,length(x));
        inc(z_exo[trunc(zx[i,1]),1]);
      end;
      if (zx[i,2] < observations)
      then begin
        x := format('%s%d%s%4.2f%s',[ 'Actor ',i,' leaves network at time ', zx[i,2],'.'])+eol;
        outf.Write(pchar(x)^,length(x));
        inc(z_exo[trunc(zx[i,2]),2]);
      end;
      readln(df);
    end; (* for i *)

    closefile(df);
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    for h:=1 to observations-1
    do begin
      x := format('%s%d%s%d',[ 'In period ',h,', ',z_exo[h,1]]);
      outf.Write(pchar(x)^,length(x));
      if (z_exo[h,1] = 1) then outf.write(pchar(string(' actor '))^,7)
                          else outf.write(pchar(string(' actors '))^,8);
      x := format('%s%d',[ 'joined and ',z_exo[h,2]]);
      outf.Write(pchar(x)^,length(x));
      if (z_exo[h,2] = 1) then outf.write(pchar(string(' actor '))^,7)
                          else outf.write(pchar(string(' actors '))^,8);
      x := 'left the network.'+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    if (nzx = 4) then ConvertExoToStruct(firstperiod, outf);
  end; (* if exo -- Read exogenous events on network composition change *)

  closefile(f); (* that is the project .IN file *)


  if multiplus then
    begin
    SetNumberOfPeriods(nmx,nz1+nzc1+multitot-1,nzz1,nzz1 + nzzc1,totper+1);
    end
   else
    begin
    SetNumberOfPeriods(nmx,nz1+nzc1,nzz1,nzz1 + nzzc1,totper+1);
    end;

    if multi then
    begin
      begin
      nzc := nz + nzc1;
      nzzc := nzz1 + nzzc1;
      end;
    (* put constant actor covariates at the end of the changing actor covariates *)
    if (nz1 >= 1) then
      begin
      for h:=1 to nz1 do for i:=1 to n do
        begin
        missi := misz(i,h);
        for per :=firstperiod to thisperiod do
          begin
          zc[i,h+nzc1,per] := z[i,h];
          setmiszc(i,h+nzc1,per,missi);
          end;
        end;
      end;
(* put constant dyadic covariates at the end of the changing dyadic covariates *)
    if (nzz1 >= 1) then
      begin
      for h:=1 to nzz1 do for i:=1 to n do for j:=1 to n do
        begin
//         missi := miszz(i,j,h);     not implemented yet
        for per :=firstperiod to thisperiod do
          begin
          zzc[i,j,h+nzzc1,per] := zz[i,j,h];          // hier
//          setmiszc(i,h+nzzc1,per,missi);  not implemented yet
          end;
        end;
      end;
    end
   else
    begin
    nzc := nzc1;
    nzzc := nzzc1;
    end;
  if halterr then exit;

  if (not lastfile) then exit;
// For multi-group projects, the preceding is carried out for all subprojects.
// The following always is executed only once.

  observations := thisperiod;
  if multi then
    begin
    nz := 0;
    nzz := 0;
    if (nz1 > nzz1) then Setlength(dumnames,nz1+1) else Setlength(dumnames,nzz1+1);
(* put constant actor covariate names after of the changing actor covariates *)
    if (nz1 >= 1) then
      begin
      for i:=1 to nz1 do dumnames[i] := ZNames[i];
      for i:=1 to nzc1 do ZNames[i] := ZNames[i+nz1];
      for i:=1 to nz1 do ZNames[nzc1+i] := dumnames[i];
      end;
    if multiplus then
      begin
(* put subproject dummies at the end of the changing actor covariates *)
      for h:=2 to multitot do
        begin
        if (h < multitot) then j := firstobs[h+1]-1 else j := observations;
        ZNames[nzc+h-1] := 'Subproject '+inttostr(h);
        for i:=1 to nmax do for per :=1 to observations-1 do
          if ((per >= firstobs[h]) and (per <= j)) then zc[i,h+nzc-1,per] := 1.0
            else zc[i,h+nzc-1,per] := 0.0;
        end;
      nzc := nzc + multitot-1;
      if (multitot <= 2) then
        begin
        wwriteln(outf,'Dummy actor variable Subproject 2 is added to the multi-group project.');
        wwriteln(outf,'This variable is not centered, in contrast to other actor variables.');
        end
       else
        begin
        wwriteln(outf,'Dummy actor variables Subproject 2 to Subproject '+inttostr(multitot)+
                      ' are added to the multi-group project.');
        wwriteln(outf,'These variables are not centered, in contrast to other actor variables.');
        end;
      end; (* if multiplus *)
(* put constant dyadic covariate names at the end of the changing dyadic covariates *)
    if (nzz1 >= 1) then
      begin
      for i:=1 to nzz1 do dumnames[i] := ZZNames[i];
      for i:=1 to nzzc1 do ZZNames[i] := ZZNames[i+nzz1];
      for i:=1 to nzz1 do ZZNames[nzzc1+i] := dumnames[i];
      end;
      wwriteln(outf,'');
      wwrite(outf,'Series of observations ');
      if multi then wwriteln(outf,'for the multi-group project: ')
        else wwriteln(outf,'');
      for h:=1 to observations do
        begin
        wwrite(outf,inttostr(h)+' ');
        if skipi[h] then wwriteln(outf,'');
        end;
      end
   else
    begin (* if not multi *)
    nzz := nzz1;
    nzzc := nzzc1;
    end;
  Checkmaxima(halterr,outf,lf);
  Calculations(true, outf);  (* defined in S_Dat *)

  for h :=1 to ccmax do param_effect[h]:=high(integer);
(* This is necessary so that the default values are used in S_Effects;
   see procedure S_Effects.nextpar *)

  n := nmax;
  NewNetworkEffects := TEffects.Create;
  if signed then DefineSignedNetworkEffects(NewNetworkEffects)
   else DefineNetworkEffects(NewNetworkEffects);

(* AA abcde start *)
  if (MaxFunctions > ccmax) then
    begin
    x := 'The number of statistics is too large.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Compile SIENA with a larger value of ccmax or use fewer covariates.'+eol;
    outf.Write(pchar(x)^,length(x));
    if halterr then exit;
    end;
(* Abcde AA end *)

  for h :=NewNetworkEffects.NumberEffects+1 to ccmax do param_effect[h] := 0;
  InitialAlphas(alpa_l,alpa_f,0,outf);

  (* WRITE DATA TO INTERNAL FILES AND SAY SO IN .LOG-File*)
  Heading(2,lf,'Internal data storage.');
  InitForm.MessageLabel.caption := 'internal data storage';
  Application.ProcessMessages;
{$if defined(batch)}
       Writeln(Initform.MessageLabel.caption);
{$ifend}

  WritePData(halterr,outf,lf);
  if halterr then exit;

(* Toms private utility for making a dyadic similarity file *)
(*      for h:=1 to nz do begin
        assignfile(fff,'var'+inttostr(h)+'.dat');
        rewrite(fff);
        for i:=1 to n do begin
          for j:=1 to n do write(fff,range[h] - abs(z[i,h]-z[j,h]):2:0);
          writeln(fff);
         end;
        closefile(fff);
       end; *)
(*
       for i:=1 to 150 do
         begin
         fffmat[i,i] := false;
         for j:=i+1 to 150 do
           if (random < 0.03) then
             begin
             fffmat[i,j] := true;
             fffmat[j,i] := true;
             end
            else
             begin
             fffmat[i,j] := false;
             fffmat[j,i] := false;
             end;
           end;
 *)
(*        assignfile(fff,'n150.txt');
        rewrite(fff);
        for i:=1 to 150 do begin
          for j:=1 to 150 do
            if fffmat[i,j] then write(fff,' 1') else write(fff,' 0');
          writeln(fff);
         end;
        closefile(fff);
*)


  (* Networks *)
  for h:=1 to observations do begin
  (* digraph data: *)
    x := 'Network data file';
    lf.Write(pchar(x)^,length(x));
    if (observations >=2) then begin
      x := format('%s%d',[' for observation ',h]);
      lf.Write(pchar(x)^,length(x));
    end;
    x := ' written for internal use to file '+pname+'.d';
    lf.Write(pchar(x)^,length(x));
    if (h < 10) then begin
      x := format('%s%d%s',['0',h,'.'])+eol;
      lf.Write(pchar(x)^,length(x));
    end
      else begin
        x := format('%d%s',[h,'.'])+eol;
        lf.Write(pchar(x)^,length(x));
      end;
  (* missing data: *)
    x := 'Network data missings';
    lf.Write(pchar(x)^,length(x));
    if (observations >= 2) then begin
      x:= format('%s%d',[' for observation ',h]);
      lf.Write(pchar(x)^,length(x));
    end;
    x := ' written for internal use to file '+pname+'.m';
    lf.Write(pchar(x)^,length(x));
    if (h < 10) then begin
      x := format('%s%d%s',['0',h,'.'])+eol;
      lf.Write(pchar(x)^,length(x));
    end
       else begin
        x := format('%d%s',[h,'.'])+eol;
        lf.Write(pchar(x)^,length(x));
       end;
  end;

  (* Constant actor covariates *)
  if (nz > 0) then begin
  (* covariate data *)
    x := 'Centered non-changing covariates written for internal use to file '+pname+'.dav.'+eol;
    lf.Write(pchar(x)^,length(x));
    x := 'The precision used is 4 decimal places.'+eol;
    lf.Write(pchar(x)^,length(x));
  (* missing data: *)
    x := 'Missings on non-changing covariates written for internal use to file '+pname+'.miv.'+eol;
    lf.Write(pchar(x)^,length(x));
    //writeln(lf)
    lf.Write(pchar(eol)^,length(eol));
  end; (* if nz > 0 *)

  (* Changing actor covariates and actor variables *)
  if (nzc > 0) then begin
  (* covariate data *)
    x := 'Changing covariates and variables written for internal use to file '+pname+'.dac.'+eol;
    lf.Write(pchar(x)^,length(x));
    x :='The precision used is 4 decimal places.'+eol;
    lf.Write(pchar(x)^,length(x));
  (* missing data: *)
    x := 'Missings on changing covariates and variables written for internal use to file '+pname+'.mac.'+eol;
    lf.Write(pchar(x)^,length(x));
    //writeln(lf)
    lf.Write(pchar(eol)^,length(eol));
  end; (* if nzc > 0 *)

  (* Constant dyadic covariates *)
  if (nzz > 0) then begin
  (* covariate data *)
    x := 'Constant dyadic covariates written for internal use to file';
    lf.Write(pchar(x)^,length(x));
    if (nzz > 1) then lf.write(pchar(string('s'))^,1);
    x := ' '+pname+'.z01';
    lf.Write(pchar(x)^,length(x));
    if (nzz < 10) then fname := pname + '.z0'+st(nzz)
      else fname := pname + '.z'+st(nzz);
    if (nzz > 1) then begin
      x := ' to '+fname;
      lf.Write(pchar(x)^,length(x));
    end;
    x := '.'+ eol;
    lf.Write(pchar(x)^,length(x));
   (* missing data *)
    x := 'Missing on dyadic covariates written for internal use to file';
    lf.Write(pchar(x)^,length(x));
    if (nzz > 1) then lf.write(pchar(string('s'))^,1);
    x := ' '+pname+'.n01';
    lf.Write(pchar(x)^,length(x));
    if (nzz < 10) then fname := pname + '.n0'+st(nzz)
      else fname := pname + '.n'+st(nzz);
    if (nzz > 1) then begin
      x := ' to '+fname;
      lf.Write(pchar(x)^,length(x));
    end;
    x := '.'+eol;
    lf.Write(pchar(x)^,length(x));
    //writeln(lf);
    lf.Write(pchar(eol)^,length(eol));
  end; (* if nzz > 0 *)

  (* Changing dyadic covariates *)
  if (nzzc > 0) then begin
  (* covariate data *)
    x :='Changing dyadic covariates written for internal use to file';
    lf.Write(pchar(x)^,length(x));
    if (nzzc > 1) then lf.write(pchar(string('s'))^,1);
    x := ' '+pname+'.c01';
    lf.Write(pchar(x)^,length(x));
    if (nzzc < 10) then fname := pname + '.c0'+st(nzzc)
      else fname := pname + '.z'+st(nzzc);
    if (nzzc > 1) then begin
      x := ' to '+fname;
      lf.Write(pchar(x)^,length(x));
    end;
    x := '.'+eol;
    lf.Write(pchar(x)^,length(x));
{   (* missing data *) not yet implemented
    write(lf,'Missing on dyadic covariates written for internal use to file');
    if (nzz > 1) then write(lf,'s');
    write(lf,' ',pname,'.n01');
    if (nzz < 10) then fname := pname + '.n0'+st(nzz)
      else fname := pname + '.n'+st(nzz);
    if (nzz > 1) then write(lf,' to ',fname);
    writeln(lf,'.');
    writeln(lf);
}
  end; (* if nzzc > 0 *)

  (* MH: exogenous events *)
  if (exo)
  then begin
     x :='Times of composition change written for internal use '+eol;
    lf.Write(pchar(x)^,length(x));
    x := 'to file '+pname+'.dex'+eol;
    lf.Write(pchar(x)^,length(x));
    //writeln(lf);
    lf.Write(pchar(eol)^,length(eol));
  end;

  if (nz + nzc >= 1)
  then begin
    x :='For the similarity variable calculated from each actor covariate, '+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'the mean is subtracted.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'These means are : '+eol;
    outf.Write(pchar(x)^,length(x));
    for h:=1 to (nz+nzc) do if poszvar[h] and (not compzvar[h]) then
      begin
      x :=format('%s%s%s%12.4f',['Similarity ',AlignString(ZNames[h],namelength),' : ',
        zsmean[h]])+eol;
      outf.Write(pchar(x)^,length(x));
      end;
   // writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
  end;

  if (observations >= 2)
  then begin
    x := 'The mean structural dissimilarity value subtracted in the'+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%14.4f%s',['balance calculations is ',balmean,' .'])+eol;
    outf.Write(pchar(x)^,length(x));
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
  end;

  (* Firstgain is initialised standard in S_Dat at 0.2.
     For observations = 1, however, experience has taught me (TS)
     that a smaller value is much better,
     because in the ERG models there often is greater instability
     in estimation near the true value of the estimate. Therefore
     firstgain must be smaller: *)
  nrun3 := 1000;
  if (observations = 1) then
    begin
    firstgain := 0.010;
    nrun3 := 500;
    end;
  if (nza >=1) then nrun3 := 1000;
end; (* Readgraph.ReadWriteData *)


begin (* unit S_Start *)
  siena1files := false;
end.
